/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** def.h: device solver class definitions
****************************************************************************/

#ifndef CLASS_H
#define CLASS_H

class Propagation;
class Cell;
class Source;

class Fourier_cell;
class Bound;
class BTE_structure;
class Fourier_structure;
class Device_structure;
class Block_coords;

/*==========================================================================*/
//stores only material specific properties
class Material{
private:
public:
    //default constructor 
    Material(){}
    
    //Material Class constructor
    Material(string mat_name, double k_conduct, double t_ref, double cv,
                double vg, double t_rel);

    //the material's name
    string name;

    //thermal conductivity of the material (W/m.K)
    double k_cond;
    //reference temperature for volumetric specific heat
    double t_ref_;
    //volumetric specific heat (Q = V * c_v * Delta T)
    double cv_;
    //group velocity, relaxation time
    double vg_, t_rel_;

    //For inputing for a single material
    int input(ifstream & fin);

};

/*==========================================================================*/
class Material_coords{
private:
public:
    //indices of the south-west-bottom cell
    int x, y, z;
    //number of cells in each direction
    int cells_x, cells_y, cells_z;

    Material material;

    //Constructor
    Material_coords(int x_index, int y_index, int z_index, int num_cells_x, 
                    int num_cells_y, int num_cells_z, Material mat);
};

/*==========================================================================*/
class Block_coords{
private:
public:
    //indices of the south-west-bottom cell
    int x, y, z;
    //number of cells in each direction
    int cells_x, cells_y, cells_z;

    //Constructor
    Block_coords(int,int,int,int,int,int);
};

/*==========================================================================*/
//holding information for BTE structures
class BTE_structure{
private:
	//residual from last iteration set
	double residual_last;

	//x,y,z coordinates of south-west-bottom corner of structure
    double x_coor, y_coor, z_coor;
    //x,y,z lengths of structure
    double length_x, length_y, length_z;

    //extents of the cells in each direction
    double del_x, del_y, del_z;

    //total amount of cells in x,y,z
    int cells_x, cells_y, cells_z;

	friend class Fourier_structure;
	friend class Cell;
	friend class Device_structure;

    int neigh[6];
    int f_cells_x, f_cells_y, f_cells_z;
    //thermal conductivity of the material (W/m.K)
    double k_cond;
    //indicates the boundaries that are connected to BTE cells
    //=1 if the direction has BTE cells at the boundary, using E_,W_,...
    int f_boundaries[6];
    //reference temperature for volumetric specific heat
    double t_ref_;
    //volumetric specific heat (Q = V * c_v * Delta T)
    double cv_;
    //group velocity, relaxation time
    double vg_, t_rel_;
    //a flag which sets all cells to be solved using bte
    int all_bte_;
    //a flag which sets all cells to be solved using fourier
    int all_fourier_;
    //set the q values of the cells according to the heat sources
    void define_sources();
    //set the q values of the fourier cells according to the heat sources
    void define_sources_f();
    //heat sources
    vector<Source> sources;
    //the material of the structure
    string material;
	//the index of the material of the structure
	int material_index;


    //Materials inside the BTE region
    vector<Material_coords> mat_coords;
	
	//BTE cells in the structure
    vector<vector<vector<Cell> > > cell;

    //cell edges and centroids
    vector<double> edge_x, edge_y, edge_z;
    vector<double> cen_x, cen_y, cen_z;

    //amount of control angles per octant
    int N_phi;
    int N_theta;
    int Nmax_angle;
    double del_theta;
    double del_phi;

    //the centroids of all the control angles
    vector< double > cen_theta;
    vector< double > cen_phi;

    //contains the direction propagation information
    vector<Propagation> propvec;

    //indicates the boundaries that are connected to Fourier cells
    //=1 if the direction has BTE cells at the boundary, using E_,W_,...
    int bte_boundaries[6];

	double T_ini; //initial temperature of the structure

    //for specular boundaries, =1 if specular, =0 else
    int spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end;

    //for diffuse boundaries, =1 if diffuse, =0 else
    int diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end;

    //for constant temperature boundaries, =1 if temp bound, =0 else
    int tbound_x_0, tbound_x_end, tbound_y_0, tbound_y_end, tbound_z_0,
        tbound_z_end;
    double tbx_0_val, tbx_end_val, tby_0_val, tby_end_val, tbz_0_val,
        tbz_end_val;

    //for insulating temperature boundaries
    int ins_x_0, ins_x_end,ins_y_0,ins_y_end,ins_z_0,ins_z_end;

    //for boundaries touching silicon dioxide
    int oxide_x_0, oxide_x_end, oxide_y_0, oxide_y_end, oxide_z_0, oxide_z_end;
    //transmittiviy factor from silicon to silicon dioxide
    vector<double> trans_factor;

    //set the q values of the cells according to the heat sources
    void define_sources(vector<Source> &sources);

    //the allowed error in f_angle, which is used to stop iterations
    double f_error;

public:

    BTE_structure();

    BTE_structure(double,double,double,
                  double,double,double,
                  int,int,int,
                  vector<Material_coords>&,
                  vector<double>&,vector<double>&,vector<double>&,
                  vector<double>&,vector<double>&,vector<double>&,
                  int[],
                  int,int,int,int,int,int,
                  int,int,int,int,int,int,
                  int,int,int,int,int,int,
                  double,double,double,double,double,double);
      
    //initialize the values for a structure
    void initialization();

    //get the structure data from input file
    int input(ifstream & fin);

    //set the 3d grid
    void set_grid(void);
	
	//set the type of solver (BTE or Fourier) for a structure
    void set_solver(Material& material);

	//check if structure is inside a volume
    bool isInside(double,double,double,double,double,double);

    //set the angular variables/component values
    void set_angular(void);

    //calculate e0 and T (in update section)
    void calc_e0_t(void);
    //obtain the matrix coefficients
    void obtain_coeffs(int l);
    //solve the structure
    int solve(int,Device_structure*);
    //execute the tdma algorithm 
    double lblTdma(int angle, double max_rel_change);
    //calculate the sum of the residuals
    double calc_residual(int angle);
    //update the e0 and temp for the specular and diffuse boundaries
    void calc_spec_diff_e0_t(void);

    //Specular boundary handling functions
    void specular();

    //diffuse boundary handling functions
    void diffuse();

    //constant temperature updating function
    void const_temp();

    //reset all coefficients to 0
    void reset(void);

    //set the initial guesses and conditions
    void ini_solve(void);

	//for updating the Fourier boundaries after BTE blocks have been solved
	void update_fourier_t_bound();
	//for updating the BTE boundaries after Fourier block has been solved
	void update_bte_t_bound(Fourier_structure &substrate,Block_coords &
																bte_coords);
    //update the BTE boundaries after Fourier block has been solved for
    //a semireflecting boundary
    void update_bte_semireflect();

	//update the fourier cell temperatures which are in BTE regions
    void update_fourier_t_cells(Fourier_structure & substrate,
                                                Block_coords & bte_coords);

	//update the BTE cell temperatures with new found Fourier values
	void initialize_bte_t_cells(Fourier_structure & substrate,
                                                Block_coords & bte_coords);

    //print output to file
    void output(char argv[]);
    
    //calculate power going from the bottom of the substrate
    double calc_qout(int index1, int index2, int dir);
    
    double calc_qsouth(int index1, int index2, int index3);

	//-------------transient functions---------------------//
	//get the peak temperature of this bte structure
	double get_peak();
	//resize all of the storage for the future
	void resize_trans_f_angle();
	//update all the energy densities with the new time values
	void update_f_angle_trans();
	//using runge-kutta with constant step time
	void trans_RK(int angle, double del_t);
	//solve the bte system for transient analysis
	void solve_transient(Device_structure* device_structure,double del_t);
	//update all the power sources depending on the time
	void set_Q(double time);
	//------------end of transient functions--------------//


};

/*==========================================================================*/
//holds information for Fourier structures
class Fourier_structure{
private:

    //residual from last iteration set
    double residual_last;

    //x,y,z coordinates of south-west-bottom corner of structure
    double x_coor, y_coor, z_coor;
    //x,y,z lengths of structure
    double length_x, length_y, length_z;

    //extents of the cells in each direction
    double del_x, del_y, del_z;

    friend class Fourier_cell;
    friend class Device_structure;
	friend class BTE_structure;

    //the materials inside the Fourier structure
    vector<Material_coords> mat_coords;

    //Fourier cells
    vector<vector<vector<Fourier_cell> > > f_cell;

    //cell edges and centroids
    vector<double> edge_x, edge_y, edge_z;
    vector<double> cen_x, cen_y, cen_z;

    //total amount of Fourier cells in x,y,z
    int f_cells_x, f_cells_y, f_cells_z;
    //indicates the boundaries that are connected to BTE cells
    //=1 if the direction has BTE cells at the boundary, using E_,W_,...
    int f_boundaries[6];

    double T_ini; //initial temperature of the structure

    //factor to maintain stability between BTE and Fourier solvers
    double k_stab;
    
    //for specular boundaries, =1 if specular, =0 else
    int spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end;

    //for diffuse boundaries, =1 if diffuse, =0 else
    int diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end;

    //for insulating temperature boundaries
    int ins_x_0, ins_x_end,ins_y_0,ins_y_end,ins_z_0,ins_z_end;

    //a flag which sets all cells to be solved using bte
    int all_bte_;

    //a flag which sets all cells to be solved using fourier
    int all_fourier_;

	//if transient analysis for bte simulation
	int transient_;
    //----------------------------transient variables----------------//
    //the transient step size
    double step_time;
    //the total simulation time
    double total_time;
    //----------------------end of transient variables----------------//

    //set the q values of the fourier cells according to the heat sources
    void define_sources_f(vector<Source> &sources);

	//for constant temperature boundaries, =1 if temp bound, =0 else
    int tbound_x_0, tbound_x_end, tbound_y_0, tbound_y_end, tbound_z_0,
        tbound_z_end;
    double tbx_0_val, tbx_end_val, tby_0_val, tby_end_val, tbz_0_val,
        tbz_end_val;

    //the allowed error in f_angle, which is used to stop iterations
    double f_error;

public:
    Fourier_structure();

	//calculate the vertical heat flow and print it
	void get_vertical_heat_flow(string argv);

	//calculate the horizontal x dimension heat flow and print it (heat flow
	//from right to left)
	void get_horizontal_x_heat_flow(string argv);
	//calculate the horizontal y dimension heat flow and print it (heat flow
    //from right to left)
    void get_horizontal_y_heat_flow(string argv);

    //initialize the values for a structure
    void initialization();

    //get the structure data from input file
    int input(ifstream & fin);

    //set the 3d grid
    void set_grid(void);

    //calculate the sum of the residuals for a single run for Fourier
    double calc_residual();
    //solve the structure using Fourier model
    int solve_fourier(int,Device_structure*);
    //solve the fourier system for transient analysis
    void solve_fourier(Device_structure*,double);
    //execute the tdma algorithm
    double lblTdma_fourier();
    //set the initial guesses and conditions
    void ini_solve_fourier(void);
    //reset all coefficients to 0
    void reset_fourier(void);

     //obtain the matrix coefficients and set insulating boundaries
    void obtain_coeffs_fourier();

	//condition the coefficients so the boundaries are set correctly
	void condition_coeffs_fourier();

    //set the conductances of the boundary cells to zero if they are insulating
    //boundaries
    void f_ins();

    //print output to file
    void output(char argv[]);

	//material of structure
    string material;
	//keeps the full contents of the material
	Material * material_full;

	//Function to initialize the null materials in the Fourier structure
	//and the material of all the other cells to the substrate material
	void set_null();

	//update the boundary conditions for the Fourier region for semi-reflecting
	//boundaries
	void update_Fourier_semireflect(BTE_structure & bte_systemupdate, 
												Block_coords & bte_coords);

    //calculate power going out of the whole system
    double calc_qout(vector<Block_coords> & bte_coords,
                     vector<BTE_structure> & bte_system);

    //calculate vetical heat flow of each cell
    void calc_heat_flow_cell_bottom(string output_file,
                                    vector<Block_coords> & bte_coords);

	void solve_transient(Device_structure* device_structure,double del_t);
    //solve the system for the current time step using the 4th order RK method
    void trans_RK_fourier(double);

	void update_T_trans();

};

/*==========================================================================*/
//Each cell centroid in the mesh has an associated class
class Cell{
private:
	double AP_,AWest_,AEast_,ASouth_,ANorth_,ABottom_,ATop_; //the coefficients of the linear equation to be solved
	double B_; //the B term

	double e0_; //angular average of energy density in the cell
	double T_; //the temperature of the cell

	 //f(current_angle) for this cell, where current_angle covers all angles
	vector<double> f_angle; //the energy of the cell at a certain angle
	vector<double> f_angle_next_t;
	
	double q_vol; //volumetric heat generation term
	
	//------------------transient variables------------//
	double q_vol_on;
	//on time (total time on, the heat source shuts off after this time)
    double on_time;
    //offset (this is the time delay until the heat source starts)
    double offset_time;
	//---------------------end of transient variables---------//

	//store the indices of the Fourier boundaries around this cell
	vector<Bound> boundaries;

	//points to the cell at the boundary
    vector<Fourier_cell*> cell_boundaries;

	friend class BTE_structure;
    friend class Device_structure;
	friend class Fourier_cell;
	friend class Fourier_structure;

	double T_old;

public:

	//class constructor
	Cell();

	//calculate this cells e0 and T
	void calc_cell_e0_t(vector< Propagation > &propvec);

	//get all the surrounding coefficients
	double coeffs() { return (AWest_+AEast_+ASouth_+ANorth_+ABottom_+ATop_); }

	//reset the coefficients to 0
	void reset_coeffs(void);

	//set the guesses and conditions of the cell
	void ini_cell_solve(BTE_structure & system, double temper);

	//get the boundary indices
	Bound get_boundary(int dir);

	//indicates the material properties
	Material * material;

	//set the conditions of the cell if it has a Fourier boundary
	void fbound_cell_solve(int Nmax_angle);

};

/*==========================================================================*/
//Fourier cell
class Fourier_cell{
private:
	//the coefficients of the linear equation to be solved
	//Note there are no West, South, or Bottom coefficients, these are
	//equivalent to the East, North, or Top coefficients of adjacent cells
    double AP_,AEast_,ANorth_,ATop_;
    double B_; //the B term

    double T_; //the temperature of the cell
    double T_next_; //time t+1

    double q_vol; //volumetric heat generation term

	double Qprime; //stores the Q passing through the top boundary

    double Bprime;

	//store the indices of the Fourier boundaries around this cell
    vector<Bound> boundaries;

	//points to the cell at the boundary 
	vector<Cell*> cell_boundaries;

    //a flag which indicates this cell is in a BTE region or not
    int locked;

	friend class BTE_structure;
	friend class Fourier_structure;
    friend class Device_structure;
	friend class Cell;

public:

    //class constructor
    Fourier_cell();

    //get only the surrounding coefficients in North, Top, East
    double coeffs() { return (AEast_+ANorth_+ATop_); }

    //reset the coefficients to 0
    void reset_coeffs(void);

    //sets the initial temperatures for the cells for Fourier cells
    void ini_cell_solve_fourier(double temper);

	//get the boundary indices
	Bound get_boundary(int dir);

	//indicates the material properties
    Material * material;

	//------------------transient variables------------//
	double q_vol_on;
	//on time (total time on, the heat source shuts off after this time)
    double on_time;
    //offset (this is the time delay until the heat source starts)
    double offset_time;
	//---------------------end of transient variables---------//
	
};

/*==========================================================================*/
//For angular partitioning (either phi or theta)
class Angle{
private:
	double centroid_; //the centroid of the angular cell

public:

};

/*==========================================================================*/
//Contains propagation vector and extent for the current direction
class Propagation{
private:

	friend class BTE_structure;
	friend class Fourier_structure;

	
	double extent_; //DeltaOmega, the extent of the current angle
	double S_i_; //S vector component in the 'i' direction
	double S_j_; //S vector component in the 'j' direction
	double S_k_; //S vector component in the 'k' direction

public:
	//set the extent and S vector components for this direction
	void set_components(double &cen_phi, double &cen_theta, double &del_phi,
		double &del_theta);
	
	double extent() {return extent_;}
	double S_i() {return S_i_;}
	double S_j() {return S_j_;}
	double S_k() {return S_k_;}


};

/*==========================================================================*/
//information for source terms
class Source{
private:

public:
	Source();
	
	//source boundary limits, l1b<l2b
	double l1x,l2x,l1y,l2y,l1z,l2z;

	//heat generation term (W)
	double Q_source;
	//volumetric heat generation term (W/m^3)
	double q_vol;
	
	//on time (total time on, it shuts off after this time)
	double on_time;
	
	//offset (this is the time delay until the power starts)
	double offset_time;

	//get volumetric heat generation value
	void get_volumetric_val();

};

/*==========================================================================*/
//stores boundary cells
class Bound{
private:
public:
	Bound();

	//direction
	int dir; //E=0,W=1,N=2,S=3,T=4,B=5
    //indices
    int x,y,z;

	//used for setting the values
	void set_values(int dir_,int i,int j,int k);
};

/*==========================================================================*/
//Defines the device structure, including all materials and BTE and Fourier 
//sections
class Device_structure{
private:

    friend class Fourier_structure;

    //contains multiple materials
    vector<BTE_structure> bte_system;
    //contains multiple devices and their pieces
    vector<vector<BTE_structure> > device;
    //iteration parameter. ={# of fourier iterations and # of BTE iterations}
    int iter_param;

    //variables which are going to be used to initialize all BTE_structures
    //amount of control angles per octant
    vector<int> N_phi;
    vector<int> N_theta;

    //if boundary with silicon dioxide oxide=1 else =0
    vector<int> oxide_x_0;
    vector<int> oxide_x_end;
    vector<int> oxide_y_0;
    vector<int> oxide_y_end;
    vector<int> oxide_z_0;
    vector<int> oxide_z_end;

    //transmittivity factor for each BTE structure
    vector<vector<double> > trans_factor;
    
    //heat sources
    vector<Source> sources;

	//integer for the number of times times the mean free path
	int bte_region_expand_factor;

    //total input heat
    double qin;
	
public:
    Device_structure();
    
    //Generate the LUT containing the BTE thermal resistance and the Fourier
    //thermal resistance
    void generateLUT(string filename);

    //Find the peak temperature in the device
    double get_peak_temp();

	//check to make sure the reflecting boudaries don't hit walls
    void check_reflecting_boundaries();

    //contains information about each BTE structure
    vector<Block_coords> bte_coords;
    //parse the input file
    int input(string filename);

    //substrate structure
    Fourier_structure substrate;
    //Need input structure to determine multiple material areas
    //Need to initially partition the system into Fourier and BTE cells 

    //print output to file
    void output(string argv);
    //initialize the values for the device structure
    void initialization();
    //define the BTE regions and create a BTE_structre for each
    void set_solvers();
    //set all system to be solved using BTE
    void set_all_bte();
    //set the rest of the BTE_structure variables that are not
    //initialized in set_solvers() (BTE_structure-independent variables)
    void set_BTE_structures();
    //set neighbours for all structures
    void set_neigh();
    //solve the device structure by solving the BTE and Fourier sections of the
    //individual systems
    void solve();
    //get the index
    int get_index(double coor, char direction, int max);
    //get coordinate information for all materials in device structure
    void get_material_coords();
    //material specific values
    vector<Material> materials;
    //material with conductance set to 0
    Material null_material;

    //get the total input heat in the whole system
    void get_total_input_heat();

	//Function to indicate which BTE boundaries are touching the Fourier
	//boundaries
	void get_adjacent_boundaries();

	//Function to set the material of each cell in BTE regions
	void set_material();

    //lock the fourier cells that are in the BTE region
    void lock_bte_cells();

	//for updating the BTE boundaries after Fourier block has been solved
	void update_bte_t_bound();
    //update the BTE boundaries after Fourier block has been solved for
    //a semireflecting boundary
    void update_bte_semireflect();
	//for updating the BTE boundaries associated with a certain Fourier block
	void update_fourier_t_bound();

	//set the fourier cell temperatures which are in BTE regions
	void update_fourier_t_cells();

	//set boundaries of BTE region to const temperature boundaries if they
	//are touching Fourier regions
	void set_const_bound();

	//get information that is specific to devices
	void get_device_info(ifstream & fin);

	//initialize the BTE cells temperature after Fourier block has been solved
	void initialize_bte_t_cells();

	//----------------bte transient analysis functions--------------//
	//resize all of the storage for the future
	void resize_trans_f_angle();
    //For ICCAD09 paper: BTE Solver validation
	//Output flux_y at t=1.5 ps and t=10ps
    void output_yflux(ofstream & fout_tran);
	//get the peak temperature of the structure and print it, along with time
	void output_peak_trans(ofstream & fout_tran, double time);
	void output_peak_trans_fourier(ofstream & fout_tran, double time);
	//update all the power sources depending on the time
	void set_Q(double time);
	void set_Q_fourier(double time);
	//-------------end of bte transient analysis functions--------------//

    //void timer_set(timeval *t);
    //void timer_report(timeval *t, int f_or_b);

};

#endif



