/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** dyrk4.cc: code for the time-domain, 4th-order Runge Kutta dynamic thermal
** analysis method
****************************************************************************/

#include "solver.h"
#include "util.h"
#include <vector>
#include <cmath>
#include <sys/time.h>
#include "dysol.h" 

using namespace std;

//*==========================================================================*/
//* Constructor of dysol class. */
dysol::dysol():
    alpha(),
    cur_slope(),
    cur_intval(),
    cur_tslot(),
    tao(),
    cap_(),
    cur_1_slope(),
    cur_2_slope()
{ }


//*===========================================================================*/
//* Using global interval to update the temperature,
//* each time, based on the current temperature, obtainning the 
//* ks for all the grids, and then update. */
void dysol::rk4(RC_network_type & AGC, vector<double> & temp, 
	vector<double> & power, double interval){
 	
	int i,j;
	vector<double> k1, k2, k3, k4, int_re;
	double t_val;
	int index_;

	int size = AGC.size_;
	k1.resize(size);
	k2.resize(size);
	k3.resize(size);
	k4.resize(size);
	int_re.resize(size);
	double resi_ = 0;

    //* Compute K1 array and restore the common summations in int_re arrray, 
	//* to be used later. */
	for (i = 0; i < size; i++){
		t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[i].size());

		//* nei_cnt is the number of neighbors, plus 1
                //* neighbor 0 is self
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * temp[index_];
			}
		}
		resi_ += fabs(power[i] - t_val)/6;
		int_re[i] = t_val;
		k1[i] = (power[i] - t_val)/cap_[i];
	}

	//* Compute K2 array. */
	for (i = 0; i < size; i++) {
		t_val = int_re[i];
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] *  
					interval * 0.5 * k1[index_];
			}
		}
		resi_ += fabs(power[i] - t_val)/3;
		k2[i] = (power[i] - t_val)/cap_[i];
	}

	//* Compute K3 array. */
	for (i = 0; i < size; i++) {
		t_val = int_re[i];
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * ( 
					interval * 0.5 * k2[index_]);
			}
		}
		resi_ += fabs(power[i] - t_val)/6;
		k3[i] = (power[i] - t_val)/cap_[i];
	}

	//* Compute K4 array. */
	for (i = 0; i < size; i++) {
		t_val = int_re[i];
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * ( 
					interval * k3[index_]);
			}
		}
		resi_ += fabs(power[i] - t_val)/6;
		k4[i] = (power[i] - t_val)/cap_[i];
	}

	for (i =0; i < size; i++){
		temp[i] += interval * (k1[i] + 2 * k2[i] + 2 * k3[i]
			+ k4[i])/6.0;
	}
}


/*===========================================================================*/
//* Calculate the temperature based on the current AGC after one step size 
//* without updating the AGC, used to adjust the step size. */ 
void dysol::forward_step(RC_network_type & AGC, vector<double> & temp, 
	vector<double> & power, double & interval,vector<double> & c_temp){

	int i,j;
	vector<double> k1, k2, k3, k4;
	double t_val;
	int index_;

	int size = AGC.size_;
	k1.resize(size);
	k2.resize(size);
	k3.resize(size);
	k4.resize(size);

	for (i = 0; i < size; i++) {
		t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * c_temp[index_];
			}
		}
		k1[i] = (power[i] - t_val)/cap_[i];
	}

	for (i = 0; i < size; i++) {
		t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * (c_temp[index_]
					+ interval * 0.5 * k1[index_]);
			}
		}
		k2[i] = (power[i] - t_val)/cap_[i];
	}

	for (i = 0; i < size; i++) {
		t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * (c_temp[index_]
					+ interval * 0.5 * k2[index_]);
			}
		}
		k3[i] = (power[i] - t_val)/cap_[i];
	}

	for (i = 0; i < size; i++) {
		t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[i].size());
		for ( j = 0; j < nei_cnt; j++){
			index_ = AGC.A_[i][j];
			if (index_ >= 0){
				t_val += AGC.G_[i][j] * (c_temp[index_]
					+ interval * k3[index_]);
			}
		}
		k4[i] = (power[i] - t_val)/cap_[i];
	}

	for (i =0; i < size; i++){
		c_temp[i] += interval * (k1[i] + 2 * k2[i] + 2 * k3[i]
			+ k4[i])/6.0;
	} 
}

//*===========================================================================*/
//* Determine the approximate step size each time before calling
//* the rk4 function size. */
void dysol::adj_ssize(RC_network_type & AGC, vector<double> & temp,
	vector<double> & power, double & err_thres, double & in_inval, 
	double & out_inval){

	int size = AGC.size_;
	vector <double> c_temp;
	vector <double> h_temp;
	c_temp.resize(size);
	h_temp.resize(size);
	for (int i = 0; i < size; i++){
		c_temp[i] = temp[i];
		h_temp[i] = temp[i];
	}
	double inval_1 = 0.75 * in_inval;
	double inval_2 = 1.5 * in_inval;

    //* Try the in_val twice to obtain y(t + 2*in_inval). */
	forward_step(AGC, temp, power, inval_1, c_temp);
	forward_step(AGC, temp, power, inval_1, c_temp);

    //* Try the 2 * in_val once to obtain y(t + 2*in_inval). */
	forward_step(AGC, temp, power, inval_2, h_temp);
	double ratio;

    //* Compare the results obtained by two different approaches. */
	double m_err = 0;
	double min_err = FMAX_;
	for (int i = 0; i < size; i++){
	    if (fabs(c_temp[i] - h_temp[i]) > m_err){
		m_err = fabs(c_temp[i] - h_temp[i]);
	    }
	    if (fabs(c_temp[i] - h_temp[i]) > 1e-20){
		min_err = min(min_err, fabs(c_temp[i]-h_temp[i]));
	    }
	}
	ratio = err_thres/m_err;
	out_inval = pow(ratio, 0.25) * in_inval;
//#warning delete here later
    //out_inval = in_inval;
}


//*===========================================================================*/
//* Run-time solver using 4th order Runge-Kutta algorithm. */
void dysol::rk4_glb_solver(RC_network_type & AGC, vector<double> & temp, 
	vector<double> & power, double interval, double err_thres, double & t_end,
	double & final_intval){

	bool stop = false;
	int size = AGC.size_;
    
	vector<double> new_temp;
	new_temp.resize(size);
	
	int count = 0;
	double simu_time = 0;

	double in_inval = interval;
	double out_inval = interval;

	while(1) {
	    adj_ssize(AGC, temp, power, err_thres,in_inval, out_inval);
	    //out_inval = 1e-6;
	    in_inval = min(t_end - simu_time, out_inval);

	    simu_time += in_inval;
	    double resi_ = 0;

	    rk4(AGC, temp, power, in_inval);
	    for (int ii = 0; ii < size; ii++){
		double t_val = 0.0;
		int nei_cnt = static_cast<int>(AGC.A_[ii].size());
		for ( int j = 0; j < nei_cnt; j++){
		    int index_ = AGC.A_[ii][j];
		    if (index_ >= 0){
			t_val += AGC.G_[ii][j] * temp[index_];
		    }
		}
		resi_ += fabs(power[ii] - t_val);
	    }

	    count ++;
	    if (simu_time >= t_end){
		stop = true;}
	    if (stop){break;}

	}
	final_intval = in_inval;
}


//*===========================================================================*/
//* Output the peak temperature for a vector of double. */
void display_peakt(vector<double> temp){

	double max_t = 0;
	double min_t = FMAX_;

    for (int i = 0; i < int(temp.size()); i++){
		max_t = max(max_t, temp[i]);
		min_t = min(min_t, temp[i]);
		//cout << temp[i] << endl;
	}

    cout << "MAX/MIN temp: " << max_t << " " << min_t << endl;
}
