/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** dytrd.cc: code for the semi-implicit time-domain method
****************************************************************************/

#include <cmath>
#include <sys/time.h>
#include <algorithm>
#include "solver.h"
#include "util.h"
#include "dysol.h"

using namespace std;

/*###########################################################################*/
TIQueue::TIQueue():
     impl_()
{ }

//*===========================================================================*/
//* Semi_th method, which update temperature of grids one by one:
//* update the tepmerature and the slope of block[index] using 
//* trapezoidal method

void dysol::semi_th(const RC_network_type & AGC,
	std::vector<double>& temp, 
        std::vector<vector<double> >& old_temp,
        std::vector<double> & power, int index, double interval) {

    /* calculate the slope of block[index] at time cur_tslot + interval
     * based on equation (17) 
     * attention:  first element of A_ vector in AGC stores the thermal
     * conductivity of block[index] itself, A_[1] ~ A_[6] store the 
     * thermal conductivity of its neighbours 
     * attention : temp is the temperture vector that records current
     * temperature information for all the thermal elements, while old_temp[0]
     * through old_temp[2] record history temperature value with old_temp[3]
     * being the oldest. */

	double beta = power[index];
	int nei_cnt = static_cast<int>(AGC.A_[index].size());
	for (int i = 1; i < nei_cnt; ++i){
		int id_ = AGC.A_[index][i];
		double ther_con = - AGC.G_[index][i];
		if ( id_ >= 0){
			beta += (temp[id_] + (cur_tslot[index] + interval
				  - cur_tslot[id_]) * cur_slope[id_]) * ther_con;
		}
	}

	// calculate the new temperature slope 
	beta -= alpha[index] * (temp[index] + cur_slope[index] * interval / 2.0);
	double new_slope = beta / (cap_[index] + alpha[index] * interval / 2.0);

	// record the previous temperature information for 
	// step size computation using DD3 method
	for (int i = 2; i > 0; i--)
	    old_temp[i][index] = old_temp[i-1][index];
	old_temp[0][index] = temp[index];

	//* calculate new temperature of block[index], and store it in temp 
	//* variable. */
	temp[index] += interval * (cur_slope[index] + new_slope) / 2.0;

	//* update the slope at target time. */
	cur_slope[index] = new_slope;

	//* quiescent state detection:
	//* we assume the element reaches steady state if (1) the slope is 
	//* smaller than dysol::DERIV_THRES or (2) slope is smaller than
	//* 5 * dysol::DERIV_THRES and temperature difference is smaller 
	//* than dysol::TEMP_THRES.*/

	if (fabs(cur_slope[index]) < dysol::DERIV_THRES || ((fabs(cur_slope[index]) <
	5 * dysol::DERIV_THRES) && \
	(fabs(temp[index] - old_temp[0][index]) < dysol::TEMP_THRES))) {
		if (!quies_flag[index])
			quies_sum++;
		quies_flag[index] = true;
	} else {
		if (quies_flag[index])
			quies_sum--;
		quies_flag[index] = false;
	}
}

//*============================================================================
//* Calculate new step based on third-order divided temperature difference, which
//* is then used to estimate the error caused by trapezoidal method during temperature 
//* update 
double dysol::adjsize_semi_th_implct(const RC_network_type & AGC,
	std::vector<double>& temp, 
        std::vector<std::vector<double> >& old_temp,
        double err_thres, int index, 
        double c_tslot, double in_inval, double /* t_end */){

    double first_deriv[3], second_deriv[2], third_deriv;
    double lte, numerator, denominator, out_inval;
    double abs_err = err_thres, rel_err = 0.1 * err_thres;

    //* calculate first-order temperature derivative
    first_deriv[0] = (temp[index] - old_temp[0][index]) / cur_intval[0][index];
    for (int i = 1; i < 3; i++)
	first_deriv[i] = (old_temp[i-1][index] - old_temp[i][index]) / cur_intval[i][index];

    //* calculate second-order temperature derivative
    for (int i = 0; i < 2; i++)
	second_deriv[i] = (first_deriv[i] - first_deriv[i+1]) / (cur_intval[i][index] + cur_intval[i+1][index]);

    //* calculate third-order temperature derivative
    third_deriv = (second_deriv[0] - second_deriv[1]) / (cur_intval[0][index] + cur_intval[1][index] + cur_intval[2][index]);

    //* estimate integration error due to trapezoidal method
    lte = fabs(third_deriv) * pow(in_inval, 3) / 24.0;

    //* estimate new step size, where rel_err can be treated as maximum 
    //* tolerable relative temperature error and abs_err is maximum 
    //* tolerable absolute temperature error. An abs_err of 1e-6 is 
    //* recommended.
    numerator = 2 * (rel_err * fabs(temp[index]) + abs_err);

    //* adjust step size based on integration error
    denominator = max(abs_err, lte);

    //* calculate new step size
    out_inval = 1.5 * pow(numerator / denominator, 0.3) * in_inval;

    //* bound the step size depending on the neighbors' step sizes
    //* to address the problem of global propagation. More specifically, 
    //* current step size cannot exceed 6 * any_neighbor_step_size
    double nei_bound = dysol::FMAX_;
    int nei_cnt = static_cast<int>(AGC.A_[index].size());

    for (int i = 1; i < nei_cnt; ++i) {
	int id_ = AGC.A_[index][i];
	if (id_ >= 0) {
	    double n_bound = 2 * cur_intval[0][id_];
	    if (n_bound > 0){
		nei_bound = 3 * min(nei_bound, n_bound);
	    }
	}
    }
    
    out_inval = min(out_inval, nei_bound);

    return out_inval;
}

//*===========================================================================*/
//* Run-time solver using forward Euler order method, temperatures are 
//* updated block by block. */
void dysol::semi_th_implct_solver(const RC_network_type & AGC, 
		std::vector<double>& temp, 
                std::vector<std::vector<double> >& old_temp, 
		std::vector<double> & power, double interval, double err_thres, 
                double t_end){
  
	//* Create the AGC structure 
        //* An AGC structure is an vector of a structure hoct_interface, while
        //* there is a A_ vector, power_ and temp_ elements in hoct_interface. The
        //* structure stores the thermal conductivies and power, temp information
        //* of one specific block.
        //* A_ vector has seven elements, the first one is the thermal conductivity
        //* of the block itself, the 2-7 elements are the thermal conductivities between
        //* this block with and its neighbours respectively, power_ and temp_ are power 
        //* consumption value and temperature value of this block. */

	int size = AGC.A_.size();

	//* Initialize quiesence sum, which is used to record how many elements have 
	//* reached their steady states. 

	quies_sum = 0;

	//* Resize vectors for current interval, the alpha vector, slope vector, 
	//* capacitance vector, quies_flag, etc. */
	cur_tslot.resize(size);
	cur_slope.resize(size);
	alpha.resize(size);
	quies_flag.resize(size);
	tao.resize(size);
	cur_intval.resize(3);
	old_temp.resize(3);
	for (int i = 0; i < 3; i++) {
	    cur_intval[i].resize(size);
	    old_temp[i].resize(size);
	}
	
	//* Initialize vectors for current time, the alpha vector, quiescence
	//* flag vector, temperature vector, etc.
	for (int i = 0; i < size; ++i){
		cur_tslot[i] = 0;
		alpha[i] = 0;
		quies_flag[i] = false;
		for (int j = 2; j > 0; j--)
			old_temp[j][i] = old_temp[j-1][i];
		old_temp[0][i] = temp[i];

	    int nei_cnt = static_cast<int>(AGC.A_[i].size());
	    for (int j = 1; j < nei_cnt; j++) {
		alpha[i] += - AGC.G_[i][j];
	    }
	    tao[i] = (1.0/AGC.G_[i][0]) * cap_[i]; 
	}

	//* Calculate the initial slope for each thermal element
	init_slope(AGC, temp, power);

	double out_inval;

	//* Adpative h here, each block has a unique block size. */
	vector<int> it_cnt;
	it_cnt.resize(size);
	for (int i = 0; i < size; ++i){
	    it_cnt[i] = 0;
	    for (int j = 0; j < 3; j++)
		cur_intval[j][i] = interval;
	}

	//* Initialize the heap with current time 0s, current interval, 
	//* the current intervals are calculated based on the same 
	//* initial interval.  The heap is sorted by the target time of the 
	//* blocks. */
	double start_t = 0.0;

	TIQueue run_q;
	for (int index = 0; index < size; ++index){
		out_inval = adjsize_semi_th_implct(AGC, temp, old_temp, err_thres, index,
			start_t , interval, t_end);  
		for(int j = 2; j > 0; j--)
		    cur_intval[j][index] = cur_intval[j-1][index];
		cur_intval[0][index] = out_inval;

		TimeIndex ti(start_t + out_inval, index);
		run_q.push(ti);
	}

	//* Start from the top element of the queue, update its temperature
	//* and slope, then update its cur_tslot. Then adjust the next step
	//* size based on the existing information. And update the cur_intval
	//* with the determined step size. pop the top element out and push
	//* in an element with updated target time(old target time + new step
	//* size. */

	/* the following lines till "control[i] = r_num" seem unuseful */
	int r_num = 1;
	vector<int> control;
	control.resize(size);
	for (int i = 0; i < size; i++){
	    control[i] = r_num;
	} 

	// determine when it should ends
	int endsum = 0;
	
	while (true) {
	        //* Start updating the block at the top of the queue. */
		TimeIndex cur_ti = run_q.smallest();
		++it_cnt[cur_ti.index];
		
		//* when an element reaches the simulation end, increment
		//* endsum. When all elements reach the simulation end, i.e.,
		//* endsum == size, return.
		if (cur_ti.time >= t_end){
		    for (int i = 2; i > 0; i--)
			cur_intval[i][cur_ti.index] = cur_intval[i-1][cur_ti.index];
		    cur_intval[0][cur_ti.index] -= (cur_ti.time - t_end); 
		    cur_ti.time = t_end;
		    endsum += 1;
		    if (endsum == size)
			return;
		}

		//* update temperature for element associated with cur_ti.index
		semi_th(AGC, temp, old_temp, power, cur_ti.index, cur_intval[0][cur_ti.index]);

		//* update element's current time
		cur_tslot[cur_ti.index] += cur_intval[0][cur_ti.index];

		//* deal with quiescent state
		if (quies_sum == size) {
		    for (int index = 0; index < size; index++) {
			cur_tslot[index] = t_end;
		    }
		    break;
		}
		
		--control[cur_ti.index];
		out_inval = cur_intval[0][cur_ti.index];
		if (0 == control[cur_ti.index]){
		    //* adjust element's step size
		    out_inval = adjsize_semi_th_implct(AGC, temp, old_temp, err_thres, cur_ti.index, 
						       cur_ti.time, cur_intval[0][cur_ti.index], t_end);
		    control[cur_ti.index] = r_num;
		}  

		// update element's step size
		for (int i = 2; i > 0; i--)
		    cur_intval[i][cur_ti.index] = cur_intval[i-1][cur_ti.index];
		cur_intval[0][cur_ti.index] = out_inval;

		// pop the event from the event queue
		run_q.pop();

		//* Needs to update the new target time for the current element to
		//* a large value to avoid keeping updating the same element */
		if (cur_ti.time >= t_end){
		    TimeIndex new_ti(dysol::FMAX_, cur_ti.index);
		    run_q.push(new_ti);
		}
		else {
		    TimeIndex new_ti(cur_ti.time + cur_intval[0][cur_ti.index], 
				     cur_ti.index);
		    run_q.push(new_ti);
		}
	}
}


//*===========================================================================*/
//* Obtain the initial slope of all blocks based on the initial temperature
//* use equation 17.
//* All temperatures are current temperatures (temperature at time 0, slope
//* at time 0. */
void dysol::init_slope(const RC_network_type & AGC, 
	std::vector<double>& temp, std::vector<double> & power){

        //int size = AGC.size_;
        int size = AGC.A_.size();

	for (int index = 0; index < size; ++index){
		double beta =  power[index];
		int nei_cnt = static_cast<int>(AGC.A_[index].size());
		for (int i = 0; i < nei_cnt; ++i){ 
			int id_ = AGC.A_[index][i];
			double ther_con = - AGC.G_[index][i];
			if ( id_ >= 0){
				beta += ther_con * temp[id_];
			}
		}
		cur_slope[index] = beta / cap_[index];
	}
}



