// Copyright 2018 by Robert Dick.
// All rights reserved.

#ifndef MEM_CHECKER_
#define MEM_CHECKER_

/*###########################################################################*/
/* Tracks currently allocated memory and watches for some out-of-bounds writes.
.cc file redefines std's new, delete, new[], and delete[].  Sets memory to a
nasty value on allocation and another nasty value on deallocation.  Records
address of allocation to allow a user to jump to the line in code that caused
leaks.  Make a MemAlloc object to start tracing in its scope.

Normalized CPU times:
0.71: Normal new
0.72: wall_new, wall_delete, no tracing or blanking
0.94: tracing and blanking while recording no ignored addresses
1.0 : tracing and blanking while recording ignored addresses
0.83: hash_map tracing and blanking while recording ignored addresses */

/* MC_PORTABLE causes the standard STL map data structure to be used.  MC_FAST
causes the non-standard ext/hash_map data structure but is faster. */
// Define MC_PORTABLE or MC_FAST here.
//#define MC_PORTABLE
#define MC_FAST

/* Need a non-new based allocator.  Standard libraries should provide
malloc_allocator<>.  However, its location may not be standard. */
#include <ext/malloc_allocator.h>

#include <functional>

#if defined MC_PORTABLE
#	include <map>
#elif defined MC_FAST
#	include <unordered_map>
#endif

#include <stack>
#include <deque>
#include <string>
#include <cstddef>
#include <memory>
#include <functional>

namespace rstd {
/*===========================================================================*/
// Put one of these in the outer scope of main to monitor memory management.
// Can make more for local monitoring.
class MemChecker {
public:
	struct AllocRecord {
		AllocRecord(size_t b, const void * c) : bytes(b), caller(c) {}
		size_t bytes;
		const void * caller;
	};

// Location byte -> requested bytes, caller address?

#if defined MC_FAST
	typedef std::unordered_map<std::size_t, AllocRecord,
		std::hash<std::size_t>, std::equal_to<std::size_t>,
		__gnu_cxx::malloc_allocator<void *> > MemMap;
#elif defined MC_PORTABLE
	typedef std::map<std::size_t, AllocRecord, std::less<std::size_t>,
		__gnu_cxx::malloc_alloc<void *> > MemMap;
#endif

// Checker for current stack frame.
	static MemChecker * mc();

// Put this on stack.  Don't new it.
	MemChecker(const char * name);

// Set maximum number of leaks to report.
	void max_print(size_t mprint);

// Set range or number of leaked objects to expect.
	void expect_alloc_cnt(size_t min_alloc_cnt, size_t max_alloc_cnt);
	void expect_alloc_cnt(size_t alloc_cnt);

// Set range or number of leaked bytes to expect.
	void expect_byte_cnt(size_t min_byte_cnt, size_t max_byte_cnt);
	void expect_byte_cnt(size_t byte_cnt);
 
/* Used to ignore leaks generated by a range of addresses owned by system
libraries. */
	void expect_caller(size_t ignore_loc_min, size_t ignore_loc_max);
	void expect_caller(size_t ignore_loc);

	bool is_ignore(const void * addr) const;

	~MemChecker();

	void * checked_alloc(size_t bytes);
	void checked_dealloc(void * data);

	size_t out_allocs() const;
	size_t out_bytes() const;

// Checks for leaks, generating at most mprint specific leak report.
	void check(size_t mprint);

private:
	typedef std::stack<MemChecker *, std::deque<MemChecker *,
		__gnu_cxx::malloc_allocator<MemChecker *> > > MemCheckStack;

// Location to size.
		MemMap mem_map_;
		size_t min_alloc_;
		size_t max_alloc_;
		size_t min_byte_;
		size_t max_byte_;
		void * ignore_min_;
		void * ignore_max_;
		size_t max_print_;
		std::string name_;
		static MemCheckStack mc_stack_;
};

/*===========================================================================*/
void MemChecker_test();
}

/*###########################################################################*/
#endif
