/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** fourier.cc: solve the fourier equations using line by line TDMA
****************************************************************************/

#include "header.h"

//*==========================================================================*/
//* this function is only called once per iteration and solves AT=P, for T
double Fourier_structure::lblTdma_fourier()
{

    int i,j,k;	//counters for 3 directions
	
    int l;

    vector<double> coeffVector1,coeffVector2,coeffVector3,bVector,energyVector;
    vector<int> lockVector;
	
    double rel_change; //the relative change of energy after each iteration
    double max_rel_change=0; //the maximum relative change of energy of all cells
	
    double energy_new; //the value of energy after iteration
    double energy_old; //the value of energy before iteration
	
    //reserve memory space for the vectors
    coeffVector1.resize(f_cells_z-2);	//boundaries are not taken
    coeffVector2.resize(f_cells_z-2);
    coeffVector3.resize(f_cells_z-2);

    bVector.resize(f_cells_z-2);

    lockVector.resize(f_cells_z-2);
	
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(f_cells_z-2);

    //loop for all cartesian directions

    //Sweeping the directions (6 loops)
    /* ------------------------  First Loop  ------------------------ */
    
    for(i=1;i<f_cells_x-1;i++)
    //loop for all cells in x-direction (forward)
    {
        for(j=1;j<f_cells_y-1;j++)
        //loop for all cells in y-direction (forward)
        {
            for(k=1;k<f_cells_z-1;k++)
            //loop for all cells in z-direction
            {
                l=k-1;
                coeffVector1[l]=f_cell[i][j][k].AP_; //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].ATop_; //AT1D(k) = AT(i,j,k);
                coeffVector3[l]=-f_cell[i][j][k-1].ATop_; //AB1D(k) = AB(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_+
                        f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_+
                        f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_+
                        f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_+
                        f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
                        f_cell[i][j][k].Bprime;

                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;
                //B1D(k) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                  + AE(i,j,k)∗X(i+1,j,k)
                //                  + AS(i,j,k)∗X(i,j-1,k)
                //                  + AN(i,j,k)∗X(i,j+1,k);
                    
                //add effects of two cells at the boundary to B term
                if(k==1) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][0].ATop_*f_cell[i][j][0].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_;
                }
                if(k==(f_cells_z-2)) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_;
                }
            }
            // Call TDMA(AP1D,AT1D,AB1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            // Write elements back in structure
            for(k=1;k<f_cells_z-1;k++)
            {
                //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //X(i,j,k) = X1D(k);
                    f_cell[i][j][k].T_=energyVector[k-1];
                }
            }	
        }
    }
        
    //	/ ------------------------  Second Loop  ------------------------ /
    
    
    for(i=f_cells_x-2;i>=1;i--)
    //loop for all cells in x-direction (backward)
    {
        for(j=f_cells_y-2;j>=1;j--)
        //loop for all cells in y-direction (backward)
        {
            for(k=1;k<f_cells_z-1;k++)
            //loop for all cells in z-direction
            {
                l=k-1;
                coeffVector1[l]=f_cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].ATop_;  //AT1D(k) = AT(i,j,k);
                coeffVector3[l]=-f_cell[i][j][k-1].ATop_;  //AB1D(k) = AB(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_
                        +f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_
                        +f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_
                        +f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_
                        +f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
                         f_cell[i][j][k].Bprime;

                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;

                //B1D(k) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                  + AE(i,j,k)∗X(i+1,j,k)
                //                  + AS(i,j,k)∗X(i,j-1,k)
                //                  + AN(i,j,k)∗X(i,j+1,k);
                
                //add effects of two cells at the boundary to B term
                if(k==1) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][0].ATop_*f_cell[i][j][0].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_;
}               

                if(k==(f_cells_z-2)) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_;
                }
            }
            // Call TDMA(AP1D,AT1D,AB1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            // Write elements back in structure
            for(k=1;k<f_cells_z-1;k++)
            {
                //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //X(i,j,k) = X1D(k);
                    f_cell[i][j][k].T_=energyVector[k-1];	

                }
            }	
        }
    }	
    
    //reserve memory space for the vectors
    coeffVector1.resize(f_cells_y-2);
    coeffVector2.resize(f_cells_y-2);
    coeffVector3.resize(f_cells_y-2);
    bVector.resize(f_cells_y-2);

    lockVector.resize(f_cells_y-2);
            
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(f_cells_y-2);

    // ------------------------  Third Loop  ------------------------ 
    
    for(k=1;k<f_cells_z-1;k++)
    //loop for all cells in z-direction (forward)
    {
        for(i=1;i<f_cells_x-1;i++)
        //loop for all cells in x-direction (forward)
        {
            for(j=1;j<f_cells_y-1;j++)
            //loop for all cells in y-direction
            {
                l=j-1;
                coeffVector1[l]=f_cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].ANorth_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-f_cell[i][j-1][k].ANorth_;  //AS1D(j) = AS(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_+
                        f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_+
                        f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_+
                        f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_+
                        f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
                        f_cell[i][j][k].Bprime;
 

                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;

                //   B1D(j) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                     + AE(i,j,k)∗X(i+1,j,k)
                //                     + AB(i,j,k)∗X(i,j,k-1)
                //                     + AT(i,j,k)∗X(i,j,k+1);
                
                //add effects of two cells at the boundary to B term
                if(j==1) {
                    bVector[l] = bVector[l]
                        +f_cell[i][0][k].ANorth_*f_cell[i][0][k].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_;
                }

                if(j==(f_cells_y-2)) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_;
                }	
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            // Write elements back in structure
            for(j=1;j<f_cells_y-1;j++)
            {
                //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //X(i,j,k) = X1D(k);
                    f_cell[i][j][k].T_=energyVector[j-1];
                }
            }	
        }
    }
    
    // ------------------------  Fourth Loop  ------------------------ 
    
    for(k=f_cells_z-2;k>=1;k--)
    //loop for all cells in z-direction (backward)
    {
        for(i=f_cells_x-2;i>=1;i--)
        //loop for all cells in x-direction (backward)
        {
            for(j=1;j<f_cells_y-1;j++)
            //loop for all cells in y-direction
            {
                l=j-1;
                coeffVector1[l]=f_cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].ANorth_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-f_cell[i][j-1][k].ANorth_;  //AS1D(j) = AS(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_+
                        f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_+
                        f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_+
                        f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_+
                        f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
                        f_cell[i][j][k].Bprime;

                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;

                //   B1D(j) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                     + AE(i,j,k)∗X(i+1,j,k)
                //                     + AB(i,j,k)∗X(i,j,k-1)
                //                     + AT(i,j,k)∗X(i,j,k+1);
                    
                //add effects of two cells at the boundary to B term
                if(j==1) {
                    bVector[l] = bVector[l]
                        +f_cell[i][0][k].ANorth_*f_cell[i][0][k].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_;
                }

                if(j==(f_cells_y-2)) {
                    bVector[l] = bVector[l]
                        +f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_;
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            // Write elements back in structure
            for(j=1;j<f_cells_y-1;j++)
            {
                //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //X(i,j,k) = X1D(k);
                    f_cell[i][j][k].T_=energyVector[j-1];
                }
            }
        }
    }
    
    //reserve memory space for the vectors
    coeffVector1.resize(f_cells_x-2);
    coeffVector2.resize(f_cells_x-2);
    coeffVector3.resize(f_cells_x-2);
    bVector.resize(f_cells_x-2);

    lockVector.resize(f_cells_x-2);
            
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(f_cells_x-2);
        
    // ------------------------  Fifth Loop  ------------------------ 
    
    for(j=1;j<f_cells_y-1;j++)
    //loop for all cells in y-direction (forward)
    {
        for(k=1;k<f_cells_z-1;k++)
        //loop for all cells in z-direction (forward)
        {
            for(i=1;i<f_cells_x-1;i++)
            //loop for all cells in x-direction
            {
                l=i-1;
                coeffVector1[l]=f_cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].AEast_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-f_cell[i-1][j][k].AEast_;  //AS1D(j) = AS(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_+
                        f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_+
                        f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
                        f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_+
                        f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
                        f_cell[i][j][k].Bprime;


                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;
                //add effects of two cells at the boundary to B term
                if(i==1) {
                    bVector[l] = bVector[l]+
                        f_cell[0][j][k].AEast_*f_cell[0][j][k].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_;
                }

                if(i==(f_cells_x-2)) {
                    bVector[l] = bVector[l]+
                        f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_;
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            for(i=1;i<f_cells_x-1;i++)
            {
                //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //X(i,j,k) = X1D(k);
                    f_cell[i][j][k].T_=energyVector[i-1];
                }
            }	
        }
    }


    // ------------------------  Sixth Loop  ------------------------ 

    for(j=f_cells_y-2;j>=1;j--)
    //loop for all cells in y-direction (backward)
    {
        for(k=f_cells_z-2;k>=1;k--)
        //loop for all cells in z-direction (backward)
        {
            for(i=1;i<f_cells_x-1;i++)
            //loop for all cells in x-direction
            {
                l=i-1;
                coeffVector1[l]=f_cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);

                coeffVector2[l]=-f_cell[i][j][k].AEast_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-f_cell[i-1][j][k].AEast_;  //AS1D(j) = AS(i,j,k);
                lockVector[l]=f_cell[i][j][k].locked;
                energyVector[l]=f_cell[i][j][k].T_;
                bVector[l]=f_cell[i][j][k].B_+
                        f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_+
                        f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
                        f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_+
                        f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
                        f_cell[i][j][k].Bprime;

                bVector[l] = bVector[l] + f_cell[i][j][k].q_vol;
                            
                //add effects of two cells at the boundary to B term
                if(i==1) {
                bVector[l] = bVector[l]+
                        f_cell[0][j][k].AEast_*f_cell[0][j][k].T_;
                }
                else if(lockVector[l-1]==1) {
                    bVector[l]=bVector[l]+f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_;
                }

                if(i==(f_cells_x-2)) {
                    bVector[l] = bVector[l]+
                        f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_;
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma_f(coeffVector1,coeffVector2,coeffVector3,lockVector,bVector,energyVector);
            for(i=1;i<f_cells_x-1;i++)
            {
            //if this cells is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
                    //the value of energy after iteration
                    energy_new=energyVector[i-1];
                    //the value of energy before iteration
                    energy_old=f_cell[i][j][k].T_;
                    //the relative change
                    rel_change=abs((energy_new-energy_old)/energy_old);
                    //max_rel_change is used to stop the iterations
                    //if its value goes below the required accuracy
                    if(rel_change>max_rel_change)
                    {
                        max_rel_change=rel_change;   		     
                    }
                    // Write elements back in structure
                    f_cell[i][j][k].T_=energy_new;
                }
            }
        }
    }
     
	return max_rel_change;

}

//*==========================================================================*/
void Tdma_f(vector<double> &coeffVector1,vector<double> &coeffVector2,vector<double> &coeffVector3,vector<int> &lockVector,vector<double> &bVector,vector<double> &energyVector)
{
	
    int cell;	//counter for the cell number in the line
    double r;
    int size;	

    size=coeffVector1.size();	//size of the vectors

    for(cell=1;cell<size;cell++)	//start elimination from the second line
    {
        if(!lockVector[cell-1]) {
                // r = AW(i)/AP(i-1);
                r=coeffVector3[cell]/coeffVector1[cell-1];
                // AP(i) = AP(i) - r*AE(i-1);
                coeffVector1[cell]=coeffVector1[cell]-r*coeffVector2[cell-1];
                //B(i) = B(i) - r*B(i-1);
                bVector[cell]=bVector[cell]-r*bVector[cell-1];
        }
    }
    //--back substitution--
    //X(N) = B(N)/AP(N);
    if(!lockVector[size-1])
    {
        energyVector[size-1]=bVector[size-1]/coeffVector1[size-1];
    }
    for(cell=size-2;cell>=0;cell--)
    {
        if(!lockVector[cell])
        {
            //X(i) = (B(i) - AE(i)*X(i+1))/AP(i);
            energyVector[cell]=(bVector[cell]-coeffVector2[cell]*energyVector[cell+1])/coeffVector1[cell];	
        }
    }
}


/*==========================================================================*/
//* solve the system using Fourier model
int Fourier_structure::solve_fourier(int iter_param, Device_structure *
device_structure) {

    f_error=TEMP_ERROR; // /(4*PI);

//    double residual=f_error*10; //the residual of the system
    //initialized with any value greater than ACCURACY in order for the loop to execute the first time
    
    double max_rel_change=f_error*10; //the maximum relative change of energy of all cells;
    //initialized with any value greater than ACCURACY in order for the loop to execute the first time

//    double residual_old=0;

    //start loop until solution converges
    
    int count=0;
    
    device_structure->update_fourier_t_cells();
  
	//quit the loop when it has completed the desired amount of solutions    
	while(count<iter_param) {

		count++;

//        residual_old=residual;        

//    	residual=0;

        //Choose which TDMA function to call
		//execute lbl-tdma
        max_rel_change=lblTdma_fourier();
		  
		//calculate residual
//		residual=calc_residual();
			calc_residual();
        //cout<<"Max rel change="<<max_rel_change<<endl;
        //cout<<"Residual="<<residual<<endl;


    }

	//return a 0 if solution converged
    //cout<<(residual)<<endl;
    //cout<<max_rel_change<<" res "<<residual<<" "<<residual_old<<endl;

	//if(max_rel_change>f_error||((residual>1e-6)&&(residual_old!=residual)))
	if(max_rel_change>f_error)
		return 1;
	else
    {
        //cout<<"Max rel change="<<max_rel_change<<endl;
        //cout<<"Residual="<<residual<<endl;
		return 0;
    }

}

/*==========================================================================*/
//* calculate the sum of the residuals for Fourier
double Fourier_structure::calc_residual()
{
	int i,j,k;	//counters for 3 directions
	
	double residual=0; //residuals are accumulated in this variable
	
	//loop on all blocks and sum residuals
	for(i=1;i<f_cells_x-1;i++)
	{
		for(j=1;j<f_cells_y-1;j++)
		{
			for(k=1;k<f_cells_z-1;k++)
			{
                //if the cell is not in the BTE region
                if(!(f_cell[i][j][k].locked))
                {
				    //res=ApXp-(AwXw+AeXe+AsXs+AnXn+AbXb+AtXt+B)
				    residual=residual+abs(f_cell[i][j][k].AP_*f_cell[i][j][k].T_-
			                 (f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_+
			                  f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_+
			                  f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_+
			                  f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
			                  f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_+
			                  f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
			                  f_cell[i][j][k].B_+f_cell[i][j][k].q_vol));
                }
			}
		}
	}
	return residual;
}

/*==========================================================================*/
//* set the initial guesses and conditions
void Fourier_structure::ini_solve_fourier(void) {

    //set the boundary temperatures to the initial temperature if constant
	    //boundary temperatures were not specified
    double temper;
    if(tbound_x_0==0) {
        tbx_0_val = T_ini;
    }
    if(tbound_x_end==0) {
        tbx_end_val = T_ini;
    }
    if(tbound_y_0==0) {
        tby_0_val = T_ini;
    }
    if(tbound_y_end==0) {
        tby_end_val = T_ini;
    }
    if(tbound_z_0==0) {
        tbz_0_val = T_ini;
    }
    if(tbound_z_end==0) {
        tbz_end_val = T_ini;
    }	

	//go over all cells
	for(int i=0; i<f_cells_x; i++) {
		for(int j=0; j<f_cells_y; j++) {
			for(int k=0; k<f_cells_z; k++) {
                //set all the cells initially as unlocked
                f_cell[i][j][k].locked = 0;
				if(i==0) {
                    temper = tbx_0_val;
                }
                else if(i==(f_cells_x-1)) {
                    temper = tbx_end_val;
                }
                else if(j==0) {
                    temper = tby_0_val;
                }
                else if(j==(f_cells_y-1)) {
                    temper = tby_end_val;
                }
                else if(k==0) {
                    temper = tbz_0_val;
                }
                else if(k==(f_cells_z-1)) {
                    temper = tbz_end_val;
                }
                else
                    temper = T_ini;

				//set the guesses and conditions of the cell
				f_cell[i][j][k].ini_cell_solve_fourier(temper);
			}
		}
	}

}

/*==========================================================================*/
//* obtain the matrix coefficients for Fourier
void Fourier_structure::obtain_coeffs_fourier() {

    double flow; //temporary variable

    //get north and south coefficients
    for(int i=1; i<f_cells_x-1; i++) {
        for(int k=1; k<f_cells_z-1; k++) {

            for(int j=0; j<f_cells_y-1; j++) {
				//currently set up for uniform grid (width is constant)
    	        //g_th = kA/L
	            flow = del_x*del_z*(f_cell[i][j][k].material->k_cond)/del_y;

                f_cell[i][j][k].ANorth_ = flow;
            }
        }
    }

    //get east and west coefficients
    for(int j=1; j<f_cells_y-1; j++) {
        for(int k=1; k<f_cells_z-1; k++) {

            for(int i=0; i<f_cells_x-1; i++) {
				//currently set up for uniform grid (width is constant)
	            //g_th = kA/L
    	        flow = del_y*del_z*(f_cell[i][j][k].material->k_cond)/del_x;

                f_cell[i][j][k].AEast_ = flow;
            }
        }
    }
    //get top and bottom coefficients
    for(int i=1; i<f_cells_x-1; i++) {
        for(int j=1; j<f_cells_y-1; j++) {

            for(int k=0; k<f_cells_z-1; k++) {
				//currently set up for uniform grid (width is constant)
	            //g_thermal = kA/L
    	        flow = del_x*del_y*(f_cell[i][j][k].material->k_cond)/del_z;
                f_cell[i][j][k].ATop_ = flow;
            }
        }
    }

    //set boundary coefficients for insulating boundaries
    f_ins();

    //obtain centroid point coefficient Ap and b term
    for(int i=1; i<f_cells_x-1; i++) {
        for(int j=1; j<f_cells_y-1; j++) {
            for(int k=1; k<f_cells_z-1; k++) {
                //get the centroid coeff.
                //gnn = gnm + gnl + ...
                f_cell[i][j][k].AP_ = f_cell[i][j][k].coeffs() +
                        f_cell[i-1][j][k].AEast_ + f_cell[i][j-1][k].ANorth_ +
                        f_cell[i][j][k-1].ATop_;

                //get the b term (all set to 0 for now)
                f_cell[i][j][k].B_ = 0;

            }
        }
    }
}

/*==========================================================================*/
//* condition the coefficients so the boundaries are set correctly
void Fourier_structure::condition_coeffs_fourier() {

	//g_th = kA/L, at the boundaries (also at boundaries of locked cells), the
	//length is half that of normal, so times the conductance by 2
	//This note is no longer valid: Note: this function assumes that the 
	//entire fourier area is mode of a single material

	//make sure not to disrupt the insulating coefficient values


    //get north and south coefficients
    for(int i=1; i<f_cells_x-1; i++) {
        for(int k=1; k<f_cells_z-1; k++) {

			//Priority 1, set the boundary coefficients to the proper values
            if(ins_y_0==0) {
				if(f_cells_y>1) {
    	            f_cell[i][0][k].ANorth_ = f_cell[i][1][k].ANorth_*2;
        	    }
            	else {
                	f_cell[i][0][k].ANorth_ = f_cell[i][0][k].ANorth_*2;
	            }
			}
            for(int j=1; j<f_cells_y-2; j++) {
				
				//Priority 2
				//if there is a locked cell beside this one to the north
				if(f_cell[i][j+1][k].locked) {
                    f_cell[i][j][k].ANorth_ = f_cell[i][j][k].ANorth_*2;
                }
				//Priority 3
				//if the materials of the two cells are not equal, then combine
				//their values to get the overall right conductance
				else if(f_cell[i][j+1][k].material!=f_cell[i][j][k].material) {
					f_cell[i][j][k].ANorth_ = 1/(1/(f_cell[i][j][k].ANorth_*2)
										+ 1/(f_cell[i][j+1][k].ANorth_*2));
				}
				//Priority 1
				//if this cell is locked, set its coefficient of twice that of
				//the one above it
				if((f_cell[i][j][k].locked)&&(ins_y_end==0)) {
                    f_cell[i][j][k].ANorth_ = f_cell[i][j+1][k].ANorth_*2;
                }


            }
			
			f_cell[i][f_cells_y-2][k].ANorth_=
                                    f_cell[i][f_cells_y-2][k].ANorth_*2;

        }
    }


    //get east and west coefficients
    for(int j=1; j<f_cells_y-1; j++) {
        for(int k=1; k<f_cells_z-1; k++) {

			//Priority 1, set the boundary coefficients to the proper values
			if(ins_x_0==0) {
	            if(f_cells_x>1) {
    	            f_cell[0][j][k].AEast_ = f_cell[1][j][k].AEast_*2;
        	    }
            	else {
                	f_cell[0][j][k].AEast_ = f_cell[0][j][k].AEast_*2;
	            }
			}

            for(int i=1; i<f_cells_x-2; i++) {

				//Priority 2
				//if there is a locked cell beside this one to the east
				if(f_cell[i+1][j][k].locked) {
                    f_cell[i][j][k].AEast_ = f_cell[i][j][k].AEast_*2;
                }

				//Priority 3
                //if the materials of the two cells are not equal, then combine
                //their values to get the overall right conductance
                else if(f_cell[i+1][j][k].material!=f_cell[i][j][k].material) {
 	               f_cell[i][j][k].AEast_ = 1/(1/(f_cell[i][j][k].AEast_*2)
                                  + 1/(f_cell[i+1][j][k].AEast_*2));
                }
				//Priority 1
				//if this cell is locked, set its coefficient of twice that of
				//the one above it
				if((f_cell[i][j][k].locked)&&(ins_x_end==0)) {
					f_cell[i][j][k].AEast_ = f_cell[i+1][j][k].AEast_*2;
				}

            }

			f_cell[f_cells_x-2][j][k].AEast_=
                                f_cell[f_cells_x-2][j][k].AEast_*2;

        }
    }
    //get top and bottom coefficients
    for(int i=1; i<f_cells_x-1; i++) {
        for(int j=1; j<f_cells_y-1; j++) {

			//Priority 1, set the boundary coefficients to the proper values
			if(ins_z_0==0) {
	            if(f_cells_z>1) {
    	            f_cell[i][j][0].ATop_ = f_cell[i][j][1].ATop_*2;
        	    }
            	else {
                	f_cell[i][j][0].ATop_ = f_cell[i][j][0].ATop_*2;
	            }
			}

            for(int k=1; k<f_cells_z-2; k++) {

				//Priority 2
				//if there is a locked cell beside this one to the north
				if(f_cell[i][j][k+1].locked) {
                    f_cell[i][j][k].ATop_ = f_cell[i][j][k].ATop_*2;
                }

				//Priority 3
				//if the materials of the two cells are not equal, then combine
                //their values to get the overall right conductance
				else if(f_cell[i][j][k+1].material!=f_cell[i][j][k].material) {
                	f_cell[i][j][k].ATop_ = 1/(1/(f_cell[i][j][k].ATop_*2)
                                       + 1/(f_cell[i][j][k+1].ATop_*2));
                }
				//Priority 1
                //if this cell is locked, set its coefficient of twice that of
                //the one above it
                if((f_cell[i][j][k].locked)&&(ins_z_end==0)) {
                	f_cell[i][j][k].ATop_ = f_cell[i][j][k+1].ATop_*2;
                }
                
			}
			
			f_cell[i][j][f_cells_z-2].ATop_=
                                    f_cell[i][j][f_cells_z-2].ATop_*2;

        }
    }

    //obtain centroid point coefficient Ap and b term
    for(int i=1; i<f_cells_x-1; i++) {
        for(int j=1; j<f_cells_y-1; j++) {
            for(int k=1; k<f_cells_z-1; k++) {
                //get the centroid coeff.
                //gnn = gnm + gnl + ...
                f_cell[i][j][k].AP_ = f_cell[i][j][k].coeffs() +
                        f_cell[i-1][j][k].AEast_ + f_cell[i][j-1][k].ANorth_ +
                        f_cell[i][j][k-1].ATop_;

            }
        }
    }
}

/*==========================================================================*/
//* set boundary temperatures between BTE and Fourier sections
//* this function is for updating the BTE boundaries after Fourier block has been
//* solved
void Device_structure::update_bte_t_bound() {

    //go through all BTE structures
    int size = bte_system.size();
    for(int h=0; h<size; h++) {
        //update the fourier BTE boundaries associated with the BTE block
        bte_system[h].update_bte_t_bound(substrate,bte_coords[h]);
    }
}

/*==========================================================================*/
//* set boundary temperatures between BTE and Fourier sections
//* this function is for updating the BTE boundaries after Fourier block has been
//* solved
void BTE_structure::update_bte_t_bound(Fourier_structure &substrate, 
												Block_coords & bte_coords) {

	Bound temp; //temporary variable to hold x,y,z indices

	//for this BTE region, check all the boundaries
    for(int h=0; h<6; h++) {

        //if this is a boundary in contact with Fourier cells
        if(bte_boundaries[h]==1) {
            //then set the boundary cells temperatures
            switch (h) {
                case E_:
                    //go through all the cells along the boundary
                    for(int j=1; j<cells_y-1; j++) {
                        for(int k=1; k<cells_z-1; k++) {
							//get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[cells_x-1][j][k].get_boundary(E_);
                            cell[cells_x-1][j][k].T_ =
                            cell[cells_x-1][j][k].T_old +
                            (cell[cells_x-1][j][k].cell_boundaries[h]->T_
                            +cell[cells_x-1][j][k].cell_boundaries[h]->Qprime
                            /(substrate.f_cell[bte_coords.x+cells_x-3]
                            [bte_coords.y+j-1][bte_coords.z+k-1].AEast_)
                              -cell[cells_x-1][j][k].T_old)*substrate.k_stab;

                            cell[cells_x-1][j][k].T_old=
                                                    cell[cells_x-1][j][k].T_;
							
							//update the cell's e0
							cell[cells_x-1][j][k].fbound_cell_solve(Nmax_angle); 				
                        }
                    }
                    break;
                case W_:
                    //go through all the cells along the boundary
                    for(int j=1; j<cells_y-1; j++) {
                        for(int k=1; k<cells_z-1; k++) {
                    
					        //get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[0][j][k].get_boundary(W_);
                            cell[0][j][k].T_ =
							cell[0][j][k].T_old +
                            (cell[0][j][k].cell_boundaries[h]->T_
                            +cell[0][j][k].cell_boundaries[h]->Qprime
                            /(cell[0][j][k].cell_boundaries[h]->AEast_)
                              -cell[0][j][k].T_old)*substrate.k_stab;

                            cell[0][j][k].T_old=cell[0][j][k].T_;

							//update the cell's e0
                            cell[0][j][k].fbound_cell_solve(Nmax_angle);
                        }
                    }
                    break;
                case N_:
                    //go through all the cells along the boundary
                    for(int i=1; i<cells_x-1; i++) {
                        for(int k=1; k<cells_z-1; k++) {
                            //get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[i][cells_y-1][k].get_boundary(N_);

                            cell[i][cells_y-1][k].T_ =
							cell[i][cells_y-1][k].T_old +
                            (cell[i][cells_y-1][k].cell_boundaries[h]->T_
                            +cell[i][cells_y-1][k].cell_boundaries[h]->Qprime
                            /(substrate.f_cell[bte_coords.x+i-1]
                            [bte_coords.y+cells_y-3][bte_coords.z+k-1].ANorth_)
                              -cell[i][cells_y-1][k].T_old)*substrate.k_stab;

							cell[i][cells_y-1][k].T_old=
													cell[i][cells_y-1][k].T_;

							//update the cell's e0
                            cell[i][cells_y-1][k].fbound_cell_solve(Nmax_angle);
                        }
                    }
                    break;
                case S_:
                    //go through all the cells along the boundary
                    for(int i=1; i<cells_x-1; i++) {
                        for(int k=1; k<cells_z-1; k++) {

                            //get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[i][0][k].get_boundary(S_);
                            cell[i][0][k].T_ =
							cell[i][0][k].T_old + 
							(cell[i][0][k].cell_boundaries[h]->T_
                            +cell[i][0][k].cell_boundaries[h]->Qprime
                            /(cell[i][0][k].cell_boundaries[h]->ANorth_)
                              -cell[i][0][k].T_old)*substrate.k_stab;

							cell[i][0][k].T_old=cell[i][0][k].T_;

							//update the cell's e0
                            cell[i][0][k].fbound_cell_solve(Nmax_angle);
                        }
                    }
                    break;
                case T_:
                    //go through all the cells along the boundary
                    for(int i=1; i<cells_x-1; i++) {
                        for(int j=1; j<cells_y-1; j++) {

                            //get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[i][j][cells_z-1].get_boundary(T_);
                            cell[i][j][cells_z-1].T_ =
								cell[i][j][cells_z-1].T_old+
								(cell[i][j][cells_z-1].cell_boundaries[h]->T_
							+cell[i][j][cells_z-1].cell_boundaries[h]->Qprime
							/(substrate.f_cell[bte_coords.x+i-1]
							[bte_coords.y+j-1][bte_coords.z+cells_z-3].ATop_)
                              -cell[i][j][cells_z-1].T_old)*substrate.k_stab;

							cell[i][j][cells_z-1].T_old=
													cell[i][j][cells_z-1].T_;
							//update the cell's e0
                            cell[i][j][cells_z-1].fbound_cell_solve(Nmax_angle);
                        }
                    }
                    break;
                case B_:
                    //go through all the cells along the boundary
                    for(int i=1; i<cells_x-1; i++) {
                        for(int j=1; j<cells_y-1; j++) {

                            //get the indices to the corresponding BTE cell
                            //update the temperature for this cell
                            temp = cell[i][j][0].get_boundary(B_);
                            cell[i][j][0].T_ =
                                cell[i][j][0].T_old+( cell[i][j][0].cell_boundaries[h]->T_
                                +
                                cell[i][j][0].cell_boundaries[h]->Qprime/cell[i][j][0].cell_boundaries[h]->ATop_
                                -cell[i][j][0].T_old)*substrate.k_stab;
                                cell[i][j][0].T_old=cell[i][j][0].T_;

							//update the cell's e0
                            cell[i][j][0].fbound_cell_solve(Nmax_angle);
                        }
                    }
                    break;
                default:
                    break;
            }
        }
    }

}

/*==========================================================================*/
//* get the boundary indices
Bound Cell::get_boundary(int dir) {
	int size = boundaries.size();
	for(int h=0; h<size; h++) {
		//only get the indices if they are in the right direction
		if(boundaries[h].dir == dir) {
			return boundaries[h];
		}
	}
	//indicate error and no boundary cell information available
    //return 6;
	Bound none;
    return none;
}

/*==========================================================================*/
Bound Fourier_cell::get_boundary(int dir) {
    int size = boundaries.size();
    for(int h=0; h<size; h++) {
        //only get the indices if they are in the right direction
        if(boundaries[h].dir == dir) {
			return boundaries[h];
        }
    }
	//indicate error and no boundary cell information available
	//return 6;
	Bound none;
	return none;
}

/*==========================================================================*/
//* set the fourier cell temperatures which are in BTE regions
void Device_structure::update_fourier_t_cells() {

    for(int i=1;i<substrate.f_cells_x-1;i++)
    {
        for(int j=1;j<substrate.f_cells_y-1;j++)
        {
            for(int k=1;k<substrate.f_cells_z-1;k++)
            {
                substrate.f_cell[i][j][k].Bprime = 0;
            }
        }
    }


    //go through all BTE structures
    int size = bte_system.size();
    for(int h=0; h<size; h++) {
        //update the fourier cells in BTE regions
        bte_system[h].update_fourier_t_cells(substrate, bte_coords[h]);

    //when calling this function from the output function, this second function
    //"update_Fourier_semireflect" should not be called
		//update the q values if there are semi-reflecting boundaries
		substrate.update_Fourier_semireflect(bte_system[h], bte_coords[h]);

    }

}

/*==========================================================================*/
//* update the fourier cell temperatures which are in BTE regions
void BTE_structure::update_fourier_t_cells(Fourier_structure & substrate,
												Block_coords & bte_coords) {

	//offset indices for the fourier cells
	int x = bte_coords.x;
	int y = bte_coords.y;
	int z = bte_coords.z;

	//go through all cells in the BTE region, but not counting their boundary
	//loop this region from 1 to cells_V-2
	for(int i=1; i<(cells_x-1); i++) {
		for(int j=1; j<(cells_y-1); j++) {
			for(int k=1; k<(cells_z-1); k++) {
				//set the temperature of each Fourier cell to the corresponding
				//BTE cell temperature
				substrate.f_cell[x+i-1][y+j-1][z+k-1].T_ = cell[i][j][k].T_;
			}
		}
	}

}

/*==========================================================================*/
//* initialize the BTE cells temperature after Fourier block has been solved
void Device_structure::initialize_bte_t_cells() {

    //go through all BTE structures
    int size = bte_system.size();
    for(int h=0; h<size; h++) {
        //update the fourier BTE boundaries associated with the BTE block
        bte_system[h].initialize_bte_t_cells(substrate,bte_coords[h]);
    }
}

/*==========================================================================*/
//* update the BTE cell temperatures with new found Fourier values
void BTE_structure::initialize_bte_t_cells(Fourier_structure & substrate,
                                                Block_coords & bte_coords) {

    //offset indices for the fourier cells
    int x = bte_coords.x;
    int y = bte_coords.y;
    int z = bte_coords.z;

    //go through all cells in the BTE region, but not counting their boundary
    //loop this region from 1 to cells_V-2
    for(int i=1; i<(cells_x-1); i++) {
        for(int j=1; j<(cells_y-1); j++) {
            for(int k=1; k<(cells_z-1); k++) {
                //set the temperature of each BTE cell to the corresponding
                //Fourier cell temperature
                cell[i][j][k].T_ = substrate.f_cell[x+i-1][y+j-1][z+k-1].T_;
				//initialize e0 and f_angle of this cell based on the
				//temperature
				cell[i][j][k].fbound_cell_solve(Nmax_angle);
            }
        }
    }

}

/*==========================================================================*/
//* class Fourier_structure constructor
Fourier_structure::Fourier_structure() {

    spec_x_0 = 0;
    spec_x_end = 0;
    spec_y_0 = 0;
    spec_y_end = 0;
    spec_z_0 = 0;
    spec_z_end = 0;
    diff_x_0 = 0;
    diff_x_end = 0;
    diff_y_0 = 0;
    diff_y_end = 0;
    diff_z_0 = 0;
    diff_z_end = 0;
    tbound_x_0 = 0;
    tbound_x_end = 0;
    tbound_y_0 = 0;
    tbound_y_end = 0;
    tbound_z_0 = 0;
    tbound_z_end = 0;
    ins_x_0 = 0;
    ins_x_end = 0;
    ins_y_0 = 0;
    ins_y_end = 0;
    ins_z_0 = 0;
    ins_z_end = 0;

    all_bte_ = 0;
    all_fourier_ = 0;
	transient_ = 0;

    k_stab = 1.0;
}

/*==========================================================================*/
//* calculate vetical heat flow of each cell
void Fourier_structure::calc_heat_flow_cell_bottom(string output_file,
                                     vector<Block_coords> & bte_coords)
{
    ofstream fout_heat;
    output_file = output_file + "_heat_plot";
    fout_heat.open(output_file.c_str());

    double heat_flow_bottom;

    fout_heat<<f_cells_x-2<<" "<<f_cells_y-2<<" "<<bte_coords[0].z-1<<endl;
    for(int i=1;i<f_cells_x-1;i++)
    {
        for(int j=1;j<f_cells_y-1;j++)
        {
            for(int k=1;k<bte_coords[0].z;k++)
            {
                heat_flow_bottom = (f_cell[i][j][k].T_-
                                    f_cell[i][j][k-1].T_)*
                                    f_cell[i][j][k-1].ATop_/(del_x*del_y);
                fout_heat<<cen_x[i]<<" "<<cen_y[j]<<" "<<cen_z[k]<<" ";
                fout_heat<<heat_flow_bottom<<endl;
            }
        }
    }
}

/*==========================================================================*/
//* calculate power going out of the whole system
double Fourier_structure::calc_qout(vector<Block_coords> & bte_coords,
                                    vector<BTE_structure> & bte_system)
{
    double qout_east=0; //heat going out from east boundary
    double qout_west=0; //heat going out from west boundary
    double qout_north=0; //heat going out from north boundary
    double qout_south=0; //heat going out from south boundary
    double qout_top=0; //heat going out from top boundary
    double qout_bottom=0; //heat going out from bottom boundary

    double q_total; //total heat going out

    int i,j,k;

    int size = bte_coords.size();

    //calculate the heat flowing out of each boundary

    //east boundary

    //if the east boundary is not insulating
    if(!ins_x_end)
    { 
        for(j=0;j<f_cells_y;j++)
        {
            for(k=0;k<f_cells_z;k++)
            {
                //check if the cell is a BTE one by checking if it is locked
                if(f_cell[f_cells_x-2][j][k].locked)
                {
                    //calculate the heat flow out using phonons
                
                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].x+bte_coords[h].cells_x-1)==
                            f_cells_x)
                        {
                            if(j>=bte_coords[h].y&&
                               j<=bte_coords[h].y+bte_coords[h].cells_y-2)
                            {
                                if(k>=bte_coords[h].z&&
                                   k<=bte_coords[h].z+bte_coords[h].cells_z-2)
                                {
                                    //calculate the output heat flow from
                                    //this cell
                                    qout_east=qout_east + 
                                              bte_system[h].calc_qout(
                                              j-bte_coords[h].y+1,
                                              k-bte_coords[h].z+1,
                                              E_);
                                    break;
                                }
                            }
                        }
                    }                   
                }
                else
                {
                    //calculate the heat flow out using the conduction equation
                    qout_east=qout_east +
                              (f_cell[f_cells_x-2][j][k].T_-
                               f_cell[f_cells_x-1][j][k].T_)*
                               f_cell[f_cells_x-2][j][k].AEast_;
                }
            }
        }
    }


    //west boundary
    
    //if the west boundary is not insulating
    if(!ins_x_0)
    {
        for(j=0;j<f_cells_y;j++)
        {
            for(k=0;k<f_cells_z;k++)
            {
                //check if the cell is a BTE one by checking if it is locked
                if(f_cell[1][j][k].locked)
                {
                    //calculate the heat flow out using phonons
                
                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].x)==1)
                        {
                            if(j>=bte_coords[h].y&&
                               j<=bte_coords[h].y+bte_coords[h].cells_y-2)
                            {   
                                if(k>=bte_coords[h].z&&
                                   k<=bte_coords[h].z+bte_coords[h].cells_z-2)
                                {
                                    //calculate the output heat flow from this cell
                                    qout_west=qout_west +
                                              bte_system[h].calc_qout(
                                              j-bte_coords[h].y+1,
                                              k-bte_coords[h].z+1,
                                              W_);
                                    break;
                                }
                            }
                        }
                    }
                }
                else
                {
                    //calculate the heat flow out using the conduction equation
                    qout_west=qout_west +
                              (f_cell[1][j][k].T_-
                               f_cell[0][j][k].T_)*
                               f_cell[0][j][k].AEast_;
                }
            }
        }
    }


    //north boundary

    //if the north boundary is not insulating
    if(!ins_y_end)
    {
        for(i=0;i<f_cells_x;i++)
        {
            for(k=0;k<f_cells_z;k++)
            {
                //check if the cell is a BTE one by checking if it is locked
                if(f_cell[i][f_cells_y-2][k].locked)
                {
                    //calculate the heat flow out using phonons

                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].y+bte_coords[h].cells_y-1)==f_cells_y)
                        {
                            if(i>=bte_coords[h].x&&
                               i<=bte_coords[h].x+bte_coords[h].cells_x-2)
                            {
                                if(k>=bte_coords[h].z&&
                                   k<=bte_coords[h].z+bte_coords[h].cells_z-2)
                                {
                                    //calculate the output heat flow from this cell
                                    qout_north=qout_north +
                                               bte_system[h].calc_qout(
                                               i-bte_coords[h].x+1,
                                               k-bte_coords[h].z+1,
                                               N_);
                                    break;
                                }
                            }
                        }
                    }
                }
                else
                {
                    //calculate the heat flow out using the conduction equation
                    qout_north=qout_north +
                               (f_cell[i][f_cells_y-2][k].T_-
                                f_cell[i][f_cells_y-1][k].T_)*
                                f_cell[i][f_cells_y-2][k].ANorth_;
                }
            }
        }
    }

    //south boundary

    //if the south boundary is not insulating
    if(!ins_y_0)
    {
        for(i=0;i<f_cells_x;i++)
        {
            for(k=0;k<f_cells_z;k++)
            {
                //check if the cell is a BTE one by checking if it is locked
                if(f_cell[i][1][k].locked)
                {
                    //calculate the heat flow out using phonons

                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].y)==1)
                        {
                            if(i>=bte_coords[h].x&&
                               i<=bte_coords[h].x+bte_coords[h].cells_x-2)
                            {
                                if(k>=bte_coords[h].z&&
                                   k<=bte_coords[h].z+bte_coords[h].cells_z-2)
                                {
                                    //calculate the output heat flow from this cell
                                    qout_south=qout_south +
                                               bte_system[h].calc_qout(
                                               i-bte_coords[h].x+1,
                                               k-bte_coords[h].z+1,
                                               S_);
                                    break;
                                }
                            }
                        }
                    }
                }
                else
                {
                    //calculate the heat flow out using the conduction equation
                    qout_south=qout_south +
                               (f_cell[i][1][k].T_-
                                f_cell[i][0][k].T_)*
                                f_cell[i][0][k].ANorth_;
                }
            }
        }
    }

    //top boundary

    //if the top boundary is not insulating
    if(!ins_z_end)
    {
        for(int i=0;i<f_cells_x;i++)
        {
            for(int j=0;j<f_cells_y;j++)
            {
                if(f_cell[i][j][f_cells_z-2].locked)
                {
                    //calculate the heat flow out using phonons

                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].z+bte_coords[h].cells_z-1)==f_cells_z)
                        {
                            if(i>=bte_coords[h].x&&
                               i<=bte_coords[h].x+bte_coords[h].cells_x-2)
                            {
                                if(j>=bte_coords[h].y&&
                                   j<=bte_coords[h].y+bte_coords[h].cells_y-2)
                                {
                                    //calculate the output heat flow from this cell
                                    qout_top=qout_top +
                                             bte_system[h].calc_qout(
                                             i-bte_coords[h].x+1,
                                             j-bte_coords[h].y+1,
                                             T_);
                                    break;
                                }
                            }
                        }
                    }
                }
                else
                {
                    qout_top=qout_top +
                             (f_cell[i][j][f_cells_z-2].T_-
                              f_cell[i][j][f_cells_z-1].T_)*
                              f_cell[i][j][f_cells_z-2].ATop_;
                }
            }
        }
    }

    //bottom boundary

    //if the bottom boundary is not insulating
    if(!ins_z_0)
    {
        for(int i=0;i<f_cells_x;i++)
        {
            for(int j=0;j<f_cells_y;j++)
            {
                if(f_cell[i][j][1].locked)
                {
                    //calculate the heat flow out using phonons
    
                    //first find out which BTE structure this cell belongs to
                    for(int h=0;h<size;h++)
                    {
                        //check if it is inside BTE structure h
                        if((bte_coords[h].z)==1)
                        {
                            if(i>=bte_coords[h].x&&
                               i<=bte_coords[h].x+bte_coords[h].cells_x-2)
                            {
                                if(j>=bte_coords[h].y&&
                                   j<=bte_coords[h].y+bte_coords[h].cells_y-2)
                                {
                                    //calculate the output heat flow from this cell
                                    qout_bottom=qout_bottom +
                                                bte_system[h].calc_qout(
                                                i-bte_coords[h].x+1,
                                                j-bte_coords[h].y+1,
                                                B_);
                                    break;
                                }
                            }
                        }
                    }
                }
                else
                {
                    qout_bottom=qout_bottom +
                                (f_cell[i][j][1].T_-
                                 f_cell[i][j][0].T_)*
                                 f_cell[i][j][0].ATop_;
                }
            }
        }
    }

    //calculate total heat flowing out
//    cout<<"East "<<qout_east<<" W,  West "<<qout_west<<" W"<<endl;
//    cout<<"North "<<qout_north<<" W,  South "<<qout_south<<" W"<<endl;
//    cout<<"Top "<<qout_top<<" W,  Bottom "<<qout_bottom<<" W"<<endl;
    
    q_total = qout_east + qout_west + qout_north + qout_south +
              qout_top + qout_bottom;

    return q_total;
}
