/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** heatflow.cc:
****************************************************************************/

#include "header.h"

/*==========================================================================*/
void Fourier_structure::get_vertical_heat_flow(string argv) {

	ofstream fout_plot;	

	//open the output file
    if(argv!="") {
        string meaty = argv;
        string plot = "_plot_v";
        fout_plot.open((meaty+plot).c_str());
    }
    else { //else use default name "output.txt"
        fout_plot.open("output_plot_v.txt");
    }

	//the vertical heat flow
	vector< vector< vector< double > > > v_heat_f;

	//the -1 is there because it is assumed there is no heat flow from the 
	//top of the structure to inside the structure (where top is in z
	//dimension)
	v_heat_f.resize(f_cells_x-2);
	for(int i=1; i<f_cells_x-1; i++) {
		v_heat_f[i-1].resize(f_cells_y-2);
		for(int j=1; j<f_cells_y-1; j++) {
			v_heat_f[i-1][j-1].resize(f_cells_z-2);
			for(int k=1; k<f_cells_z-1; k++) {
			
				//to get the vertical heat flow, use H = (T_above-T_below)*cond
				v_heat_f[i-1][j-1][k-1] = (f_cell[i][j][k+1].T_ 
								- f_cell[i][j][k].T_)*f_cell[i][j][k].ATop_;
				
			}
		}
	}

	//print the heat flow values
	fout_plot<<f_cells_x-2;
    fout_plot<<" ";
	fout_plot<<f_cells_y-2;
    fout_plot<<" ";
	fout_plot<<f_cells_z-2;
    fout_plot<<endl;


	for(int i=1;i<f_cells_x-1;i++) {        
		for(int j=1;j<f_cells_y-1;j++) {
            for(int k=1;k<f_cells_z-1;k++) {
                fout_plot<<cen_x[i]<<" "<<cen_y[j]<<" "<<cen_z[k]<<" ";
                fout_plot<<v_heat_f[i-1][j-1][k-1]<<" ";
                fout_plot<<endl;
			}
		}
	}
	fout_plot.close();

}


/*==========================================================================*/
//calculate the horizontal x dimension heat flow and print it (heat flow from
//right to left)
void Fourier_structure::get_horizontal_x_heat_flow(string argv) {

    ofstream fout_plot;

    //open the output file
    if(argv!="") {
        string meaty = argv;
        string plot = "_plot_hx";
        fout_plot.open((meaty+plot).c_str());
    }
    else { //else use default name "output.txt"
        fout_plot.open("output_plot_hx.txt");
    }

    //the vertical heat flow
    vector< vector< vector< double > > > v_heat_f;

    //the -1 is there because it is assumed there is no heat flow from the
    //right side of the structure to inside the structure
    v_heat_f.resize(f_cells_x-2);
    for(int i=1; i<f_cells_x-1; i++) {
        v_heat_f[i-1].resize(f_cells_y-2);
        for(int j=1; j<f_cells_y-1; j++) {
            v_heat_f[i-1][j-1].resize(f_cells_z-2);
            for(int k=1; k<f_cells_z-1; k++) {

                //to get the horizontal heat flow, 
				//use H = (T_right-T_left)*cond
                v_heat_f[i-1][j-1][k-1] = (f_cell[i+1][j][k].T_
                                - f_cell[i][j][k].T_)*f_cell[i][j][k].AEast_;

                //to make heat flow always positive, so that it can be plotted in logscale
                if(v_heat_f[i-1][j-1][k-1]<0) {
                    v_heat_f[i-1][j-1][k-1] = v_heat_f[i-1][j-1][k-1]*(-1);
                }

            }
        }
    }

    //print the heat flow values
    fout_plot<<f_cells_x-2;
    fout_plot<<" ";
	fout_plot<<f_cells_y-2;
    fout_plot<<" ";
	fout_plot<<f_cells_z-2;
    fout_plot<<endl;
	
    for(int i=1;i<f_cells_x-1;i++) {
		for(int j=1;j<f_cells_y-1;j++) {
			for(int k=1;k<f_cells_z-1;k++) {
                fout_plot<<cen_x[i]<<" "<<cen_y[j]<<" "<<cen_z[k]<<" ";
                fout_plot<<v_heat_f[i-1][j-1][k-1]<<" ";
                fout_plot<<endl;
            }
        }
    }
    fout_plot.close();

}

/*==========================================================================*/
//calculate the horizontal y dimension heat flow and print it (heat flow from
//right to left)
void Fourier_structure::get_horizontal_y_heat_flow(string argv) {

    ofstream fout_plot;

    //open the output file
    if(argv!="") {
        string meaty = argv;
        string plot = "_plot_hy";
        fout_plot.open((meaty+plot).c_str());
    }
    else { //else use default name "output.txt"
        fout_plot.open("output_plot_hy.txt");
    }

    //the vertical heat flow
    vector< vector< vector< double > > > v_heat_f;

    //the -1 is there because it is assumed there is no heat flow from the
    //right side of the structure to inside the structure
    v_heat_f.resize(f_cells_x-2);
    for(int i=1; i<f_cells_x-1; i++) {
        v_heat_f[i-1].resize(f_cells_y-2);
        for(int j=1; j<f_cells_y-1; j++) {
            v_heat_f[i-1][j-1].resize(f_cells_z-2);
            for(int k=1; k<f_cells_z-1; k++) {

                //to get the horizontal heat flow,
                //use H = (T_right-T_left)*cond
                v_heat_f[i-1][j-1][k-1] = (f_cell[i][j+1][k].T_
                                - f_cell[i][j][k].T_)*f_cell[i][j][k].ANorth_;

                //to make heat flow always positive, so that it can be plotted in logscale
                if(v_heat_f[i-1][j-1][k-1]<0) {
                    v_heat_f[i-1][j-1][k-1] = v_heat_f[i-1][j-1][k-1]*(-1);
                }

            }
        }
    }

    //print the heat flow values
    fout_plot<<f_cells_x-2;
    fout_plot<<" ";
    fout_plot<<f_cells_y-2;
    fout_plot<<" ";
    fout_plot<<f_cells_z-2;
    fout_plot<<endl;

    for(int i=1;i<f_cells_x-1;i++) {
        for(int j=1;j<f_cells_y-1;j++) {
            for(int k=1;k<f_cells_z-1;k++) {
                fout_plot<<cen_x[i]<<" "<<cen_y[j]<<" "<<cen_z[k]<<" ";
                fout_plot<<v_heat_f[i-1][j-1][k-1]<<" ";
                fout_plot<<endl;
            }
        }
    }
    fout_plot.close();

}


