/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** hoct_model.h
****************************************************************************/

#ifndef OCT_MODEL_H_
#define OCT_MODEL_H_

#include <iostream>
#include <string>
#include <algorithm>
#include <cstdlib>
#include <cstdio>
#include <numeric>
#include <map>
#include <vector>
#include <list>

/* Number of neighbor dimensions (X_NEG_, X_POS_,Y_NEG_, Y_POS_, 
 * Z_NEG_, Z_POS_) */
const int NEI_C_ = 6;
/* Number of dimensions (X,Y,Z) */
const int DIM_C_ = 3;
const int INV_N_ = -99;
/* ID for grids with constant temperature */
const int CONT_ID_ = -1;
/* ID for isothermal grids */
const int ISOT_ID_ = -2;
/* Number to represent different dimensions */
const int X_ = 0;
const int Y_ = 1;
const int Z_ = 2;
/* Number to represent neighbors in different dimensions */
enum nei_type_ { X_NEG_=0, X_PLU_=1, Y_NEG_=2, Y_PLU_=3, Z_NEG_=4, Z_PLU_=5};
/* Thermal conductivity of Silicon [Watt/(K-mm)] */
const double K_SIL_ = 0.157; 
/* Heat Capacitivity of Silicon [J/(mm^3-K)]*/
const double C_SIL_ = 1.63799e-3;

/* Size information for initalized grid */
const double NORM_SIZE_ = 1.0;
/* Coordinate information for initalized grid */
const int NORM_COOR_ = 0;
/* Grid type */
enum grid_type_ {ISOT_, CONT_, GRID_, CHIP_3D_, COMPONENT_, LAYER_};


/*===========================================================================*/
/* Class used to store the thermal conductivity and heat capacity maps */
struct Map_block {

    double size[2];
    double center[2];
    double value;
};

/*===========================================================================*/
/* Class of the grid used to construct the data structure 
 * Each grid has its own properties such as conductivity, heat capacity, 
 * id, level in the hoct_contour data structure, etc. 
 * Each grid also has points to its parent, children, and neighbor grids */
class par_grid {

    friend std::ostream& operator<<(std::ostream& os, const par_grid & pg);

public:
	par_grid();
	~par_grid();
 
    int id() {return id_;}
    int id() const {return id_;}
    void id_assign(int a) {id_ = a;}

    int self_level() {return self_level_;}
    int self_level() const {return self_level_;}
    void self_level_assign(int a) {self_level_ = a;}

    int act_level() {return act_level_;}
    int act_level() const {return act_level_;}
    void act_level_assign(int a) {act_level_ = a;}

    grid_type_ type() {return type_;}
    void type_assign(grid_type_ a) {type_ = a;}

    par_grid * par() {return par_;}
    par_grid * par() const {return par_;}
    void par_assign(par_grid * a) {par_ = a;}

    std::map<int, int> & levelid() { return levelid_; }
	int levelid(int i) { return levelid_[i];}
	void levelid_assign(int i, int a) {levelid_[i] = a;}
	int levelid_check(int i) { 
		if(levelid_.find(i) == levelid_.end()) {return 0;}
		else{return 1;}}
	void levelid_clear() {levelid_.clear();}

    std::vector<std::vector<par_grid *> > & nei() {return nei_;}
    const std::vector<std::vector<par_grid *> > & nei() const {return nei_;}
    std::vector<par_grid *> & nei(int i) {return nei_[i];}
    const std::vector<par_grid *> & nei(int i) const {return nei_[i];}
    par_grid * nei(int i, int j) {return nei_[i][j];}
    par_grid * nei(int i, int j) const {return nei_[i][j];}
    void nei_assign(int i, par_grid * a) {nei_[i].push_back(a);}

    void nei_clear(int i) {nei_[i].clear();}

    /* Check whether grids are neighbors or not. 
     * If they are neighbors, the returned value represents the direction 
     * (0--x direction, 1--y direction, 3--z direction)of their neighbor 
     * relationship; Otherwise, the returned value will be -1 */
    int nei_check(par_grid * a);

    /* Verify whether the current grid and grid a are neighbors in Z 
     * dimension.
     * Check whether they have intersect area in x and y dimension. 
     * If they do not have overlap area in x/y dimension, then they are 
     * actually not neighbors in Z direction.
     * Return true if they have overlap area, return false if not. */
    int xyintersect_check(par_grid * a);

    /* Calculate the overlapping area in x/y dimension for two grids */
	double area_intersect(par_grid * grid_b);
    /* Calculate the overlapping area in "dim" dimension for two grids */
    double area_intersect(par_grid * grid_b, std::vector<int> dim);

    /* Calculate the overlapping area in x/y dimension between a grid and a
     * conductivity/capacity map block */
    int blk_intersect(const Map_block blk, Map_block & intrsct);

    std::vector<par_grid *> & int_chi_array() {return int_chi_array_;}
    const std::vector<par_grid *> & int_chi_array() const {return int_chi_array_;}
    par_grid * int_chi_array(int i) {return int_chi_array_[i];}
    par_grid * int_chi_array(int i) const {return int_chi_array_[i];}
    void int_chi_array_assign(par_grid * a) {int_chi_array_.push_back(a);}

    std::vector<std::pair<int, double> > & power_rela(){return 
        power_rela_;}
    std::pair<int, double> & power_rela(int i){return 
        power_rela_[i];}
    
	double * sizes() {return sizes_;}
	const double * sizes() const {return sizes_;}

	double sizes(int i) {return sizes_[i];}
	double sizes(int i) const {return sizes_[i];}
    void sizes_assign(int i, double a) {sizes_[i]=a;}
    void sizes_assign(double a, double b, double c) 
        {sizes_[X_]=a; sizes_[Y_]=b; sizes_[Z_]=c;}
	void sizes_assign(double * a) {
		sizes_[X_] =a[X_]; sizes_[Y_] = a[Y_]; sizes_[Z_] = a[Z_];}
	void sizes_assign(const std::vector<double>& s) {
		sizes_[X_] = s[X_]; sizes_[Y_] = s[Y_]; sizes_[Z_] = s[Z_];}

    double * coors() {return coors_;}
	const double * coors() const {return coors_;}

    double coors(int i) {return coors_[i];}
    double coors(int i) const {return coors_[i];}
    void coors_assign(int i, double a) {coors_[i]=a;}
    void coors_assign(double a, double b, double c) 
        {coors_[X_]=a; coors_[Y_]=b; coors_[Z_]=c;}
	void coors_assign(const std::vector<double>& s) {
		coors_[X_] = s[X_]; coors_[Y_] = s[Y_]; coors_[Z_] = s[Z_];}

    std::vector<std::pair<int, par_grid *> > &
        power_nei() {return power_nei_;}
    std::pair<int, par_grid *> & power_nei(int i) {return power_nei_[i];}
	
	std::vector<double> & adjp_coeff() {return adjp_coeff_;}
	double adjp_coeff(int i) {return adjp_coeff_[i];}
	void adjp_coeff_assign(int i, double a) {adjp_coeff_[i] = a;}	
	void adjp_coeff_assign(double a) {adjp_coeff_.push_back(a);}	

    double power() {return power_;}
    void power_assign(double a){power_ = a;}

    double temp() {return temp_;}
    double temp() const {return temp_;}
    void temp_assign(double a) {temp_ = a;}

    double cond() {return cond_;}
    double cond() const {return cond_;}
    void cond_assign(double a) {cond_ = a;}

    double bulk_cond() {return bulk_cond_;}
    double bulk_cond() const {return bulk_cond_;}
    void bulk_cond_assign(double a) {bulk_cond_ = a;}

    std::vector<Map_block> & cond_map() {return cond_map_;}
    void cond_map_assign(const std::vector<Map_block>& a) {cond_map_ = a;}

    double ccap() {return ccap_;}
    double ccap() const {return ccap_;}
    void ccap_assign(double a) {ccap_ = a;}

    double bulk_ccap() {return bulk_ccap_;}
    double bulk_ccap() const {return bulk_ccap_;}
    void bulk_ccap_assign(double a) {bulk_ccap_ = a;}

    std::vector<Map_block> & ccap_map() {return ccap_map_;}
    void ccap_map_assign(const std::vector<Map_block>& a) {ccap_map_ = a;}

	int * partition_flag() {return partition_flag_;}
	const int * partition_flag() const {return partition_flag_;}
	int partition_flag(int i) {return partition_flag_[i];}
	int partition_flag(int i) const {return partition_flag_[i];}
	void partition_flag_assign(int x, int y, int z) {
		partition_flag_[X_] = x;
		partition_flag_[Y_] = y;
		partition_flag_[Z_] = z;
	}
	void partition_flag_assign(int * a) {
		partition_flag_[X_] = a[X_];
		partition_flag_[Y_] = a[Y_];
		partition_flag_[Z_] = a[Z_];
	}

	int gid() {return gid_;}
	void gid_assign(int a) {gid_ = a;}

    std::vector<std::vector<par_grid *> > & chi_array() { return chi_array_; }
    const std::vector<std::vector<par_grid *> > & chi_array() const
        { return chi_array_; }
    std::vector<par_grid *> & chi_array(int i) { return chi_array_[i]; }
    const std::vector<par_grid *> & chi_array(int i) const
        { return chi_array_[i]; }
    par_grid * chi_array(int i, int j) {return chi_array_[i][j];}
    par_grid * chi_array(int i, int j) const {return chi_array_[i][j];}
    void chi_array_assign(par_grid * a, int i)
        { chi_array_[i].push_back(a); }

private:

    /* The id_ of one grid equals to the index according to which
     * all the children grids of its parent grid are generated */ 
	int id_;
    
    /* The self_level_ of one grid equals to the maximum of all its
     * children's self_level decrease by 1; While all the grids in
     * the deepest level(leave nodes)' self_level_ equals to the
     * number of levels of the tree data structure. */
    int self_level_;
    
    /* The root grid's actual level (act_level) equals to 0; And 
     * one grid's act_level equals to its parent's actual level 
     * increased by 1 (par()->act_level()+1)*/
    int act_level_;
    
    /* The type of a grid, (isothermal grid, constant tempearture grid,
     * normal thermal grid), or (chip level grid, component level grid, 
     * layer level grid)*/
    grid_type_ type_;
    
    /* Point to parent grid */
	par_grid * par_;
    
    /* levelid_ is the global id for the grids in the same level of the
     * hoct_contour data structure. */
    std::map<int, int> levelid_;
    
    /* vectors of points to all the neighbors */
	std::vector<std::vector<par_grid *> > nei_;
    
    /* Points to direct children grids */
	std::vector<par_grid *> int_chi_array_;
    
    /* the int inside the map defines the direction of the power grid 
     in contrast to the current grid, the par_grid pointer points to
     the recorded power grid
     the vector is to represent the case that the current grid has
     multiple power grid neighbors */
    std::vector<std::pair<int, par_grid *> > power_nei_;    
	
    /* Store the coefficient to adjust the power vector, the vector will have
	 * the same number of elements as power_nei_, and they follow the same order */
	std::vector<double> adjp_coeff_;	

    /* Store the power mapping information/relationship 
     * from the 2D power structure, the int stores which power element (the index
     * in the vector<vector> > power_profile structure, and the double stores
     * the coefficient the power element is mapping on the grid */
    std::vector<std::pair<int, double> > power_rela_;
   
    /* Sizes information */
    double sizes_[DIM_C_];
    
    /* Coordinates information */
    double coors_[DIM_C_];	
    
    /* Power information */
	double power_;
    
    /* Temperature information */
	double temp_;
    
    /* Conductivity information */
	double cond_;
    double bulk_cond_;
    std::vector<Map_block> cond_map_;
    
    /* Heat capacitivity information */
	double ccap_;
    double bulk_ccap_;
    std::vector<Map_block> ccap_map_;
    
    /* Paritioning flag information */
	int partition_flag_[DIM_C_];
    
    /* Global id information */
	int gid_;
    
    /* Points to children grids */
    std::vector<std::vector<par_grid *> > chi_array_;

};

/* grid_contour class to be used in hoct_contour data structure */
class grid_contour {
public:

	grid_contour();
	~grid_contour();
	grid_contour(par_grid * a);
	void clear(){self_ = NULL; nei_.clear();}

    par_grid * self() {return self_;}
    std::vector<std::vector<par_grid *> > & nei() {return nei_;}
    std::vector<par_grid *> & nei(int i) {return nei_[i];}
    par_grid * nei(int i, int j) {return nei_[i][j];}

private:

    /* Point to the grid constructing this grid_contour */
    par_grid * self_;
    /* Points to the neighbors */
	std::vector<std::vector<par_grid *> > nei_;
};

/* Class to store the conductances and capacitance of elements, together
 * with the indices showing the conductances are calculated between which
 * two grids (grids are differentiated by levelid) */
class RC_network_type {
public:
	RC_network_type();

	int size_;
	std::vector<std::vector<int> > A_;
	std::vector<std::vector<double> > G_;
	std::vector<double> C_;

	void clear() {A_.clear(); G_.clear(); C_.clear();}
	void clear(RC_network_type & AGC);// {A_.clear(); G_.clear(); C_.clear();}
	void resize(int size) {size_ = size; A_.resize(size); 
		G_.resize(size); C_.resize(size,0);}
};


/* Class to maintain a grid's logic position, sizes and 
 * number of divisions information */
class par_logic_pos{
public:
	par_logic_pos();
	
	/* Logic position information */
	std::vector<std::vector<int> > lpos_;
	/* Pointer of the according grid based on the same sequence
	 * as the logic_pos_ */
	std::vector<par_grid *> grid_;
	/* Number of divisions */
	std::vector<int> divs_;
	/* Sizes info. (x/y dimension) */
	std::vector<double> sizes_;
};

/* Do the refinement for "root" based on the "par",
 * "par" defines the children of X,Y,Z dimension of "root" when further 
 * refined. */
void gen_par_tree(par_grid * root, const std::vector<int> & par);

/* Find the upper_level grid (in level "upper_level") for the "cur_grid".
 * Each grid are in levels (par()->(self_level()+1)~ grid->self_level()),
 * so if par->self_level() equal to upper_level, par() is the 
 * upper_level grid, otherwise, cur_grid is the upper_level grid */
par_grid * upper_level_grid(par_grid * cur_grid, int upper_level);

/* Generate a contour structure (for hetergeneous partition, contour is 
 * necessary) for "root" based on the tree structure */
int hoct_tree_contour(par_grid * root, std::vector<std::vector<grid_contour> > 
    & hoct_contour, std::vector<int> & lid, int depth);

/* Generate the first 4 levels of the hoct_contour, which won't change when 
 * further partition is conducted */
void hoct_cont_init(par_grid * root, std::vector<std::vector<grid_contour> > & 
    hoct_contour);

/* Generate the contour structure based on the grid (push the grid and 
 * all of its children info. to contour ) */
void push_grid_contour(par_grid * grid, 
    std::vector<std::vector<grid_contour> > & hoct_contour, int start_level);

/* Assign levelid information to isothermal and constant temperature grids */
void assign_levelid_isotcont(
    std::vector<std::vector<grid_contour> > & hoct_contour, int depth);

/* Update the hoct_contour levels greater than 3 after refinement */
void update_contour(par_grid * root, std::vector<std::vector<grid_contour> > & 
    hoct_contour, std::vector<int> & lid, int depth);

/* Update the hoct_contour levels greater than 3 after refinement */
void update_contour_elts(par_grid * root, 
    std::vector<std::vector<grid_contour> > & hoct_contour, 
    std::vector<int> & lid);

/* Assign self_level for all the elements except for the first three layers
 * of elements */
int assign_self_level(par_grid * root, int depth);

/* Assign neighbor information to the hoct_cotnour grid based on the 
 * current and neighbor information in the "upper_level" */
inline void geo_nei_assign(par_grid * curr_grid, par_grid * upper_grid, 
	int dim, int curr_level, int upper_level,
	int gid, std::vector<std::vector<grid_contour> > & hoct_contour);

/* Generate the AGC structure for one specific level of the grids in 
 * the "hoct_contour" data structure */
void hoct_contour_AGC(std::vector<grid_contour> & hoct_contour, 
	RC_network_type & AGC, int level);

/* Check whether a grid is the neighbor of the input power layer whose 
 * coordinate in Z direction is "player_zcoor" */
char layer_nei(grid_contour & grid_cont, double player_zcoor);

/* Check whether a grid is the neighbor of the input power layer whose 
 * coordinate in Z direction is "player_zcoor" */
char layer_nei(par_grid * grid, double player_zcoor);

/* Apply the correct neighbor information for the "hoct_contour" data 
 * stucture */
int hoct_contour_neighbor(std::vector<std::vector<grid_contour> > 
	& hoct_contour);

/* Check whether grid a and grid b are neighbors in dimension i */
int neighbor_check(int i, par_grid * a, par_grid * b) ;

/* Clear the levelids for the grids in levels greater than 3 */
void clear_levelids(par_grid * root, int depth);

/* Calculate the capacitance of grid "gA" */
double cal_cap(par_grid * gA);

/* Calculate the conductance between grids grid "gA" and grid "gB" in dimenion 
 * "dim" */
double cal_cond(par_grid * gA, par_grid * gB, int dim);

/* Calculate the conductance of grid "gA" in dimension "dim" */
double cal_cond(par_grid * gA, int dim);

/* Calculate the equivalent resisistance of two neighbor grids ("gA" and
 * "gB" in Z dimension, in order to calculate the power share when 2D power 
 * is transfered to 3D elements */
std::pair<double, double> cal_eq_resis(par_grid *gA, par_grid * gB);

/* 2D material layer's z position value is zpos
 * for 3D elements, to check whether it is the lower neighbor of the above 2D
 * material  
 *     2D layer (zpos) 
 *     3D els */
bool is_uppernei(const par_grid * cur_grid, const double & zpos);

/* 2D material layer's z position value is zpos
 * for 3D elements, to check whether it is the lower neighbor of the above 2D
 * material  
 *     2D layer (zpos) 
 *     3D els */
bool is_lowernei(const par_grid * cur_grid, const double & zpos);

/* Find the number of level of children for "root" */
int check_depth(par_grid * root); 

/* Check the sanity of the levelid for root and all its children */
void check_levelids(par_grid * root);

#endif /* hoct_model.h */

