/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** init.cc
****************************************************************************/

#include "init.h"
#include "util.h"

using namespace std;
using namespace isac;

/*===========================================================================*/
/* Initialize the "isot"(isothermal grid) and "cont"(constant temperature 
 * grid) grids */
void isotcont_init(par_grid & isot, par_grid & cont){

	isot.type_assign(ISOT_);
	isot.levelid_assign(0, -2);
	cont.type_assign(CONT_);
	cont.levelid_assign(0, -1);

}


/*===========================================================================*/
/* Initiate the hoct_contour data structure based on the "root" grid */
/* If the hoct_contour only has 4 levels, then hoct_cont_init() function
 * call will generate the correct hoct_contour data structure. Otherwise,
 * hoct_contour_update() function is called to generate the levels other
 * than the first 4 levels */
/* The information of the first 4 levels: which are preserved the same after
 * initiation, even though further refinement is conducted for some elements.
 * Level 0: One grid only, which reparesents the whole chip package;
 * Level 1: Grid number equal to the number of components. Each grid represents
 * one component; 
 * Level 3: Grid number equal to the number of material and interface layers.
 * Each grid represents one material layer or one interface layer;
 * Level 4: Grid representing all the elements after the initial partition.
 * Grids in Level 1 are direct children of Level 0 grid;
 * Grids in Level 2 are direct children of grids in Level 1;
 * Grids in Level 3 are direct children of grids in Level 2. */
void hoct_contour_init(vector<vector<grid_contour> > & hoct_contour, 
    par_grid * root, par_grid * isot, par_grid * cont) {

	int levels = check_depth(root);
    hoct_contour.resize(levels);

    vector<int> lid;
    lid.resize(levels,0);

	/* Form the contour */
    /* Generate the first 4 levels of the hoct_contour, and these
     * 4 levels won't change when further partition is conducted */
    hoct_cont_init(root, hoct_contour);
    
    /* If the contour has more than 4 levels, then the contour needs 
     * to be generated by hoct_tree_contour_update */
    if (levels > START_LEVEL_){
        hoct_contour_update(hoct_contour, root);
    }

    /*Set neighbors for level 0 grid */
	for(int z = 0; z < levels; z++) {
       for(unsigned int p = 0; p < hoct_contour[z].size(); p++) {
            hoct_contour[z][p].nei().resize(NEI_C_);
       }
    }
    assert(hoct_contour[0][0].self() == root);

	hoct_contour[0][0].nei(0).push_back(isot);
    hoct_contour[0][0].nei(1).push_back(isot);
    hoct_contour[0][0].nei(2).push_back(isot);
    hoct_contour[0][0].nei(3).push_back(isot);
    hoct_contour[0][0].nei(4).push_back(isot);
    hoct_contour[0][0].nei(5).push_back(cont);
    
}


/*===========================================================================*/
/* Generate the hoct_contour data structure (levels greater than 3) based on 
 * the partition of the "root" */ 
void hoct_contour_update(vector<vector<grid_contour> > & hoct_contour, 
    par_grid * root) {

	int levels = check_depth(root);

    vector<int> lid;
    lid.resize(levels, 0);

	/* Form the levels greater than 3 of the hoct_contour */
    update_contour(root, hoct_contour, lid, levels);
   
    /* Set the size of neighbors for the elements in hoct_contour 
     * data structure (6 dimensions, X_NEG_, X_POS_, Y_NEG_, Y_POS_,
     * Z_NEG_, Z_POS_), so the first diemension of the nei() of each
     * element in hoct_contour is set to be 6 (NEI_C_)) */ 
    assert(hoct_contour[0][0].self() == root);
	for(int z = START_LEVEL_; z < levels; z++) {
       for(unsigned int p = 0; p < hoct_contour[z].size(); p++) {
            hoct_contour[z][p].nei().resize(NEI_C_);
       }
    }
}


/*===========================================================================*/
/* Generate the all the neighbor information for the grid_contours in level 
 * 3 (This level of grid_contours never change) hoct_contour data structure
 * (based on the neighbor information from the "root") */
void contour_nei_init(par_grid * root, vector<vector<grid_contour> > 
	& hoct_contour, int init_level){

    /* First generate the correct neighbor information for the 
     * 4rd level, which won't change even further partition
     * if conducted */
    int check_level = START_LEVEL_ - 1;
    assert(check_level <= init_level);
	int comp_num = static_cast<int>(root->int_chi_array().size());
	int gid = 0;
    vector<grid_contour> & level_contour = hoct_contour[check_level];

    for (int i = 0; i < comp_num; ++i){
		par_grid * cur_comp = root->int_chi_array(i);
		int layer_num = cur_comp->int_chi_array().size();
		for (int j = 0; j < layer_num; ++j){
			par_grid * cur_layer = cur_comp->int_chi_array(j);
			int elts_num = cur_layer->int_chi_array().size();
			for (int k = 0; k < elts_num; ++k){
				par_grid * cur_elt = cur_layer -> int_chi_array(k);
				for (int l = 0; l < NEI_C_; ++l){
					int nei_cnt = cur_elt->nei(l).size();
					grid_contour & cur_cgrid = level_contour[gid];
					for (int m = 0; m < nei_cnt; ++m){
						cur_cgrid.nei(l).push_back(cur_elt->nei(l,m));	
					}
				}
			    gid++;
			}
		}
	}
    assign_levelid_isotcont(hoct_contour, START_LEVEL_);
    
    /* If the hoct_contour has more than 4 levels, the following
     * set the correct neighbor information for the levels greater
     * than 4 */ 
    if (init_level >= START_LEVEL_){
        hoct_contour_neighbor(hoct_contour);
    }
}



		
