/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** initialization.cc: execute grid partitioning and obtain initial values
****************************************************************************/

#include "header.h"

/*==========================================================================*/
//initialize the values for the entire structure
void Device_structure::initialization() {

    //initialize the Fourier_structure and reserve memory for cells
    substrate.set_grid();
    //and define its sources
    substrate.define_sources_f(sources);

    get_total_input_heat();

    get_material_coords();

    //initialize materials in the Fourier structure
    substrate.set_null();

    //check if required to solve all substrate using BTE or Fourier
    if(substrate.all_bte_==0&&substrate.all_fourier_==0)
    //if not
    {
        //define the BTE regions and create a BTE_structre for each
        set_solvers();
    }
    else if(substrate.all_bte_==1)
    //solve using BTE
    {
        set_all_bte();
    }
    //else there will be no initialization for BTE structures

    //set the rest of the BTE_structure variables that are not
    //initialized in set_solvers() (BTE_structure-independent variables)
    //and reserve memory for cells
    set_BTE_structures();
    
    //indicate which BTE boundaries are touching the Fourier boundaries
    get_adjacent_boundaries();

    //set the material of each cell in BTE regions
    set_material();
    
    int size=bte_system.size();
    for(int h=0;h<size;h++)
    {
        bte_system[h].ini_solve();
    }
    //set the boundaries of the BTE to constant temperature boundaries if 
    //they are touching Fourier regions
    set_const_bound();

	//initialize the Fourier block values
    //reset all coefficients to 0 for Fourier cells
    substrate.reset_fourier();
    //set the initial guesses and conditions for the Fourier cells
    substrate.ini_solve_fourier();
    //initialize fourier coefficients and set insulating boundaries
    substrate.obtain_coeffs_fourier();

}

/*==========================================================================*/
//set the rest of the BTE_structure variables that are not
//initialized in set_solvers() (BTE_structure-independent variables)
void Device_structure::set_BTE_structures()
{
    int Nmax_angle;
    double del_theta,del_phi;
    vector<double> cen_theta,cen_phi;
    vector<Propagation> propvec;


    int size=bte_system.size();

    for(int i=0;i<size;i++)
    {
        //set angular domain

	    //get control angle width
	    Nmax_angle = N_theta[i]*N_phi[i]*8;
	    del_theta = 0.5*PI/N_theta[i];
	    del_phi = 0.5*PI/N_phi[i];

	    cen_theta.reserve(2*N_theta[i]);
	    cen_phi.reserve(4*N_phi[i]);

	    //find the centroids of all the control angles
	    cen_theta.push_back(0.5*del_theta);
	    cen_phi.push_back(0.5*del_phi);

	    for(int a=1; a<(N_theta[i]*2); a++) {
		    cen_theta.push_back(cen_theta[a-1]+del_theta);
	    }

	    for(int a=1; a<(N_phi[i]*4); a++) {
		    cen_phi.push_back(cen_phi[a-1]+del_phi);
	    }

	    //get all the direction extents and the propagation vectors

	    int temp_theta = N_theta[i]*2;// - 1;
	    int temp_phi = N_phi[i]*4;//- 1;
	    int cur_angle = 0; //the indicator of the current angle

	    //make enough room for the vector
	    propvec.resize(Nmax_angle);

	    for(int a=0; a<temp_theta; a++) {
		    for(int b=0; b<temp_phi; b++) {
			    //get the current angle
			    cur_angle = b + a*4*N_phi[i];
			
			    //get S vector components
			    propvec[cur_angle].set_components(cen_phi[b], cen_theta[a], 
														del_phi,del_theta);

		    }
	    }
        
	    //reserve memory for cells
	    bte_system[i].cell.resize(bte_system[i].cells_x);
	
	    for(int a=0;a<bte_system[i].cells_x;a++)
	    {
		    bte_system[i].cell[a].resize(bte_system[i].cells_y);
	    }
	
	    for(int a=0;a<bte_system[i].cells_x;a++)
	    {
		    for(int b=0;b<bte_system[i].cells_y;b++)
		    {
			    bte_system[i].cell[a][b].resize(bte_system[i].cells_z);
		    }
	    }
        

        bte_system[i].del_x = substrate.del_x; 
        bte_system[i].del_y = substrate.del_y;
        bte_system[i].del_z = substrate.del_z;

        bte_system[i].N_phi = N_phi[i];
        bte_system[i].N_theta = N_theta[i];
        bte_system[i].Nmax_angle = Nmax_angle;
        bte_system[i].del_theta = del_theta;
        bte_system[i].del_phi = del_phi;
        bte_system[i].cen_theta = cen_theta;
        bte_system[i].cen_phi = cen_phi;
        bte_system[i].propvec = propvec;
        
        bte_system[i].T_ini = substrate.T_ini;
        
		//set the reflecting values of the bte system to the corresponding
		//values, however this may be overridden if it hits a Fourier wall
        bte_system[i].oxide_x_0 = oxide_x_0[i];
        bte_system[i].oxide_x_end = oxide_x_end[i];
        bte_system[i].oxide_y_0 = oxide_y_0[i];
        bte_system[i].oxide_y_end = oxide_y_end[i];
        bte_system[i].oxide_z_0 = oxide_z_0[i];
        bte_system[i].oxide_z_end = oxide_z_end[i];

        for(int dir=0;dir<6;dir++)
        {
            bte_system[i].trans_factor.push_back(trans_factor[i][dir]);
        }

        bte_system[i].reset();

        bte_system[i].define_sources(sources);
    }
}

/*==========================================================================*/
//Function to indicate which BTE boundaries are touching the Fourier boundaries
void Device_structure::get_adjacent_boundaries() {

	int size = bte_system.size();

	//go over all bte structures
	for(int l=0; l<size; l++) {		

		//go over all boundaries
		for(int h=0; h<6; h++) {
		
			//see if the boundaries have a Fourier neighbour
			if(bte_system[l].bte_boundaries[h]==1) {		
			//if yes
				//single boundary cell index
				int cell_b1=0, cell_b2=0, cell_b3=0;

				//if boundary is E
				if(h==0) {
					//get the Fourier index at the boundary
					//east cell of BTE region +1
					cell_b1 = bte_coords[l].x + bte_coords[l].cells_x -2;//+1

					//find the Fourier cell index at the south bottom
					cell_b2 = bte_coords[l].y ;
					cell_b3 = bte_coords[l].z ;
				}
				//else if W
				else if(h==1) {
					//get the Fourier index at the boundary
					//west cell of the BTE region -1
                    cell_b1 = bte_coords[l].x - 1;

                    //find the Fourier cell index at the south bottom
                    cell_b2 = bte_coords[l].y ;
                    cell_b3 = bte_coords[l].z ;
                }				
				else if(h==2) { //N
					//get the Fourier index at the boundary
                    cell_b2 = bte_coords[l].y + bte_coords[l].cells_y -2;//+1

                    //find the Fourier cell index at the bottom west
                    cell_b3 = bte_coords[l].z ;
                    cell_b1 = bte_coords[l].x ;
                }
                else if(h==3) { //S
                    //get the Fourier index at the boundary
                    cell_b2 = bte_coords[l].y - 1;

                    //find the Fourier cell index at the bottom west
                    cell_b3 = bte_coords[l].z ;
                    cell_b1 = bte_coords[l].x ;
                }
				else if(h==4) { //T
                	//get the Fourier index at the boundary
                    cell_b3 = bte_coords[l].z + bte_coords[l].cells_z -2;//+1

                    //find the Fourier cell index at the south west
                    cell_b2 = bte_coords[l].y ;
                    cell_b1 = bte_coords[l].x ;
				} 
				else if(h==5) { //B
					//get the Fourier index at the boundary
                    cell_b3 = bte_coords[l].z -1;

                    //find the Fourier cell index at the south west
                    cell_b2 = bte_coords[l].y ;
                    cell_b1 = bte_coords[l].x ;
                } 

				//temporary variables
				int cells_x = bte_coords[l].cells_x   -2;
				int cells_y = bte_coords[l].cells_y   -2;
				int cells_z = bte_coords[l].cells_z   -2;

				//loop over all cells in this boundary

				//if boundary is in x
				if((h==0)||(h==1)) {
					//loop for increasing y and z for all cells at the boundary
					for(int j=0; j<cells_y; j++) {
						for(int k=0; k<cells_z; k++) {

							//======set the pointers=============//
							//this cell is the boundary cell
							bte_system[l].cell[((1-h)*(cells_x-1+2))][j+1][
                                            k+1].cell_boundaries.resize(6);
                            //initialize its direction and index values
                            bte_system[l].cell[((1-h)*(cells_x-1+2))][j+1][
                                    k+1].cell_boundaries[h] =
							&(substrate.f_cell[cell_b1][cell_b2+j][cell_b3+k]);
                            //set the Fourier cell neighbour
                            //make sure to initialize this cell's boundary
                            substrate.f_cell[cell_b1][cell_b2+j][
                                        cell_b3+k].cell_boundaries.resize(6);
                            //initialize its direction and index values
                            substrate.f_cell[cell_b1][cell_b2+j][
                                            cell_b3+k].cell_boundaries[h]=
                                &(bte_system[l].cell[((1-h)*(cells_x-1+2))][j+1][k+1]);
							//======end of set the pointers========//

						}
					}
				//end if
				}
				//else if boundary is in y
				if((h==2)||(h==3)) {
					//loop for increasing z and x for all cells at the boundary
                    for(int k=0; k<cells_z; k++) {
                        for(int i=0; i<cells_x; i++) {

                            //======set the pointers=============//
                            bte_system[l].cell[i+1][((3-h)*(cells_y-1+2))][
                                            k+1].cell_boundaries.resize(6);
                            //initialize its direction and index values
                            bte_system[l].cell[i+1][((3-h)*(cells_y-1+2))][
                                    k+1].cell_boundaries[h] =
                            &(substrate.f_cell[cell_b1+i][cell_b2][cell_b3+k]);
                            //set the Fourier cell neighbour
                            //make sure to initialize this cell's boundary
                            substrate.f_cell[cell_b1+i][cell_b2][
                                        cell_b3+k].cell_boundaries.resize(6);
                            //initialize its direction and index values
                            substrate.f_cell[cell_b1+i][cell_b2][
                                            cell_b3+k].cell_boundaries[h]=
                            &(bte_system[l].cell[i+1][((3-h)*(cells_y-1+2))]
																		[k+1]);
                            //======end of set the pointers========//

                        }
                    }
				//end if
				}
				//else if boundary is in z
				if((h==4)||(h==5)) {
					//loop for increasing y and x for all cells at the boundary
                    for(int j=0; j<cells_y; j++) {
                        for(int i=0; i<cells_x; i++) {

                            //======set the pointers=============//
                            bte_system[l].cell[i+1][j+1][((5-h)*(cells_z-1+2))]
                                            .cell_boundaries.resize(6);
                            //initialize its direction and index values
                            bte_system[l].cell[i+1][j+1][((5-h)*(cells_z-1+2))]
                                    .cell_boundaries[h] =
                            &(substrate.f_cell[cell_b1+i][cell_b2+j][cell_b3]);
                            //set the Fourier cell neighbour
                            //make sure to initialize this cell's boundary
                            substrate.f_cell[cell_b1+i][cell_b2+j][
                                        cell_b3].cell_boundaries.resize(6);
                            //initialize its direction and index values
                            substrate.f_cell[cell_b1+i][cell_b2+j][
                                            cell_b3].cell_boundaries[h]=
                            &(bte_system[l].cell[i+1][j+1][((5-h)*
															(cells_z-1+2))]);
                            //======end of set the pointers========//
							
                        }
                    }
				//end if
				}

			//end if statement looking if this bte region had a neighbour
			}

		}

	//end looking at all bte structures
	}
//end function
}

/*==========================================================================*/
//Function to set the material of each cell in BTE regions
void Device_structure::set_material() {

	int size = bte_system.size();

    //go over all BTE structures
	for(int l=0; l<size; l++) {

		//initialize all cells as substrate
		//go over all cells
        for(int i=0; i<bte_system[l].cells_x; i++) {
            for(int j=0; j<bte_system[l].cells_y; j++) {
                for(int k=0; k<bte_system[l].cells_z; k++) {
					 bte_system[l].cell[i][j][k].material =
												substrate.material_full ;
				}
			}
		}

		//for each BTE region, go over all the materials in that region, not
		//counting the substrate
		int sizer = bte_system[l].mat_coords.size();

	    //go over all the materials within the the BTE region
    	for(int h=0; h<sizer; h++) {
	        int end_x = bte_system[l].mat_coords[h].x 
										+ bte_system[l].mat_coords[h].cells_x;
    	    int end_y = bte_system[l].mat_coords[h].y 
										+ bte_system[l].mat_coords[h].cells_y;
        	int end_z = bte_system[l].mat_coords[h].z 
										+ bte_system[l].mat_coords[h].cells_z;

	        //go to the first point of the region where this material is
    	    //and loop over all cells in this region
        	for (int i=bte_system[l].mat_coords[h].x; i<end_x+1; i++) {
            	for (int j=bte_system[l].mat_coords[h].y; j<end_y+1; j++) {
                	for (int k=bte_system[l].mat_coords[h].z; k<end_z+1; k++) {
                    	bte_system[l].cell[i][j][k].material = 
										&(bte_system[l].mat_coords[h].material);
	                }
    	        }
        	}
    	}

    //end BTE structures
	}

//end function
}



/*==========================================================================*/
//Function to initialize the null materials in the Fourier structure
//and the material of all the other cells to the substrate material
void Fourier_structure::set_null() {
    
	//initialize all cells in Fourier region to the substrate material
	//go over all cells in x,y,z
	for(int i=0; i<f_cells_x; i++) {
		for(int j=0; j<f_cells_y; j++) {
			for(int k=0; k<f_cells_z; k++) {
	
				//set the material to substrate (from the class itself)
				f_cell[i][j][k].material = material_full;
			}
		}
	}

	int size = mat_coords.size();
	//go over all the materials within the Fourier region and set the
	//corresponding material to the cells
	for(int	l=0; l<size; l++) {
		int end_x = mat_coords[l].x + mat_coords[l].cells_x;
		int end_y = mat_coords[l].y + mat_coords[l].cells_y;
		int end_z = mat_coords[l].z + mat_coords[l].cells_z;
		
		//go to the first point of the region where this material is
		//and loop over all cells in this region
		for (int i=mat_coords[l].x; i<end_x; i++) {
			for (int j=mat_coords[l].y; j<end_y; j++) {
				for (int k=mat_coords[l].z; k<end_z; k++) {
					f_cell[i][j][k].material = &(mat_coords[l].material);
				}
			}
		}
	}

//end function
}

/*==========================================================================*/
void Bound::set_values(int dir_,int i,int j,int k) {
	dir = dir_;
	x=i;
	y=j;
	z=k;
}

/*==========================================================================*/
//set the boundaries of the BTE to constant temperature boundaries if they are
//touching Fourier regions
void Device_structure::set_const_bound() {

	int size = bte_system.size();	

	//go over all bte regions
	for(int l=0; l<size; l++) {

		//go over all boundaries
        for(int h=0; h<6; h++) {

            //if the boundaries have a Fourier neighbour
            if(bte_system[l].bte_boundaries[h]==1) {
						
				//set the correct boundaries to constant temperature boundaries	
				switch (h) {
                	case E_:
						bte_system[l].tbound_x_end=1;
	                    break;
    	            case W_:
						bte_system[l].tbound_x_0=1;
                    	break;
					case N_:
                        bte_system[l].tbound_y_end=1;
                        break;
                    case S_:
                        bte_system[l].tbound_y_0=1;
                        break;
					case T_:
                        bte_system[l].tbound_z_end=1;
                        break;
                    case B_:
                        bte_system[l].tbound_z_0=1;
                        break;
                	default:
                    	break;
				}

            }
		}

	}
	
}


/*==========================================================================*/
//Material_coords Class constructor
Material_coords::Material_coords(int x_index, int y_index, int z_index,
                             int num_cells_x, int num_cells_y, int num_cells_z,
                             Material mat) {

    x = x_index;
    y = y_index;
    z = z_index;
    cells_x = num_cells_x;
    cells_y = num_cells_y;
    cells_z = num_cells_z;
    material = Material(mat.name,mat.k_cond,mat.t_ref_,mat.cv_,mat.vg_,
                        mat.t_rel_);
}

/*==========================================================================*/
//Material Class constructor
Material::Material(string mat_name, double k_conduct, double t_ref, double cv,
             double vg, double t_rel) {

    name = mat_name;
    k_cond = k_conduct;
    t_ref_ = t_ref;
    cv_ = cv;
    vg_ = vg;
    t_rel_ = t_rel;
}

/*==========================================================================*/
//set the type of solver (BTE or Fourier) for a structure
void BTE_structure::set_solver(Material& material) {

    double limit = 10*material.vg_*material.t_rel_;
    if(length_x<limit||length_y<limit||length_z<limit)
    {
        all_bte_ = 1;
    }
}

/*==========================================================================*/
//initialize the values for a structure
void BTE_structure::initialization() {

    //set up the cartesian grid
    set_grid();

    //set the angular variables/component values
    set_angular();

    //reset all coefficients to 0 and initialize the f_angles for BTE cells
    reset();

	//set the initial guesses and conditions for the BTE cells
    ini_solve();

}

/*==========================================================================*/
//set up the cartesian grid
void Fourier_structure::set_grid() {

	//reserve memory for fourier cells
	f_cell.resize(f_cells_x);
	
	for(int i=0;i<f_cells_x;i++)
	{
		f_cell[i].resize(f_cells_y);
	}
	
	for(int i=0;i<f_cells_x;i++)
	{
		for(int j=0;j<f_cells_y;j++)
		{
			f_cell[i][j].resize(f_cells_z);
		}
	}

	del_x = length_x/(f_cells_x-2);
	del_y = length_y/(f_cells_y-2);
	del_z = length_z/(f_cells_z-2);

	//UNIFORM GRID SETUP
	//------------------------------------//
    //get the edges of the cell
    edge_x.reserve(f_cells_x);
    edge_y.reserve(f_cells_y);
    edge_z.reserve(f_cells_z);

    //initialize the boundary positions
    edge_x.push_back(0);
    edge_y.push_back(0);
    edge_z.push_back(0);
    
    edge_x.push_back(0);
    edge_y.push_back(0);
    edge_z.push_back(0);

	for(int i=2; i<f_cells_x-1; i++) {
		edge_x.push_back(edge_x[i-1]+del_x);		
	}
    for(int i=2; i<f_cells_y-1; i++) {
        edge_y.push_back(edge_y[i-1]+del_y);
    }
    for(int i=2; i<f_cells_z-1; i++) {
        edge_z.push_back(edge_z[i-1]+del_z);
    }

    //initialize the boundary positions
    edge_x.push_back(length_x);
    edge_y.push_back(length_y);
    edge_z.push_back(length_z);
	//------------------------------------//
	//END OF UNIFORM GRID SETUP

	//get cell centroids
	cen_x.push_back(0);
	cen_y.push_back(0);
	cen_z.push_back(0);
	
    for(int i=1; i<f_cells_x-1; i++) {
        cen_x.push_back(0.5*(edge_x[i]+edge_x[i+1]));
    }
    for(int i=1; i<f_cells_y-1; i++) {
		cen_y.push_back(0.5*(edge_y[i]+edge_y[i+1]));
    }
    for(int i=1; i<f_cells_z-1; i++) {
		cen_z.push_back(0.5*(edge_z[i]+edge_z[i+1]));
    }
    
    cen_x.push_back(length_x);
    cen_y.push_back(length_y);
    cen_z.push_back(length_z);

}

/*==========================================================================*/
//set up the cartesian grid
void BTE_structure::set_grid() {

	del_x = length_x/(cells_x-2);
	del_y = length_y/(cells_y-2);
	del_z = length_z/(cells_z-2);

	//UNIFORM GRID SETUP
	//------------------------------------//
    //get the edges of the cell
    edge_x.reserve(cells_x);
    edge_y.reserve(cells_y);
    edge_z.reserve(cells_z);

    //initialize the boundary positions
    edge_x.push_back(0);
    edge_y.push_back(0);
    edge_z.push_back(0);
    
    edge_x.push_back(0);
    edge_y.push_back(0);
    edge_z.push_back(0);

	for(int i=2; i<cells_x-1; i++) {
		edge_x.push_back(edge_x[i-1]+del_x);		
	}
    for(int i=2; i<cells_y-1; i++) {
        edge_y.push_back(edge_y[i-1]+del_y);
    }
    for(int i=2; i<cells_z-1; i++) {
        edge_z.push_back(edge_z[i-1]+del_z);
    }

    //initialize the boundary positions
    edge_x.push_back(length_x);
    edge_y.push_back(length_y);
    edge_z.push_back(length_z);
	//------------------------------------//
	//END OF UNIFORM GRID SETUP

	//get cell centroids
	cen_x.push_back(0);
	cen_y.push_back(0);
	cen_z.push_back(0);
	
    for(int i=1; i<cells_x-1; i++) {
        cen_x.push_back(0.5*(edge_x[i]+edge_x[i+1]));
    }
    for(int i=1; i<cells_y-1; i++) {
		cen_y.push_back(0.5*(edge_y[i]+edge_y[i+1]));
    }
    for(int i=1; i<cells_z-1; i++) {
		cen_z.push_back(0.5*(edge_z[i]+edge_z[i+1]));
    }
    
    cen_x.push_back(length_x);
    cen_y.push_back(length_y);
    cen_z.push_back(length_z);

}

/*==========================================================================*/
//set the angular variables/component values
void BTE_structure::set_angular() {
	
	//get control angle width
	Nmax_angle = N_theta*N_phi*8;
	del_theta = 0.5*PI/N_theta;
	del_phi = 0.5*PI/N_phi;

	cen_theta.reserve(2*N_theta);
	cen_phi.reserve(4*N_phi);

	//find the centroids of all the control angles
	cen_theta.push_back(0.5*del_theta);
	cen_phi.push_back(0.5*del_phi);

	for(int i=1; i<(N_theta*2); i++) {
		cen_theta.push_back(cen_theta[i-1]+del_theta);
	}

	for(int i=1; i<(N_phi*4); i++) {
		cen_phi.push_back(cen_phi[i-1]+del_phi);
	}

	//get all the direction extents and the propagation vectors

	int temp_theta = N_theta*2;// - 1;
	int temp_phi = N_phi*4;//- 1;
	int cur_angle = 0; //the indicator of the current angle

	//make enough room for the vector
	propvec.resize(Nmax_angle);

	for(int i=0; i<temp_theta; i++) {
		for(int j=0; j<temp_phi; j++) {
			//get the current angle
			cur_angle = j + i*4*N_phi;
			
			//get S vector components
			propvec[cur_angle].set_components(cen_phi[j], cen_theta[i], 
														del_phi,del_theta);

		}
	}			

}

/*==========================================================================*/
//set the extent and S vector components for this direction
void Propagation::set_components(double &cen_phi, double &cen_theta, double
&del_phi, double &del_theta) {

    extent_ = 2*sin(cen_theta)*sin(0.5*del_theta)*del_phi;
    S_i_ = sin(cen_phi)*sin(0.5*del_phi)*(del_theta -
        cos(2*cen_theta)*sin(del_theta));
    S_j_ = cos(cen_phi)*sin(0.5*del_phi)*(del_theta -
        cos(2*cen_theta)*sin(del_theta));
    S_k_ = 0.5*del_phi*sin(2*cen_theta)*sin(del_theta);

}

/*==========================================================================*/
//BTE function: reset all coefficients to 0 and initialize the f_angles
void BTE_structure::reset(void) {

    for(int i=0; i<cells_x; i++) {
        for(int j=0; j<cells_y; j++) {
            cell[i][j][0].f_angle.resize(Nmax_angle);
            cell[i][j][cells_z-1].f_angle.resize(Nmax_angle);
                
        }
    }
    
    for(int j=0; j<cells_y; j++) {
        for(int k=0; k<cells_z; k++) {
            cell[0][j][k].f_angle.resize(Nmax_angle);
            cell[cells_x-1][j][k].f_angle.resize(Nmax_angle);
                
        }
    }
    
    for(int i=0; i<cells_x; i++) {
        for(int k=0; k<cells_z; k++) {
            cell[i][0][k].f_angle.resize(Nmax_angle);
            cell[i][cells_y-1][k].f_angle.resize(Nmax_angle);
                
        }
    }
    
    for(int i=1; i<cells_x-1; i++) {
        for(int j=1; j<cells_y-1; j++) {
            for(int k=1; k<cells_z-1; k++) {
                cell[i][j][k].reset_coeffs();
                
                //reserve memory for all f_angle and reset to zero
                cell[i][j][k].f_angle.resize(Nmax_angle);
                
            }
        }
    }
}

/*==========================================================================*/
//Fourier function: reset all coefficients to 0
void Fourier_structure::reset_fourier(void) {

    for(int i=1; i<f_cells_x-1; i++) {
        for(int j=1; j<f_cells_y-1; j++) {
            for(int k=1; k<f_cells_z-1; k++) {
                f_cell[i][j][k].reset_coeffs();
            }
        }
    }
}

/*==========================================================================*/
//default constructor for BTE_structure class
BTE_structure::BTE_structure() {

    cells_x = 0;
    cells_y = 0;
    cells_z = 0;
	material = "";
    diff_x_0 = 0;
    diff_x_end = 0;
    diff_y_0 = 0;
    diff_y_end = 0;
    diff_z_0 = 0;
    diff_z_end = 0;
    spec_x_0 = 0;
    spec_x_end = 0;
    spec_y_0 = 0;
    spec_y_end = 0;
    spec_z_0 = 0;
    spec_z_end = 0;
    tbound_x_0=0;
    tbound_x_end=0;
    tbound_y_0=0;
    tbound_y_end=0;
    tbound_z_0=0;
    tbound_z_end=0;
    tbx_0_val = 1;
    tbx_end_val = 1;
    tby_0_val = 1;
    tby_end_val = 1;
    tbz_0_val = 1;
    tbz_end_val = 1;
    x_coor=0;
    y_coor=0;
    z_coor=0;
}

/*==========================================================================*/
//define the source terms in the desired areas
void BTE_structure::define_sources(vector<Source> &sources) {
	
	int size = sources.size();

	//go over all sources	
	for(int l=0; l<size; l++) {
	
		//set the volumetric value for each of them
		sources[l].get_volumetric_val();
		
		//go over all cells and set the corresponding heat sources
		for(int i=1; i<cells_x-1; i++) {

			//(to save time check if it is in the boundaries right away)
			//if the cell centroid's 'x' value is within the heat source limits
			if((cen_x[i]>sources[l].l1x)&&(cen_x[i]<sources[l].l2x)) {

				//go over the cells in y
				for(int j=1; j<cells_y-1; j++) {
					//if the cell centroid's 'y' value is within source limits
	        	    if((cen_y[j]>sources[l].l1y)&&(cen_y[j]<sources[l].l2y)) {

						//go over the cells in z
						for(int k=1; k<cells_z-1; k++) {
				
							//set the heat source value of the cell
							//if the cell centroid is within source limits
							if((cen_z[k]>sources[l].l1z)&&
												(cen_z[k]<sources[l].l2z)) {

								cell[i][j][k].q_vol =
											sources[l].q_vol*del_x*del_y*del_z;
                                //set the source on and delay times
                                cell[i][j][k].on_time = sources[l].on_time;
                                cell[i][j][k].offset_time = sources[l].offset_time;
                                cell[i][j][k].q_vol_on = cell[i][j][k].q_vol;							
							}
						}
					}
				}
			}
		}
	}
}

/*==========================================================================*/
//define the source terms in the desired areas for fourier cells
void Fourier_structure::define_sources_f(vector<Source> & sources) {

    int size = sources.size();

    //go over all sources   
    for(int l=0; l<size; l++) {

        //set the volumetric value for each of them
        sources[l].get_volumetric_val();

        //go over all cells and set the corresponding heat sources
        for(int i=1; i<f_cells_x-1; i++) {

            //(to save time check if it is in the boundaries right away)
            //if the cell centroid's 'x' value is within the heat source limits
            if((cen_x[i]>sources[l].l1x)&&(cen_x[i]<sources[l].l2x)) {

                //go over the cells in y
                for(int j=1; j<f_cells_y-1; j++) {
                    //if the cell centroid's 'y' value is within source limits
                    if((cen_y[j]>sources[l].l1y)&&(cen_y[j]<sources[l].l2y)) {

                        //go over the cells in z
                        for(int k=1; k<f_cells_z-1; k++) {

                            //set the heat source value of the cell
                            //if the cell centroid is within source limits
                            if((cen_z[k]>sources[l].l1z)&&
                                                (cen_z[k]<sources[l].l2z)) {

                                f_cell[i][j][k].q_vol =
											sources[l].q_vol*del_x*del_y*del_z;
                                //set the source on and delay times
                                f_cell[i][j][k].on_time = sources[l].on_time;
                                f_cell[i][j][k].offset_time = sources[l].offset_time;
                                f_cell[i][j][k].q_vol_on = f_cell[i][j][k].q_vol;							

                            }
                        }
                    }
                }
            }
        }
    }
}

/*==========================================================================*/
//get the total input heat in the whole system
void Device_structure::get_total_input_heat()
{

    //go over all cells
    for(int i=1;i<substrate.f_cells_x-1;i++)
    {
        for(int j=1;j<substrate.f_cells_y-1;j++)
        {
            for(int k=1;k<substrate.f_cells_z-1;k++)
            {
                //add the heat source of the cell to the total input heat
                qin = qin + substrate.f_cell[i][j][k].q_vol;
            }
        }
    }
    cout<<"Total input power = "<<qin<<endl;
}

/*==========================================================================*/
//set the volumetric heat generation value
void Source::get_volumetric_val() {
	
	//divide the Q term by the volume of the heat source
	q_vol = Q_source/( (l2x-l1x)*(l2y-l1y)*(l2z-l1z) );

}

/*==========================================================================*/
//set the transient values to 0 initially
Source::Source() {
	on_time=0;
	offset_time=0;
}

/*==========================================================================*/
//initialize the boundary values
Bound::Bound() {
	dir=6;
	x=y=z=0;
}

/*==========================================================================*/
bool BTE_structure::isInside(double x_min, double x_max, double y_min,
                         double y_max, double z_min, double z_max) {

    if( (!(x_coor>x_max)) && (!(x_coor+length_x<x_min)) )
    {
        if( (!(y_coor>y_max)) && (!(y_coor+length_y<y_min)) )
        {
            if( (!(z_coor>z_max)) && (!(z_coor+length_z<z_min)) )
            {
                return 1;
            }
        }
    }

    return 0;
}


/*==========================================================================*/
//BTE_structure class constructor
BTE_structure::BTE_structure(double x,double y,double z,
                  double len_x,double len_y,double len_z,
                  int num_cells_x,int num_cells_y,int num_cells_z,
                  vector<Material_coords>& material_coords,
                  vector<double>& edges_x,vector<double>& edges_y,
                  vector<double>& edges_z,vector<double>& cents_x,
                  vector<double>& cents_y,vector<double>& cents_z,
                  int neighbours[6],
                  int spec_x0,int spec_xend,int spec_y0,int spec_yend,
                  int spec_z0,int spec_zend,int diff_x0,int diff_xend,
                  int diff_y0,int diff_yend,int diff_z0,int diff_zend,
                  int tbound_x0,int tbound_xend,int tbound_y0,int tbound_yend,
                  int tbound_z0,int tbound_zend,double tbx0_val,
                  double tbxend_val,double tby0_val,double tbyend_val,
                  double tbz0_val,double tbzend_val) {

    x_coor = x;
    y_coor = y;
    z_coor = z;
    length_x = len_x;
    length_y = len_y;
    length_z = len_z;
    cells_x = num_cells_x;
    cells_y = num_cells_y;
    cells_z = num_cells_z;

    mat_coords = material_coords;
    edge_x = edges_x;
    edge_y = edges_y;
    edge_z = edges_z;
    cen_x = cents_x;
    cen_y = cents_y;
    cen_z = cents_z;

    for(int i=0;i<6;i++)
    {
        bte_boundaries[i]=neighbours[i];
    }

    spec_x_0 = spec_x0;
    spec_x_end = spec_xend;
    spec_y_0 = spec_y0;
    spec_y_end = spec_yend;
    spec_z_0 = spec_z0;
    spec_z_end = spec_zend;
    diff_x_0 = diff_x0;
    diff_x_end = diff_xend;
    diff_y_0 = diff_y0;
    diff_y_end = diff_yend;
    diff_z_0 = diff_z0;
    diff_z_end = diff_zend;
    tbound_x_0 = tbound_x0;
    tbound_x_end = tbound_xend;
    tbound_y_0 = tbound_y0;
    tbound_y_end = tbound_yend;
    tbound_z_0 = tbound_z0;
    tbound_z_end = tbound_zend;
    tbx_0_val = tbx0_val;
    tbx_end_val = tbxend_val;
    tby_0_val = tby0_val;
    tby_end_val = tbyend_val;
    tbz_0_val = tbz0_val;
    tbz_end_val = tbzend_val;
    

}

/*==========================================================================*/
Block_coords::Block_coords(int x_index,int y_index,int z_index,
                           int num_cells_x,int num_cells_y,int num_cells_z) {

    x = x_index;
    y = y_index;
    z = z_index;
    cells_x = num_cells_x;
    cells_y = num_cells_y;
    cells_z = num_cells_z;
}
