/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** nhar.cc: code for the NHAR model order reduction algorithm
****************************************************************************/

#include <stdlib.h>
#include <acml.h>
#include "sol.h"
#include "util.h"
#include "init.h"
#include "solver.h"

using namespace std;
using namespace isac;

unsigned long int factorial(int number);
unsigned int sqrt(unsigned int n);

//*===========================================================================*/
//* Constructor of nhar class. */
nhar::nhar(RC_network_type & AGC):

N_(AGC.size_)
{

}

//*===========================================================================*/
//* Generate the projection matix Q */
//* the linear system can be described by (G+sC)x(s) = B u(s)
//* u(s) is expanded using Taylor series and the equation becomes
//* (G+sC)x(s) = B(u_0 + u_1 s^1 + ...)
void nhar::nhar_projection(RC_network_type& AGC){

#ifdef VERBOSE
    cout << "Calling nhar_projection()" << endl;
#endif
    
    unsigned int M_ = moments_;
    unsigned int n_order = trunc_order_;

	//Note: in matrices, use format row + column*num_rows, i.e. column-major

    //1,2. N_ is the number of thermal elements (size of the AGC_ structure)
	unsigned int N_ele = N_ ;

  	//---Matrix sizes---
	//um is P_ by n_order
    //e_vector is n_order-1 by 1
    //f_matrix is n_order-1 by n_order-1
    //bm is N_ele by n_order
	//bm_col is N_ele by 1
    //G_matrix_ is N_ele by N_ele
	//C_matrix_ is N_ele by N_ele
	//B_matrix_ is N_ele by P_
	//X_ is N_ele by M_
	//Y_ is n_order-1 by M_
	//GBU_ is N_ele by 1
	//tx is N_ele by 1
	//ty is n_order-1 by 1
	//h_ is j-1 by 1

	//3,4.
	vector<double> f_matrix;
	//fill the matrix with zeros
	unsigned int total_size = (n_order-1)*(n_order-1);
	for(unsigned int i=0;  i<total_size; i++) {
		f_matrix.push_back(0) ;
    }
	//5,6,7. put in the 1's
	for(unsigned int i=1;  i<n_order-1; i++) {
		f_matrix[i + (i-1)*(n_order-1)] = 1 ;
	}
    
	//8. get the first column of the identity matrix
	vector<double> e_vector;
	//fill the vector with zeros
    for(unsigned int i=0;  i<n_order-1; i++) {
    	e_vector.push_back(0) ;
    }
    //put in the 1
    e_vector[0] = 1 ;

	vector<double> bm_;
	bm_ = UM_matrix_;

	//11, 12. do LU factorization and calculate GBU
	//calculate G^-1 * B_0 = GBU
	
	//calculate GBU = G^-1 * B_0
	//get BM(:,1)
    vector<double> bm_col;  //N_ by 1
    for(int i = 0; i < N_ ; i++) {
        bm_col.push_back(bm_[i]);
    }
	vector<double> GBU_; //N_ele by 1
    GBU_.resize(N_ele);

    solver_type solver(AGC);
    double err_thres = 1E-15;
#ifdef DEBUG 
    cout << "Getting G^-1 * b" << endl;
#endif
    solver.ssmoother2(AGC, GBU_, bm_col, err_thres);
#ifdef DEBUG
    cout << "Done" << endl;
#endif

	//13. get the norm of GBU
	double norm = dnrm2(N_ele,&GBU_.front(),1); //dnorm2

	//14,15. make X and Y matrices
	vector<double> X_;
	total_size = N_ele*M_;
	for(unsigned int i=0; i< total_size; i++) {
		X_.push_back(0);
	}
	vector<double> Y_;
    total_size = (n_order-1)*M_;
    for(unsigned int i=0; i< total_size; i++) {
        Y_.push_back(0);
    }

    //Can change the following two operations to use one of LAPACK's functions
    //to multiply GBU by a factor (1/norm)
	//16. X(:,1) = GBU/norm
	for(unsigned int i=0; i<N_ele ; i++) {
		X_[i] = GBU_[i]/norm;
	}
	//17. Y(:,1) = e_1 / norm
	for(unsigned int i=0; i<n_order-1 ; i++) {
        Y_[i] = e_vector[i]/norm;
    }

	//get BM(:,2:end)
	vector<double> bm_sub; //N_ele by n_order-1
    for(unsigned int i=1; i<n_order ; i++) {
	//go over all rows
        for(unsigned int k=0; k<N_ele ; k++) {
            bm_sub.push_back( bm_[k+ N_ele * i] );
		}
    }
    
    vector<double> bmy;
	bmy.resize(N_ele);
    vector<double> cx;
    cx.resize(N_ele);
    vector<double> sum;
    sum.resize(N_ele);
    vector<double> tx; //N_ele by 1
    tx.resize(N_ele);
    vector<double> ty; //n_order-1 by 1
	ty.resize(n_order-1);

	for(unsigned int j=1; j<M_; j++) {
		//get Y(:,j-1)
        vector<double> y_col (Y_.begin()+(j-1)*(n_order-1), 
                              Y_.begin()+(j)*(n_order-1));  //n_order-1 by 1
     
    	//BM(:,2:end)*Y(:,j-1)
        dgemv('N', N_, n_order - 1, 1.0, &bm_sub.front(), N_, &y_col.front(),
            1, 0.0, &bmy.front(), 1);

		//get X(:,j-1)
		vector<double> x_col (X_.begin()+(j-1)*N_ele,
                              X_.begin()+j*N_ele); //N_ele by 1
	
		//-C*X(:,j-1)
		//C is a diagonal matrix, so no need to reconstruct it and then
		//multiply it by x
        
        //Don't forget the minus sign    
        for(int i = 0; i < N_; i++) {
            cx[i] = -AGC.C_[i] * x_col[i];
        }

		//-C * X(:,j-1) + BM(:,2:end) * Y(:,j-1) = SUM
		for(unsigned int i = 0; i <N_ele; i++) {
            sum[i] = cx[i] + bmy[i];
	    }	
		
#ifdef VERBOSE
        cout << "Getting G^-1 * SUM, j = " << j << endl;
#endif
        solver.ssmoother2(AGC, tx, sum, err_thres);
#ifdef VERBOSE
        cout << "Done" << endl;
#endif
        dgemv('N', n_order - 1, n_order - 1, 1.0, &f_matrix.front(),
            n_order - 1, &y_col.front(), 1, 0.0, &ty.front(), 1);

		//orthogonalization
		//get X(:,1:j-1)
		vector<double> x_cols (X_.begin(), X_.begin() + N_ele*j); //N_ele by j-1

		//get Y(:,1:j-1) //n_order-1 by j-1
        vector<double> y_cols (Y_.begin(), 
                               Y_.begin() + (n_order - 1) * j);
		//X(:,1:j-1)' * tx
		vector<double> h_; // j-1 by 1
        h_.resize(j);
		//note the T for transpose
        dgemv('T', N_, j, 1.0, &x_cols.front(), N_, &tx.front(),
            1, 0.0, &h_.front(), 1);

		//tx = tx - X(:,1:j-1) * h
		//note the alpha = -1 and beta = 1
        dgemv('N', N_, j, -1.0, &x_cols.front(), N_, &h_.front(),
            1, 1.0, &tx.front(), 1);

		//ty = ty - Y(:,1:j-1) * h
        //note the alpha = -1 and beta = 1
        dgemv('N', n_order - 1, j, -1.0, &y_cols.front(), n_order - 1,
            &h_.front(), 1, 1.0, &ty.front(), 1);

		//reorthogonalization
		//h= X(:,1:j-1)' * tx
		//note the T for transpose
		dgemv('T', N_, j, 1.0, &x_cols.front(), N_,
		    &tx.front(), 1, 0.0, &h_.front(), 1);
		
		//tx = tx - X(:,1:j-1) * h
        //note the alpha = -1 and beta = 1
        dgemv('N', N_, j, -1.0, &x_cols.front(), N_, &h_.front(),
            1, 1.0, &tx.front(), 1);

        //ty = ty - Y(:,1:j-1) * h
        //note the alpha = -1 and beta = 1
        dgemv('N', n_order - 1, j, -1.0, &y_cols.front(), n_order - 1,
            &h_.front(), 1, 1.0, &ty.front(), 1);

		double norm2 = dnrm2(N_ele,&tx.front(),1);

		//normalization
		//for this column
		for(unsigned int i=0; i< N_ele; i++ ) {
			X_[i + (N_ele)*j] = tx[i]/norm2;
		}
		for(unsigned int i=0; i< n_order-1; i++) {
            Y_[i+ (n_order-1)*j] = ty[i]/norm2;
        }
	}

    //get the projection matrix Q
    Q_matrix_ = X_;

}

//*==========================================================================*/
//Expand the input in an infinite Taylor series around zero frequency, and
//truncate to n terms
void nhar::taylor_expand(double duration, vector<double> & power) {

#ifdef VERBOSE
    cout << "Calling taylor_expand() " << endl;
#endif

    vector<double> coeff;
    //coeff.resize(moments_); trunc_order and not moments_
    coeff.resize(trunc_order_);

    //calculate the moments (coefficients of the taylor series expansion)
    //coeff[i] = (-1)^i * duration^(i+1) / (i+1)!
    for(int i = 0; i < trunc_order_; i++) {

        coeff[i] = pow(duration, i + 1);
        if(i >= 20) {
            coeff[i] = 0;
        }
        else {
            coeff[i] /= double(factorial(i+1));
        }
        if(i % 2 == 1) {
            coeff[i] *= -1;
        }
        //cout << coeff[i] << endl;
    }
    //cout << endl;

    //fill in the UM matrix
    //the UM matrix format is like this:
    //      col0        col1        col2     ...    col(trunc_order_-1)
    //row0  P1*coeff0   P1*coeff1   P1*coeff2       P1*coeff(trunc_order_-1)
    //row1  P2*coeff0   P2*coeff1   P2*coeff2       P2*coeff(trunc_order-1)
    //row2  P3*coeff0   P3*coeff1   P3*coeff2       P3*coeff(trunc_order-1)
    //...
    //rowN  PN*coeff0   PN*coeff1   PN*coeff2       PN*coeff(trunc_order-1)
    //
    //where Pi is the power consumption of thermal element i
   
    //Note: the matrices are stored in a column-major form
    UM_matrix_.resize(N_ * trunc_order_);
    //for(int i = 0; i < moments_; i++) {
    for(int i = 0; i < trunc_order_; i++) {
        for(int j = 0; j < N_; j++) {
            
            UM_matrix_[j + i * N_] = power[j] * coeff[i];
        }
    }
}

//*==========================================================================*/
// Project the system using the projection matrix Q
void nhar::project_system(vector<double> & power_, vector<double> & init_temp,
        RC_network_type & AGC) {

#ifdef VERBOSE
    cout << "Calling project_system()" << endl;
#endif


    //Note: Q is (N * n) where
    //N : number of thermal elements = N_
    //n : number of moments = moments_

    //project the capacitance matrix C
    //first get Q' * C = temp (n * N)
    vector<double> temp;
    temp.resize(moments_ * N_);

    //Q'*C is the same as Q' but with column1 multiplied by C[0], column2
    //multiplied by C[1], ....
    for(int i = 0; i < N_; i++) {
        for(int j = 0; j < moments_; j++) {
            temp[i*moments_ + j] = Q_matrix_[j*N_ + i] * AGC.C_[i];
        }
    }
    
    //second get C~ = temp * Q
    reduced_C_.resize(moments_ * moments_);
    dgemm('N', 'N', moments_, moments_, N_, 1.0, &temp.front(), moments_,
        &Q_matrix_.front(), N_, 0.0, &reduced_C_.front(), moments_);

    //Check if reduced C matrix contains any NaNs
    for(int i = 0; i < moments_; i++) {
        for(int j = 0; j < moments_; j++) {
            if(std::isnan(reduced_C_[i*moments_ + j])) {
                cout << "Error in computing the reduced C matrix";
                cout << "(contains NaNs)" << endl;
                exit(1);
            }
        }
    }


    //project the conductance matrix G
    //first get Q' * G = temp (n * N)

    //clear temp and fill it with zeros
    temp.clear();
    temp.resize(N_ * moments_);

    //get G*Q first = temp (N * n)  (easier than getting Q'*G first)
    for(int i = 0; i < N_; i++) {
        for(int j = 0; j < moments_; j++) {
            vector<int> & at = AGC.A_[i];
            vector<double> & gt = AGC.G_[i];
            for(int k = 0; k < int(at.size()); k++) {
                int at_index = at[k];
                if(at_index >= 0) {
                    temp[j*N_ + i] += gt[k] * Q_matrix_[j*N_ + at_index];
                }
            }
        }
    }
    
    //Second get G~ = Q' * temp
    reduced_G_.resize(moments_ * moments_);
    dgemm('T', 'N', moments_, moments_, N_, 1.0, &Q_matrix_.front(), N_,
        &temp.front(), N_, 0.0, &reduced_G_.front(), moments_ );


    //Project the initial temperature vector
    //x~(s) = Q'x(s)
    init_reduced_temp_.resize(moments_);
    dgemv('T', N_, moments_, 1.0, &Q_matrix_.front(), N_, &init_temp.front(),
            1, 0.0, &init_reduced_temp_.front(), 1);

    //project the input power matrix UM (B)
    //get B~ = Q' * UM
    reduced_B_.resize(moments_);
    dgemv('T', N_, moments_, 1, &Q_matrix_.front(), N_, &power_.front(),
        1, 0, &reduced_B_.front(), 1);

}

//*==========================================================================*/
// Solve the system for x~(s)
void nhar::solve(vector<vector<double> > & all_temp,
    vector<double> & power_, double duration, double tempreport_interval) {

    //Assume the input power durations are the same as the required
    //duration of simulation. Because of that, Bu~(n+1) would be constant.
    //Otherwise, need to check if current simulation time time has
    //exceeded the duration of the input power
#ifdef VERBOSE
    cout << "Solving system" << endl;
#endif

    //double h = 1e-6; //step size
    double h = tempreport_interval; //step size

    //Need to solve [G~ + C~/h] x~(n+1) = Bu~(n+1) + C~/h x~(n)

    //First, get G~ + C~/h 
    vector<double> temp;
    //temp.resize(moments_ * moments_);
    temp = reduced_G_;
    for(int i = 0; i < moments_ * moments_; i++) {
        temp[i] += reduced_C_[i] / h;
    }

    vector<int> perm_array;
    perm_array.resize(moments_);
    int info;
    //Second, do LU factorization 
    dgetrf(moments_, moments_, &temp.front(), moments_, &perm_array.front(),
        &info);

    if(info != 0) {
        cout << "Error in LU factorization" << endl;
        exit(1);
    }

    vector<double> x_n;   // x(n)
    vector<double> x_np1; // x(n+1)

    x_n.resize(moments_);
    //Initialize the temperatures with those of the previous step
    x_np1 = init_reduced_temp_;

    double simulation_time = 0;

    while(simulation_time < duration) {

        simulation_time += h;
#ifdef DEBUG
        cout << "Current time = " << simulation_time << endl;
#endif

        //Compute the right hand side of the equation
        //i.e. get Bu(n) + C~/h * x~(n) = reduced_B + 1/h*reduced_C*x_n
        
        x_n = x_np1;
        
        x_np1 = reduced_B_; //will be turned into the rhs
        dgemv('N', moments_, moments_, 1.0/h, &reduced_C_.front(), moments_,
            &x_n.front(), 1, 1.0, &x_np1.front(), 1);

        //solve the system
        dgetrs('N', moments_, 1, &temp.front(), moments_, &perm_array.front(),
            &x_np1.front(), moments_, &info);

        if(info != 0) {
            cout << "Error in solving system" << endl;
            exit(1);
        }
    }
    
    //rhs contains the solution
    reduced_temp_ = x_np1;

}


//*==========================================================================*/
// Compute the temperatures by obtaining x(s) from x~(s)
void nhar::temp_compute(vector<double> & temp) {

    //x(s) = Q * x~(s)

#ifdef VERBOSE
    cout << "Computing the final temperatures" << endl;
#endif
  
    vector<double> temporary;
    temporary.resize(N_); 
    dgemv('N', N_, moments_, 1.0, &Q_matrix_.front(), N_,
        &reduced_temp_.front(), 1, 0.0, &temporary.front(), 1);

    temp = temporary;
}

//*==========================================================================*/
unsigned long int factorial(int number) {
    unsigned long int temp;

    if(number <= 1) return 1;
    if(number > 20) return 100000000;

    temp = number * factorial(number - 1);
    return temp;
}

//*==========================================================================*/
unsigned int sqrt(unsigned int n) {
	unsigned int a;
	for (a=0;n>=(2*a)+1;n-=(2*a++)+1);
	return a;
}

