/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** sol.cc: The ISAC solution class
****************************************************************************/

#include <stdlib.h>
#include <acml.h>
#include "sol.h"
#include "util.h"
#include "init.h"
#include "solver.h"
#include "configuration.h"
#include "isac.h"

using namespace std;
using namespace isac;


//*==========================================================================*/
//* Constructor for Solution class */
Solution::Solution():
root_(),
hoct_contour_(),
AGC_(),
power_pos_(),
temp_(),
power_(),
N_(),
elts_incomp_(),
init_level_(),
isot_(),
cont_(),
chip_(),
player_btmnei_(),
hcoeff_conv_(),
btm_elts_(0),
temp_conv_(0.0),
AGC_expanded_(false)
{

}


//*==========================================================================*/
//* Constructor for Solution class. */
Solution::Solution(const ChipPkg_type & chip_3D):
temp_(),
power_(),
btm_elts_(0),
temp_conv_(0.0),
AGC_expanded_(false)
{
	chip_ = (ChipPkg_type *)(&chip_3D);

    /* Generate the basic tree structure of the chip 
     * based on information from "chip_3D". */
    isotcont_init(isot_, cont_);
    gene_solu();
    gene_neis();

}


//*==========================================================================*/
//* Form the related objects inside Solution class. */
void Solution::form_solution(
    const double ambient_resis, const double ambient_heatcap){
    
    //* Generate the hoct_contour. */
    hoct_contour_init(hoct_contour_, &root_, &isot_, &cont_);    
    init_level_ = hoct_contour_.size() - 1;
    N_ = hoct_contour_[init_level_].size();
	contour_nei_init(&root_, hoct_contour_, init_level_);
    //* Generate the AGC data strucutre of the deepest level elements in 
    //* hoct_contour. */
    hoct_contour_AGC(hoct_contour_[init_level_], AGC_, init_level_);

    //refine_zpar();
    //* Initiate power and temperature vectors. */
    init_power_temp();
    //* Find the coefficients for power mapping purpose. */
    find_power_rela();

    //Partition the chip as specified in the input chip-package information
    ChipPkg_type & chip_3D = *chip_;
    int comp_num = static_cast<int>(chip_3D.components.size());
    for(int i = 0; i < comp_num; ++i) {
        Component_type & cur_comp = chip_3D.components[i];
        int pow = 0;
        int layer_nums = static_cast<int>(cur_comp.layers.size());
        for (int j = 0; j < layer_nums; ++j) {
            Layer_type & cur_layer = cur_comp.layers[j];
            if(cur_layer.property == POWER_) {
                pow = 1;
                continue;
            }
            else {
                refine_par(i, j - pow, cur_layer.par[X_], cur_layer.par[Y_], 
                    cur_layer.par[Z_]);
            }
        }
    }
    spatial_update();
#ifdef VERBOSE
    cout << "Total number of elements = " << N_ << endl;
#endif
}


//*===========================================================================*/
//* Steady-state thermal simulation.
//* Calculate the thermal profile when time goes to infinity. */
vector<Element_type> Solution::static_solver(
    const vector<vector<double> > & power, double err_thres,
    const double ambient_resis, const double ambient_heatcap) {

    int cont_level = hoct_contour_.size() - 1;

    AGC_expanded_ = false;

    if (!AGC_expanded_){
        find_power_rela();
        broaden_AGC(ambient_resis, ambient_heatcap);
        cal_power_value(cont_level, power); 
        broaden_temp_power(temp_conv_);
        AGC_expanded_ = true;
    }

    //static_temp_compute(err_thres);
    spatial_multigrid_solver(power, ambient_resis, ambient_heatcap);
    
    vector<Element_type> thermal_profile;
    pargrid_to_element(thermal_profile, 
        hoct_contour_[hoct_contour_.size() - 1], temp_);

    return thermal_profile;
}

//*===========================================================================*/
//* Steady-state thermal simulation.
//* Calculate the thermal profile when time goes to infinity. */
vector<Element_type> Solution::static_solver_adaptive(
    const vector<vector<double> > & power, double err_thres,
    const double ambient_resis, const double ambient_heatcap) {

    int cont_level = hoct_contour_.size() - 1;

    AGC_expanded_ = false;

    if (!AGC_expanded_){
        find_power_rela();
        broaden_AGC(ambient_resis, ambient_heatcap);
        cal_power_value(cont_level, power); 
        broaden_temp_power(temp_conv_);
        AGC_expanded_ = true;
    }

    //static_temp_compute(err_thres);
    spatial_multigrid_solver(power, ambient_resis, ambient_heatcap);

    vector<Element_type> thermal_profile;
    pargrid_to_element(thermal_profile, 
        hoct_contour_[hoct_contour_.size() - 1], temp_);

    return thermal_profile;
}

//* Spatial refinement and static thermal simulation.
// * Do spatial refinement and steady-state thermal simulation until the
// * thermal gradient meets the predefined threshold. */
/*============================================================================*/
void Solution::spatial_multigrid_solver(
    const vector<vector<double> > & power, 
    const double ambient_resis, const double ambient_heatcap) {

    int global_flag = 0;
   	double err_thres = 1E-4;

    static_temp_compute(err_thres);

    double max_t = 0.0;
    double min_t = FMAX_;
    for (int i = 0; i < N_; i++){
        max_t = max(max_t, temp_[i]);
        min_t = min(min_t, temp_[i]);
    }
    float thermal_thre = (max_t - min_t)/20.0;

    global_flag = spatial_refinement(thermal_thre);
    int cont_level = hoct_contour_.size() - 1;
    find_power_rela();
    broaden_AGC(ambient_resis, ambient_heatcap);
    cal_power_value(cont_level, power);
    broaden_temp_power(temp_conv_);

    while(global_flag) {
        static_temp_compute(err_thres);
        //find_player_btmnei();
        global_flag = spatial_refinement(thermal_thre);
    	cont_level = hoct_contour_.size() - 1;
        find_power_rela();
        broaden_AGC(ambient_resis, ambient_heatcap);
        cal_power_value(cont_level, power);
        broaden_temp_power(temp_conv_);
    }
    static_temp_compute(err_thres);
}


//*==========================================================================*/
//* Generate the Solution object and call iterative smoother function
// * call to calculate the steady state thermal profile. */
void Solution::static_temp_compute(double err_thres) {

	solver_type solver(AGC_);
	solver.ssmoother(AGC_, temp_, power_, err_thres);

    //* Store the average temperature value of the convection elements. */
    for (int i = 0; i < btm_elts_; ++i){
        temp_conv_ += temp_[i]; 
    }temp_conv_ /= btm_elts_;

	temperature_back_tree();
}


//*==========================================================================*/
//* Dynamic thermal simulation function. */
vector<Element_type> Solution::dynamic_solver(vector<vector<double> > 
	& all_temp, double duration, double tempreport_interval, 
    const vector<vector<double> > & power, char select_domain)  {

    if (!AGC_expanded_){
        cout <<"Init_dynamic_function is not called "<<endl;
        exit(1);
    }
    
    int cont_level = hoct_contour_.size() - 1;
    cal_power_value(cont_level, power);
    
    switch(select_domain){
        case 't':
            time_domain_solver(duration);
            break;
        case 'f':
            frequency_domain_solver(all_temp, duration, tempreport_interval);
            break;
        default:
            cerr << "Should not have this option, must choose time domain "
                << " solver or frequency domain solver " << endl;
            exit(1);
            break;
    } 

    vector<Element_type> thermal_profile;
    pargrid_to_element(thermal_profile, 
        hoct_contour_[hoct_contour_.size() - 1], temp_);
	
    return thermal_profile;
}

//*===========================================================================*/
//* Frequency domain dynamic thermal analysis based on the NHAR method. */
void Solution::frequency_domain_solver(vector<vector<double> > & all_temp,
    double duration, double tempreport_interval){

#ifdef VERBOSE
    cout << "Calling the NHAR solver " << endl;
#endif

    bool adaptive = configuration::ap().nhar_adap();

    if(!adaptive) {
        nhar_mm(all_temp, duration, tempreport_interval);
    }
    else {
        nhar_mm_adap(all_temp, duration, tempreport_interval);
    }
}

//*===========================================================================*/
//* The NHAR moment-mathcing thermal analysis method. */
void Solution::nhar_mm(vector<vector<double> > & all_temp,
    double duration, double tempreport_interval){

    nhar cur_nhar(AGC_);
    cur_nhar.moments() = configuration::ap().moments();
    cur_nhar.trunc_order() = 20;

#ifdef VERBOSE
    cout << "with " << cur_nhar.moments() << " moments. " << endl;
#endif

    //Expand the input in an infinite Taylor series around zero frequency, and
    //truncate to n terms
    cur_nhar.taylor_expand(duration, power_);
   
    //Generate the projection matrix Q 
    cur_nhar.nhar_projection(AGC_);

    //Project the matrices using Q
    cur_nhar.project_system(power_, temp_, AGC_);
    
    //Solve the system for x~(s)
    cur_nhar.solve(all_temp, power_, duration, tempreport_interval);
    
    //Obtain x(s) from x~(s) which is the vector of temeperatures
    cur_nhar.temp_compute(temp_);
    
}

//*===========================================================================*/
//* The NHAR moment-matching thermal analysis method with adaptive moment
//determination. */
void Solution::nhar_mm_adap(vector<vector<double> > & all_temp,
    double duration, double tempreport_interval){

    nhar cur_nhar(AGC_);
//start with a small number of moments
    static int num_moments = configuration::ap().moments();
    cur_nhar.moments() = num_moments;
#ifdef VERBOSE
    cout << "with " << cur_nhar.moments() << " moments. " << endl;
#endif

    double error_thres = 1.0E-3;

    //Begin executing algorithm 2

    vector<double> temp1, temp2;

    //Expand the input in an infinite Taylor series around zero
    //frequency, and truncate to n terms
    cur_nhar.taylor_expand(duration, power_);
  
    //Generate the projection matrix Q by employing algorithm 1
    cur_nhar.nhar_projection(AGC_);

    //Project the matrices using Q
    cur_nhar.project_system(power_, temp_, AGC_);
    
    //Solve the system for x~(s)
    cur_nhar.solve(all_temp, power_, duration, tempreport_interval);
    
    //Obtain x(s) from x~(s) which is the vector of temeperatures
    cur_nhar.temp_compute(temp1);

    bool converged = 0;   

    while(!converged) {

#ifdef VERBOSE
        cout << endl << "Calling the NHAR solver " << endl;
#endif
        cur_nhar.moments() = cur_nhar.moments() + 1;

#ifdef VERBOSE
        cout << "with " << cur_nhar.moments() << " moments. " << endl;
#endif

        //Expand the input in an infinite Taylor series around zero
        //frequency, and truncate to n terms
        cur_nhar.taylor_expand(duration, power_);
   
        //Generate the projection matrix Q by employing algorithm 1
        cur_nhar.nhar_projection(AGC_);

        //Project the matrices using Q
        cur_nhar.project_system(power_, temp_, AGC_);
    
        //Solve the system for x~(s)
        cur_nhar.solve(all_temp, power_, duration, tempreport_interval);
    
        //Obtain x(s) from x~(s) which is the vector of temeperatures
        cur_nhar.temp_compute(temp2);

        converged = compute_error(temp1, temp2, error_thres);

        temp1 = temp2;

    }

    temp_ = temp2;
    num_moments = cur_nhar.moments() - 1;
    
}

//*===========================================================================*/
//* Time domain dynamic thermal analysis. */
void Solution::time_domain_solver(double duration){

    //double interval = 1e-6;
    //double interval = 1e-12;
    double interval = 1e-11;
    double err_thres = 1e-06;
    int sop = 2;

    vector<double> temp_out;
    double f_intval;

    dysol cur_dysol;
    int c_size = AGC_.C_.size();
    cur_dysol.cap_.resize(c_size/*N_*/);
    for (int i = 0; i < c_size/*N_*/; i++){
        cur_dysol.cap_[i] = AGC_.C_[i];
    }
    
    switch(sop){
        case 0:
            cur_dysol.rk4_glb_solver(AGC_, temp_, power_, 
                interval, err_thres, duration, f_intval);
            break;
        case 2:
            cur_dysol.semi_th_implct_solver(AGC_, temp_, old_temp, power_,
                interval, err_thres, duration);
            break;
        default:
            break;
    }
}

//*=========================================================================*/
//* Generate the object of Solution class. 
// * Setup the first level element of the hoct_contour data structure  
// * (root_). */
void Solution::gene_solu(){

    const int chip_level = 0;
    const int chip_id = 0;

	root_.type_assign(CHIP_3D_);
	root_.levelid_assign(chip_level, chip_id);
	root_.id_assign(chip_id);
	root_.act_level_assign(chip_level);
	root_.self_level_assign(chip_level);
	gene_comps();
}


//*=========================================================================*/
// * Setup the second level elements of the hoct_contour data structure  
// * (compnonents). */
void Solution::gene_comps(){

	const int comp_level = 1;
    int layer_id = 0;
    int elem_id = 0;
	int comp_nums = static_cast<int>(chip_->components.size());
	elts_incomp_.resize(comp_nums);
	
	for (int comp_id = 0; comp_id < comp_nums; ++comp_id){
        Component_type & cur_comp = chip_->components[comp_id];
	    par_grid * chi = new par_grid;
		chi -> type_assign(COMPONENT_);
		chi -> par_assign(&root_);
		chi -> levelid_assign(comp_level, comp_id);
		chi -> id_assign(comp_id);
		chi -> act_level_assign(comp_level);
		chi -> self_level_assign(comp_level);
		root_.int_chi_array_assign(chi);

		gene_layers(*chi, cur_comp, layer_id, elem_id);
		elts_incomp_[comp_id] = elem_id;
	}
}


/*============================================================================*/
//* Setup the third level elements of the hoct_contour data structure  
//* (layers). */
void Solution::gene_layers(par_grid & root, Component_type & comp,  
    int & layer_id, int & elem_id) {

	const int layer_level = 2;
	int material_layer_num = 0;
	int layers_num = static_cast<int>(comp.layers.size());

    for (int i = 0; i < layers_num; ++i){
        Layer_type & cur_layer = comp.layers[i];
		if (cur_layer.property == MATERIAL_ || 
            cur_layer.property == INTERFACE_){
			par_grid * chi = new par_grid;
			chi->type_assign(LAYER_);
            chi->cond_assign(cur_layer.cond);
            chi->cond_map_assign(toMap(cur_layer.cond_map));
            chi->ccap_assign(cur_layer.cap);
            chi->ccap_map_assign(toMap(cur_layer.cap_map));
            chi->id_assign(material_layer_num);
            chi->levelid_assign(layer_level, layer_id);
			chi->sizes_assign(cur_layer.size[X_], cur_layer.size[Y_],
				cur_layer.size[Z_]);
			chi -> coors_assign(cur_layer.coor[X_], cur_layer.coor[Y_],
                cur_layer.coor[Z_]);
			chi -> par_assign(&root);
			chi -> act_level_assign(layer_level);
		    chi -> self_level_assign(layer_level);
			root.int_chi_array_assign(chi);	

			//* Create the initial partition(the leaf nodes) based on 
            //* the initial partition stored in chip_ */
			gene_elements(*chi, cur_layer, elem_id);

			material_layer_num ++;
            layer_id ++;
		}
	}
}


/*=========================================================================*/
//* Setup the third level elements of the hoct_contour data structure  
//* (initial partitioned thermal elements). */
//* Generate the finest grid elements for the initial partition. */
void Solution::gene_elements(par_grid & root, Layer_type & layer, 
    int & elem_id){ 

    par_grid * init_par = chip_->init_partition.
        int_chi_array(root.par()->id());

	const int elts_level = 3;
	int pgrid_no = static_cast<int>(init_par->int_chi_array().size());
	for (int i = 0; i < pgrid_no; ++i){
		par_grid * chi = new par_grid;
		par_grid * ref_grid = init_par->int_chi_array(i);
		chi->type_assign(GRID_);
        chi -> cond_assign(layer.cond);
        chi -> bulk_cond_assign(layer.cond);
        chi -> cond_map_assign(toMap(layer.cond_map));
        chi -> ccap_assign(layer.cap);
        chi -> bulk_ccap_assign(layer.cap);
        chi -> ccap_map_assign(toMap(layer.cap_map));
		chi->id_assign(i);
		chi->levelid_assign(elts_level, elem_id);
		chi->sizes_assign(ref_grid -> sizes(X_), ref_grid -> sizes(Y_),
			root.sizes(Z_));
		chi->coors_assign(ref_grid -> coors(X_), ref_grid -> coors(Y_),
			root.coors(Z_));
		chi->par_assign(&root);
		chi->partition_flag_assign(0,0,0);
		chi -> act_level_assign(elts_level);
		chi -> self_level_assign(elts_level);
		root.int_chi_array_assign(chi);
		elem_id ++;
	}	
}


//*=========================================================================*/
//* Generate the neighbor information for each partition level (the whole
//* packge level, the component level, the layer level) of the chip package 
//* based on the init_partition input */
void Solution::gene_neis(){

    par_grid & init_partition = chip_->init_partition;
    gene_comp_neis(init_partition);
    gene_layers_neis(root_, init_partition);
    gene_elts_neis(init_partition);
    
}


//*=========================================================================*/
//* Generate the neighbor relationship for the components. */
void Solution::gene_comp_neis(par_grid & init_partition){

    int tot_comps = root_.int_chi_array().size();
    
    int nei_cnt = 0;
    for (int i = 0; i < tot_comps; ++i){
        par_grid & cur_compgrid = *root_.int_chi_array(i);
        par_grid & ref_compgrid = *init_partition.int_chi_array(i);
        for (int j = 0; j < NEI_C_; ++j){
            nei_cnt = ref_compgrid.nei(j).size();
            for (int k = 0; k < nei_cnt; ++k){
                par_grid & nei_grid = *ref_compgrid.nei(j,k);
                int nei_id = nei_grid.id();
                if((nei_id == CONT_ID_) || (nei_id == ISOT_ID_)) {
                    cur_compgrid.nei_assign(j, &nei_grid);
                } else{
                    cur_compgrid.nei_assign(j, root_. 
                        int_chi_array(nei_id));
                }
            }
        }
    }
}


//*=========================================================================*/
//* Generates the neighbor information for the layers and/or the initial
//* finest elements. */
void Solution::gene_layers_neis(par_grid & root, par_grid & init_partition) {

    int par_num = static_cast<int>(root.int_chi_array().size());
    par_grid * cur_par_node=NULL;
    par_grid * pre_par_node=NULL;
    par_grid * nex_par_node=NULL;

    for (int i = 0; i < par_num; ++i){
        cur_par_node = root.int_chi_array(i);
        if(i > 0) {
            pre_par_node = root.int_chi_array(i-1);
        }
        if(i < par_num - 1) {
            nex_par_node = root.int_chi_array(i+1);
        }
        par_grid * ref_par_node = init_partition.int_chi_array(i);
        int chi_num = static_cast<int> (cur_par_node->int_chi_array().size());
        par_grid * cur_chi_node=NULL;
        par_grid * pre_chi_node=NULL;
        par_grid * nex_chi_node=NULL;

        for (int j = 0; j < chi_num; ++j){

            cur_chi_node = cur_par_node -> int_chi_array(j);
            if(j > 0) {
                pre_chi_node = cur_par_node->int_chi_array(j-1);
            }
            if(j < chi_num -1 ) {
                nex_chi_node = cur_par_node->int_chi_array(j+1);
            }

            //* The following loop deals with the X_NEG_, X_PLU_, Y_NEG_ and 
            // * Y_PLU_ deirection. */
            for(int l = 0; l < (NEI_C_ - 2); ++l) {
                for(int m = 0; m < static_cast<int> 
                        (ref_par_node->nei(l).size()); ++m) {
                    int par_nei_id = ref_par_node->nei(l,m)->id();
                    if((par_nei_id == CONT_ID_) || (par_nei_id == ISOT_ID_)) {
                        cur_chi_node->nei_assign(l, ref_par_node->nei(l,m));
                    }else {
                        cur_chi_node->nei_assign(l, cur_par_node->
                                int_chi_array(par_nei_id));
                    }
                }
            }
            if(j == 0) {
                for(int m = 0; m < static_cast<int> 
                        (ref_par_node->nei(Z_NEG_).size()); ++m) {
                    int par_nei_id = ref_par_node->nei(Z_NEG_,m)->id();
                    if((par_nei_id == CONT_ID_) || (par_nei_id == ISOT_ID_)) {
                        cur_chi_node->nei_assign(Z_NEG_, 
                            ref_par_node->nei(Z_NEG_, m));
                    }else {
                        assert(i != 0);
                        cur_chi_node->nei_assign(Z_NEG_, pre_par_node->
                            int_chi_array(pre_par_node->int_chi_array().size()-1));
                    }
                }
                for(int m = 0; m < static_cast<int> 
                        (ref_par_node->nei(Z_PLU_).size()); ++m) {
                    int par_nei_id = ref_par_node->nei(Z_PLU_,m)->id();
                    if((par_nei_id == CONT_ID_) || (par_nei_id == ISOT_ID_)) {
                        cur_chi_node->nei_assign(Z_PLU_, 
                            ref_par_node->nei(Z_PLU_, m));
                    }else {
                        par_grid * nei_node = NULL;
                        if ( j != (chi_num -1)){
                            nei_node = nex_chi_node;
                        } else{
                            nei_node = nex_par_node -> int_chi_array(0);
                        }
                        cur_chi_node->nei_assign(Z_PLU_, nei_node); 
                    }
                }
            }else if (j == chi_num -1) {//*j won't be equal to 0 here (else if)*/
                for (int m = 0; m < static_cast<int>
                        (ref_par_node->nei(Z_PLU_).size()); ++m) {
                    int par_nei_id = ref_par_node->nei(Z_PLU_,m)->id();
                    if((par_nei_id == CONT_ID_) || (par_nei_id == ISOT_ID_)) {
                        cur_chi_node->nei_assign(Z_PLU_, 
                            ref_par_node->nei(Z_PLU_, m));
                    }else {
                        assert(i != (par_num-1));
                        cur_chi_node->nei_assign(Z_PLU_, nex_par_node->
                            int_chi_array(0));
                    }
                }
                for(int m = 0; m < static_cast<int> 
                        (ref_par_node->nei(Z_NEG_).size()); ++m) {
                    int par_nei_id = ref_par_node->nei(Z_NEG_,m)->id();
                    if((par_nei_id == CONT_ID_) || (par_nei_id == ISOT_ID_)) {
                        cur_chi_node->nei_assign(Z_NEG_, 
                            ref_par_node->nei(Z_NEG_, m));
                    }else {
                        par_grid * nei_node = NULL;
                        //if ( j != 0){
                            nei_node = pre_chi_node;
                        /*} else{
                            nei_node = pre_par_node -> int_chi_array
                                (pre_par_node->int_chi_array().size() - 1);

                        }*/
                        cur_chi_node->nei_assign(Z_NEG_, nei_node); 
                    }
                }
            }else {
                cur_chi_node->nei_assign(Z_NEG_, pre_chi_node);
                cur_chi_node->nei_assign(Z_PLU_, nex_chi_node);
            }
        }
    }
}


//*=========================================================================*/
//* Generate the neighbor information for the elements in the chip package 
//* based on the init_partition input. */
void Solution::gene_elts_neis(par_grid & init_partition){

    int comp_nums = static_cast<int>(init_partition.int_chi_array().size());
    par_grid * cur_compgrid = NULL;
    par_grid * pre_compgrid = NULL;
    par_grid * nex_compgrid = NULL;

    par_grid & root = root_;
    
    for (int i = 0; i < comp_nums; ++i){
        cur_compgrid = root.int_chi_array(i);
        par_grid * ref_compgrid = init_partition.int_chi_array(i);
        if(i > 0) {
            pre_compgrid = root.int_chi_array(i-1);
        }
        if(i < comp_nums - 1) {
            nex_compgrid = root.int_chi_array(i+1);
        }
        int layer_nums = static_cast<int>
            (root.int_chi_array(i)->int_chi_array().size());
        for (int j = 0; j < layer_nums; ++j){
            par_grid * cur_layergrid = NULL;
            par_grid * pre_layergrid = NULL;
            par_grid * nex_layergrid = NULL;
            cur_layergrid = cur_compgrid -> int_chi_array(j);

            if(j > 0) {
                pre_layergrid = cur_compgrid->int_chi_array(j-1);
            }
            if(j < layer_nums -1 ) {
                nex_layergrid = cur_compgrid ->int_chi_array(j+1);
            }

            for (int k = 0; k < static_cast<int>
                (cur_layergrid->int_chi_array().size()); ++k){
                par_grid * cur_grid = cur_layergrid -> int_chi_array(k);
                par_grid * ref_grid = ref_compgrid -> int_chi_array(k);
                //* Find the neighbors in the X_NEI_, X_PLU_, Y_NEG_, Y_PLU_
                //* directions first. */
                for (int l = 0; l < (NEI_C_ - 2); ++l){
                    for (int m = 0; m < static_cast<int>
                        (ref_grid -> nei(l).size()); ++m){
                        int nei_id = ref_grid -> nei(l, m) -> id();
                        assert(nei_id >= -2);
                        if ((nei_id ==CONT_ID_ ) || ( nei_id == ISOT_ID_)){
                            cur_grid -> nei_assign(l, ref_grid -> nei(l, m));
                        } else if (nei_id >= 0){
                            cur_grid -> nei_assign(l, cur_layergrid -> int_chi_array(nei_id));
                        }
                    }
                }
                //* Take care of the neighbors in the Z_NEG_/Z_PLU_ direction when the elements 
                //* are in the layer on the top of a component. */
                if (j == 0) {
 					//* Deal with the Z_NEG_ when (j == 0). */
                    for (int m = 0; m < static_cast<int>
                        (ref_grid -> nei(Z_NEG_).size()); ++m){
                        int nei_id = ref_grid -> nei(Z_NEG_,m) -> id();
                        assert(nei_id >= -2);
                        if ((nei_id == ISOT_ID_) || (nei_id == CONT_ID_)){
                            cur_grid -> nei_assign(Z_NEG_, ref_grid -> nei(Z_NEG_, m));
                        }   
                        else{
                            assert(i!=0);// i won't be zero under this circumstance.
                            int nei_layer_nums = pre_compgrid->int_chi_array().size();
                            par_grid * nei_grid = pre_compgrid ->
                                int_chi_array(nei_layer_nums - 1) -> int_chi_array(nei_id);
                            cur_grid -> nei_assign(Z_NEG_, nei_grid);
                        }
                    }
 					//* Deal with the Z_PLU_ when (j == 0). */
					if ( j == (layer_nums - 1)){
						for (int m = 0; m < int(ref_grid->nei(Z_PLU_).size()); ++m){
                        	int nei_id = ref_grid->nei(Z_PLU_,m)->id();
                            if((nei_id == CONT_ID_) || (nei_id == ISOT_ID_)) {
                                cur_grid->nei_assign(Z_PLU_, 
                                    ref_grid->nei(Z_PLU_, m));
                            } else{//* cur_layer won't be the last layer under this case
								   //* since the neighbors on Z_PLU_ are not isothermal
                                   //* grid or constant temperature grid. */
                                par_grid * nei_grid = nex_compgrid -> int_chi_array(0)->
                                    int_chi_array(nei_id);
                                cur_grid->nei_assign(Z_PLU_, nei_grid);
                            }
							
						}
					}
					else{
                       	par_grid * nei_grid = nex_layergrid -> int_chi_array(cur_grid->id());
                       	cur_grid->nei_assign(Z_PLU_, nei_grid);
					}
                }
                //* Take care of the neighbors in the Z_NEG_/Z_PLU_ direction when the elements
                // * are in the layer on the bottom of a component. */
                else if ( j == (layer_nums - 1)){//* j won't be 0 here since. */
                    par_grid * nei_grid = pre_layergrid -> int_chi_array
						(cur_grid->id());;
                    cur_grid->nei_assign(Z_NEG_, nei_grid); 
                    for (int m = 0; m < static_cast<int>
                        (ref_grid -> nei(Z_PLU_).size()); ++m){
                        int nei_id = ref_grid -> nei(Z_PLU_,m) -> id();
                        assert(nei_id >= -2);
                        if ((nei_id == ISOT_ID_) || (nei_id == CONT_ID_)){
                            cur_grid -> nei_assign(Z_PLU_, ref_grid -> nei(Z_PLU_, m));
                        }
                        else{
                            assert(i!= comp_nums - 1);//*won't be the last component.*/
                            par_grid * nei_grid = nex_compgrid ->
                                int_chi_array(0) -> int_chi_array(nei_id);
                            cur_grid -> nei_assign(Z_PLU_, nei_grid);
                        }
                    }
                }
                else {
                    //* Generate the Z_NEG_, and Z_PLU_ neighbors when the elements
                    //* are in a layers that is not top or bottom of a component. */
                    for (int m = 0; m < static_cast<int>
                        (ref_grid -> nei(Z_NEG_).size()); ++m){
                        int nei_id = cur_grid -> id();
                        par_grid * pre_nei_grid = pre_layergrid 
                            -> int_chi_array(nei_id);
                        par_grid * nex_nei_grid = nex_layergrid ->
                            int_chi_array(nei_id);
                        cur_grid -> nei_assign(Z_NEG_, pre_nei_grid);
                        cur_grid -> nei_assign(Z_PLU_, nex_nei_grid);
                    }
                }
            }
        }
    }
}


//*===========================================================================*/
//* Find the elements that have upper/lower neighbor of power
//* for 3D thermal elements. */ 
void Solution::find_powernei(int level, char direction){

	/* hoct_contour does not contain the 2D power grids */
	int power_nums = static_cast<int>(power_pos_.size());
    ChipPkg_type & chip_3D = *chip_;
    int comp_nums = static_cast<int>(chip_3D.components.size());

    int cur_power_seq = 0;
    for (int i = 0; i < comp_nums; ++i){
        Component_type & cur_comp = chip_3D.components[i];
        int layer_nums = static_cast<int>(cur_comp.layers.size());
		int mat_num = 0;
        for (int j = 0; j < layer_nums; ++j){
            Layer_type & cur_layer = cur_comp.layers[j];
            if (cur_layer.property == MATERIAL_ ||
                cur_layer.property == INTERFACE_){
                for (int k = cur_power_seq; k < power_nums; ++k){
                    //double zpos = power_pos_[cur_power_seq].second
                    //    ->int_chi_array(0) -> coors(Z_);
                    double zpos = power_pos_[k].second
                        ->int_chi_array(0) -> coors(Z_);
                    if ((direction == 'u')
                        && ((fabs(cur_layer.coor[Z_] - cur_layer.size[Z_]/2.0 
                        - zpos)) < FMIN_)){
						//* Power layer is at the top of the current element
						//* layer (attached to each other). */
                        vector<par_grid *> & cur_power_layer = power_pos_[k].
                            second -> int_chi_array();
                        par_grid * cur_grid = root_.int_chi_array(i)
                            -> int_chi_array(mat_num);
                        check_nei(cur_grid, cur_power_layer, Z_NEG_);
                        cur_power_seq++;
                        continue;
                    } else if ((direction == 'd') &&
                    (fabs(cur_layer.coor[Z_] + cur_layer.size[Z_]/2.0 
                        - zpos) < FMIN_)){
						//* Power layer is at the bottom of the current element
						//* layer (attached to each other). */
                        vector<par_grid *> & cur_power_layer = power_pos_[k].
                            second -> int_chi_array();
                        par_grid * cur_grid = root_.int_chi_array(i)
                            -> int_chi_array(mat_num);
                        check_nei(cur_grid, cur_power_layer, Z_PLU_); 
                        cur_power_seq++;
                        continue;
                    }
                } 
				mat_num ++;
            }
        }
    }
}

//*===========================================================================*/
//* Parition all the elements inside a box homogeneously and map the power
//profile to the new grid
void Solution::zoom_box(double x_min, double x_max, double y_min,
    double y_max, double plyr_z, const vector<vector<double> > & power) {

    //Find out all the elements inside the box. Do that by finding if the
    //element's center is inside the box

    //First, find the active layer z-coordinate. The active layer is the one
    //adjacent to the power layer whose index is given by actv_lyr_ind
    double layer_zcoor = -1;
    double layer_zsize = 0;
    char dir = '\0';

    //Loop over all components and layers, and find the one that is adjacent
    //to the power layer given by plyr_z
    for(unsigned int i = 0; i < chip_->components.size(); i++) {
        for(unsigned int j = 0; j < chip_->components[i].layers.size(); j++) {
            //Make sure it is a material layer, not an interface layer or a
            //power layer
            if(chip_->components[i].layers[j].property==MATERIAL_) {
            //check if this layer is on top or on the bottom of the power layer
                if(fabs(chip_->components[i].layers[j].coor[Z_] -
                chip_->components[i].layers[j].size[Z_]/2.0 - plyr_z) 
                < FMIN_) {
                    
                    layer_zcoor = chip_->components[i].layers[j].coor[Z_];
                    layer_zsize = chip_->components[i].layers[j].size[Z_];
                    dir = 'u'; //up
                }
                else if(fabs(chip_->components[i].layers[j].coor[Z_] +
                chip_->components[i].layers[j].size[Z_]/2.0 - plyr_z)
                < FMIN_) {
                
                    layer_zcoor = chip_->components[i].layers[j].coor[Z_];
                    layer_zsize = chip_->components[i].layers[j].size[Z_];
                    dir = 'd'; //down
                }
            }
        }
    }

    //Make sure a layer was found
    if(layer_zcoor < 0) {
        cout << "No layer with the given z is found." << endl;
        exit(1);
    }

    //Find the z-coordinate of the top/bottom elements in this layer, depending
    //on whether the material layer is above or below the power layer
    //If it is above the power layer, then it is the minimum z that satisfies:
    //layer_zcoor - layer_zsize/2 < z_coor < layer_zcoor
    //Otherwise, it is the maximum z that satisfies
    //layer_zcoor < z_coor < layer_zcoor + layer_zsize/2
    
    vector<grid_contour> & c_contour = hoct_contour_[hoct_contour_.size()-1]; 

    double elt_zcoor;
    elt_zcoor = c_contour[0].self()->coors(Z_);
    if(dir == 'u') {
        for(int i = 1; i < N_; i++) {
            if(c_contour[i].self()->coors(Z_) < layer_zcoor &&
               c_contour[i].self()->coors(Z_) > layer_zcoor - layer_zsize/2.0) {
                if(c_contour[i].self()->coors(Z_) < elt_zcoor) {
                    elt_zcoor = c_contour[i].self()->coors(Z_);
                }
            }
        }
    }
    else {
        for(int i = 1; i < N_; i++) {
            if(c_contour[i].self()->coors(Z_) < layer_zcoor + layer_zsize/2.0 &&
               c_contour[i].self()->coors(Z_) > layer_zcoor ) {
                if(c_contour[i].self()->coors(Z_) > elt_zcoor) {
                    elt_zcoor = c_contour[i].self()->coors(Z_);
                }
            }
        }
    }

#ifdef DEBUG
    cout << "Z-coordinate of active layer = " << elt_zcoor << endl;
#endif

    //Now go through all the elements and find the ones with z_coor =
    //layer_zcoor and lie inside the box

    vector<int> layer_elts_ind;

    for(int i = 0; i < N_; i++) {
        if(fabs(c_contour[i].self()->coors(Z_) - elt_zcoor) < FMIN_) {
#ifdef DEBUG
            cout << c_contour[i].self()->coors(X_) << " ";
            cout << c_contour[i].self()->coors(Y_) << " ";
#endif
            if(c_contour[i].self()->coors(X_) > x_min &&
               c_contour[i].self()->coors(X_) < x_max &&
               c_contour[i].self()->coors(Y_) > y_min &&
               c_contour[i].self()->coors(Y_) < y_max) { 

                layer_elts_ind.push_back(i);
#ifdef DEBUG
                cout << "Element lies inside box";
#endif
            }
#ifdef DEBUG
            cout << endl;
#endif
        }
    }

    //Partition all the elements obtained homogeneously into 16x16x4
    for(unsigned int i = 0; i < layer_elts_ind.size(); i++) {
        FC_spatial_refinement(layer_elts_ind[i], 16, 16, 4);
    }
    
    FC_spatial_update(power);
 
}

//*===========================================================================*/
//* Further partition hoct_tree here as needed */
void Solution::refine_par(int comp, int layer, int x_div,
        int y_div, int z_div){

    /* Partition parameters. */
    vector<int> par_dim;
    par_dim.push_back(x_div);
    par_dim.push_back(y_div);
    par_dim.push_back(z_div);
    
    //*Refine the leaf nodes(the children of the layer nodes) based on
    //* the par_dim parameter. */
    //* Further partition the children of each layer grid (the children here 
    //* are leaf nodes of the root tree. */
    par_grid * curlayergrid = root_.int_chi_array(comp)->int_chi_array(layer);
    assert(curlayergrid);
    for (unsigned int k = 0; 
    k < curlayergrid->int_chi_array().size(); ++k){
	    gen_par_tree(curlayergrid->int_chi_array(k), par_dim);
    }
}

//*============================================================================*/
//* Further partition hoct_tree here as needed. */
void Solution::refine_par(int comp, int layer, int elt, int x_div,
    int y_div, int z_div){

    vector<int> par_dim;
    par_dim.push_back(x_div);
    par_dim.push_back(y_div);
    par_dim.push_back(z_div);
    
    /*Refine the leaf nodes(the children of the layer nodes) based on
     * the par_dim parameter */
    /* Further partition the children of each layer grid (the children here 
     * are leaf nodes of the root tree */
    par_grid * curlayergrid = root_.int_chi_array(comp)->int_chi_array(layer);
    assert(curlayergrid);
    assert(curlayergrid->int_chi_array(elt));
	gen_par_tree(curlayergrid->int_chi_array(elt), par_dim);

}

//*============================================================================*/
//* Check whether the children of the par grid are the neighbor of the
//* power layer(a vector of par grids) with a coor of z_pos. */
//* push the according grids into power_nei_ when correct
//* neighbors are found. */
//* Only check neighbor in Z_ direction. */
void Solution::check_nei(par_grid * par, vector<par_grid *> & cur_power_layer,
int direction){
    //* Here we already know the children are neighbors in Z dimension, so only
    //* need to check xy dimension. */
    int material_chi_nums = static_cast<int>(par->int_chi_array().size());
    int power_chi_nums = static_cast<int>(cur_power_layer.size());
    for (int i = 0; i < power_chi_nums; ++i){
        par_grid * cur_power_grid = cur_power_layer[i];
        //* First make sure the chi_grid is actually within the range of the 
        //* interface layer in x/y dimension. */
        for (int j = 0; j < material_chi_nums; j++){
			int dir_flag = par->int_chi_array(j) -> 
				xyintersect_check(cur_power_grid);
			if (dir_flag == 1){
                par->int_chi_array(j) -> power_nei().push_back
					(pair<int, par_grid *>(direction, cur_power_grid));
			}
        }
    }
}


/*===========================================================================*/
/* Initialize the power_ and temp_ with the correct size and 0.0 values*/
void Solution::init_power_temp(){

    int elts_nums = static_cast<int>(AGC_.A_.size());
    power_.resize(elts_nums);
    temp_.resize(elts_nums);
    for (int i = 0; i < elts_nums; ++i){
        power_[i] = 0.0;
        temp_[i] = 0.0;
    }
}

/*===========================================================================*/
//* Calcuate the mapped power value to the 3D elements from the 2D power 
//* input. */
void Solution::cal_power_value(int level, 
    const vector<vector<double> > & power){

    vector<double> power_vec;
    for (unsigned int i = 0; i < power.size();++i){
        for (unsigned int j = 0; j < power[i].size();++j){
            power_vec.push_back(power[i][j]);
        }
    }
    
    power_.clear();
    power_.resize(hoct_contour_[level].size());

    for (unsigned int i = 0; i < hoct_contour_[level].size(); ++i){
        power_[i] = 0.0;
        for (unsigned int j = 0; 
            j < hoct_contour_[level][i].self()->power_rela().size(); ++j){
            power_[i] += hoct_contour_[level][i].self()->power_rela(j).second
                *power_vec[hoct_contour_[level][i].self()->power_rela(j).first];
        }
    }

}


//*===========================================================================*/
//* Calcuate the coefficients that will be used to calculate the power vlaues
//* of the 3D elements from the 2D power input. */
void Solution::find_power_rela(){

    for (unsigned int i = 0; i < hoct_contour_.size(); ++i){
        for (unsigned int j = 0; j < hoct_contour_[i].size(); ++j){
            hoct_contour_[i][j].self()->power_rela().clear();
            find_power_rela_elt(hoct_contour_[i][j], i);
        }
    }
}


//*===========================================================================*/
//* Find the 3D elements that should be mapped with power values from the
//* 2D power input. */
void Solution::find_power_rela_elt(grid_contour & grid_cont, int level){

    int pelts_offset = 0;
    for (unsigned int i = 0; i < chip_->components.size();++i){
        for (unsigned int j = 0; j < chip_->components[i].layers.size();++j){
            if (chip_->components[i].layers[j].property==POWER_){
                char p_nei=layer_nei(grid_cont, 
                    chip_->components[i].layers[j].coor[Z_]);       
                if ((p_nei == 'u') || (p_nei =='d')){
                    cal_power_rela_coeff(grid_cont, p_nei, 
                        chip_->components[i].layers[j], pelts_offset, level);
                }
                pelts_offset += chip_->components[i].layers[j].power_profile.size();
            }
        }
    }
}


//*===========================================================================*/
//* Calculate the coefficient for power value calculation for one grid, which
//* is the neighbor of power_layer in Z direction. */
void Solution::cal_power_rela_coeff(grid_contour & grid_cont, char nei_rela,
    const Layer_type & power_layer, int pelt_index_offset, int level){
    
    const vector<Element_type> & power_pfile = power_layer.power_profile;
    par_grid * cur_grid = grid_cont.self();
    if (nei_rela == 'u'){
        if (fabs(power_layer.coor[Z_]) < FMIN_){
            for (unsigned int i = 0; i < power_pfile.size();++i){
                double xyar_intsect = area_intersect(cur_grid, power_pfile[i]); 
                double xyar_elt = power_pfile[i].size[X_] 
                    * power_pfile[i].size[Y_];
                if (xyar_intsect > DMIN_){
                    double xyar_share =xyar_intsect/xyar_elt;
                    int pelt_index = pelt_index_offset + i;
                    cur_grid->power_rela().push_back
                        (pair<int,double>(pelt_index, xyar_share));
                }
            }
        }else{
            for (unsigned int i = 0; i < power_pfile.size();++i){
                double gp_overlap = area_intersect(cur_grid, power_pfile[i]);
                double counted_gp_overlap = 0.0;
                double xyar_pelt = power_pfile[i].size[X_] 
                    * power_pfile[i].size[Y_];
                int pelt_index = pelt_index_offset + i;
                for (int j = 0; j < int(grid_cont.nei(Z_NEG_).size());++j){
	                par_grid * nei_grid = grid_cont.nei(Z_NEG_,j);
                    double ggp_overlap = area_intersect
                        (cur_grid, nei_grid, power_pfile[i]); 
                    if (ggp_overlap > FMIN_){
                        counted_gp_overlap += ggp_overlap;
                        double ggp_share = ggp_overlap/xyar_pelt;
    	                pair<double,double> resis=cal_eq_resis(cur_grid, nei_grid);
    	                double tot_resis = resis.first + resis.second;
    	                double coeff = (1 - resis.first/tot_resis) * ggp_share;
                        cur_grid->power_rela().
                            push_back(pair<int,double>(pelt_index, coeff));

                    }
                }
                double rem_gp_overlap = gp_overlap - counted_gp_overlap;
                if ((rem_gp_overlap)>FMIN_){
                    double coeff = rem_gp_overlap/xyar_pelt;     
                    cur_grid->power_rela().
                        push_back(pair<int,double>(pelt_index, coeff));
                }
            }
        }
    }else{
        for (unsigned int i = 0; i < power_pfile.size();++i){
            double gp_overlap = area_intersect(cur_grid, power_pfile[i]);
            double counted_gp_overlap = 0.0;
            double xyar_elt = power_pfile[i].size[X_] 
                * power_pfile[i].size[Y_];
            int pelt_index = pelt_index_offset + i;
            for (int j = 0; j < int(grid_cont.nei(Z_PLU_).size());++j){
	            par_grid * nei_grid = grid_cont.nei(Z_PLU_,j);
                double ggp_overlap = area_intersect
                    (cur_grid, nei_grid, power_pfile[i]); 
                if (ggp_overlap > FMIN_){
                    counted_gp_overlap += ggp_overlap;
                    double ggp_share = ggp_overlap/xyar_elt;
	                pair<double,double> resis=cal_eq_resis(cur_grid, nei_grid);
    	            double tot_resis = resis.first + resis.second;
    	            double coeff = (1 - resis.first/tot_resis) * ggp_share;
                    cur_grid->power_rela().
                        push_back(pair<int,double>(pelt_index, coeff));
                }
            }
            double rem_gp_overlap = gp_overlap - counted_gp_overlap;
            if ((rem_gp_overlap)>FMIN_){
                double coeff = rem_gp_overlap/xyar_elt;     
                cur_grid->power_rela().
                    push_back(pair<int,double>(pelt_index, coeff));
            }
        } 
    }
}


//*==========================================================================*/
//* Set the initial temperature to be 0.0 for all the thermal grids. */
void Solution::set_init_temp(){

	for (int i = 0; i < int(temp_.size()); ++i){
		temp_[i] = 0.0; 
	}
}


//*===========================================================================*/
//* Put the temperature valus back from temp_ to hoct_tree grids. */
void Solution::temperature_back_tree(){

	int elts_num = temp_.size() - btm_elts_;
	vector<grid_contour> & cur_contour = hoct_contour_[hoct_contour_.size()-1];
	for (int i = 0; i < elts_num; ++i){
		cur_contour[i].self()->temp_assign(temp_[i]);
	}
}


//*===========================================================================*/
//* Spatial refinement based on the thermal gradient.
//* Check the thermal gradient between each two neighbor elements, and refine
//* them if the thermal gradient between them are larger than the input 
//* thermal gradient threshold ("thermal_thre"). */
int Solution::spatial_refinement(double thermal_thre) {

    int global_flag = 0;

    vector<grid_contour> & c_contour = hoct_contour_[hoct_contour_.size()-1];
    for(int i = 0; i < N_; i++) {
        par_grid * cur_grid = c_contour[i].self();

        int nei_cnt = AGC_.A_[i].size();
        vector<int> par_dim;
        par_dim.resize(DIM_C_, 1);

        double cur_temp = temp_[i];
        int flag = 0;
        for(int j = 0; j < nei_cnt; j++) {
            int nei_index = AGC_.A_[i][j];
			if (nei_index >= 0 && nei_index < N_){
				par_grid * nei_grid = c_contour[nei_index].self();
				if(nei_grid->act_level() >= (init_level_ + 5)){
                    continue;
				}
				int dir_flag = cur_grid -> nei_check(nei_grid);
                if (i != nei_index){
                    assert(dir_flag >= 0);
                }
            	double ther_grad = fabs(cur_temp - temp_[nei_index]);
            	if(ther_grad > thermal_thre) {
                    int par = int(exp2(int(ceil(log2(ther_grad/thermal_thre)))));
                    if (cur_grid->sizes(X_) >= nei_grid->sizes(X_) && (dir_flag/2) == 0){
                        global_flag = 1;
                        flag = 1;
                   		par_dim[X_] = par>par_dim[X_]?par:par_dim[X_];
                    }
                    if (cur_grid->sizes(Y_) >= nei_grid->sizes(Y_) && (dir_flag/2) == 1){
                        global_flag = 1;
                        flag = 1;
                   		par_dim[Y_] = par>par_dim[Y_]?par:par_dim[Y_];
                    }
                    if (cur_grid->sizes(Z_) >= nei_grid->sizes(Z_) && (dir_flag/2) == 2){
                        //Disable adaptation in z-direction because it
                        //messes up the zooming in functions
                        //global_flag = 1;
                        flag = 1;
                   		par_dim[Z_] = par>par_dim[Z_]?par:par_dim[Z_];
                    }
            	}
			}
        }
	    if(flag) {
	        if(par_dim[X_]>=2){
	            par_dim[X_] = 2;
	        }
	        if(par_dim[Y_]>=2){
	            par_dim[Y_] = 2;
	        }
	        if(par_dim[Z_]>=2){
	            par_dim[Z_] = 1; //Disable adaptation in z-direction because it
                                 //messes up the zooming in functions
	        }
	    }
		gen_par_tree(cur_grid, par_dim);
    }
    spatial_update();

    return global_flag;
}


//*===========================================================================*/
//* Update the data structures after the spatial refinement. */
void Solution::spatial_update() {

	clear_contour(hoct_contour_);
	AGC_.clear();

	//* Update hoct_contour, together with the neighbor information. */
	hoct_contour_update(hoct_contour_, &root_);
	hoct_contour_neighbor(hoct_contour_);

	//* Update AGC */
	hoct_contour_AGC(hoct_contour_[hoct_contour_.size()-1], AGC_, 
		hoct_contour_.size()-1);
	//* Update the number of thermal elements. */
	N_ = AGC_.size_;

	//* Update the power and temp vectors. */
	power_.resize(N_, 0.0);
	temp_.resize(N_, 0.0);
	vector<grid_contour> & l_contour = hoct_contour_[hoct_contour_.size()-1];
	for (int i = 0; i < N_; ++i){
		power_[i] = 0.0; 
		temp_[i] = l_contour[i].self() -> temp();
	}

}


/*==========================================================================*/
/* Calculate the hcoeff_conv_. */
void Solution::cal_hcoeff(){
    
    double Rebstar = U_conv * b_conv * b_conv /(gamma_conv * L_conv);
    double Nuiterm1 = 1.0/pow((Rebstar*Pr_conv/2.0),3.0);
    double Nuiterm2 = 0.664 * sqrt(Rebstar)*pow(Pr_conv, 1.0/3.0)* 
        sqrt(1+3.65/sqrt(Rebstar));
    double Nuiterm3 = 1.0/pow(Nuiterm2, 3.0);
    double Nuiterm = Nuiterm1 + Nuiterm3;
    double Nui = pow(Nuiterm, -1.0/3.0);
    double Nubterm2 = 
        sqrt(2*Nui*kf_conv*H_conv*H_conv/(k_conv*b_conv*t_conv)*
        (t_conv/L_conv + 1.0));
    double Nubterm1 = tanh(Nubterm2);

    double Nub = Nubterm1/Nubterm2 * Nui;

    hcoeff_conv_ = Nub * kf_conv / b_conv;
   
}


//*========================================================================*/
//* Find out the smallest sizes for all the layers. */
double Solution::min_zsize(){

    double min_sizez = FMAX_;

    int comp_nums = chip_->components.size();
    for (int i = 0; i < comp_nums; ++i){
        Component_type & cur_comp = chip_->components[i];
        int layer_nums = cur_comp.layers.size();
        for (int j = 0; j < layer_nums; ++j){
            min_sizez = min(min_sizez, cur_comp.layers[j].size[Z_]);
        }
    }

    return min_sizez;
}


/*===========================================================================*/
/* Run the ssmoother iteratively among multiple AGCs, or several parts of 
 * AGCs */
void Solution::multiple_ssmoother(double err_thres){

	double diff_thres = 1E-2;
	int size = static_cast<int>(chip_->components.size());
	vector<bool> flag(size, true);
	solver_type sol(AGC_);

	int mul_iter = 0;
	while (flag_true(flag)){
		for (int i = 0; i < size; ++i){
			mul_iter++;
			int start_id = 0;
			if (i > 0){	
				start_id = elts_incomp_[i-1];
			}
			int end_id = elts_incomp_[i];
			vector<double> otemp(end_id - start_id, 0.0);
            for (int j = 0; j < (end_id-start_id); ++j){
                otemp[j] = temp_[j + start_id];
            }
			sol.ssmoother(AGC_, temp_, power_, err_thres, start_id, end_id);
			if (size >= 2){
				int adjstart_id = 0;
				int adjend_id = 0;
				if ((size-1) == i){	
					adjend_id = elts_incomp_[0];
				}else{
					adjstart_id = elts_incomp_[i];
					adjend_id = elts_incomp_[i+1];
				}
				if (max_diff(otemp, temp_, start_id) < diff_thres){
                	flag[i] = false;
            	}
			}else{
				flag[0] = false;
			}
		}
	}
}


/*============================================================================*/
/* Calculate the sizes and coordinates information based on top_zpos 
 * information */
void cal_sizecoor(vector<double> & sizes, vector<double> & coors, 
	const Component_type & component, const double & top_zpos){

	int layers_size = static_cast<int>(component.layers.size());
	sizes.resize(layers_size, 0.0);
	if (component.layers.size() > 0){
		sizes[0] = component.layers[0].size[0];
		sizes[1] = component.layers[0].size[1];
	}	
	coors.resize(layers_size, 0.0);
	for (int i = 0; i < layers_size; ++i){
		sizes[2] += component.layers[i].size[2];	
	}
	coors[2] = top_zpos + sizes[2]/2.0;

}


//*===========================================================================*/
//* Here adjust the convection effect(resistance) into the thermal
//* implementation by adding nodes into the AGC data structure(not
//* real grids to the tree or the hoct_contour). The added node also
//* presents resistance and heat capacity as other normal nodes. 
//* Since the power_ and temp_ vectors also need to be revised to conduct
//* the thermal modeling, according revisions needs to be done when the 
//* temp_ vector was mapped back to the tree (contour) */
//* Change the power_ and temp_ vectors here by adding the
// * additional elements. */
void Solution::broaden_AGC(const double ambient_resis, const double
    ambient_heatcap){

    if (ambient_resis < FMIN_){
        btm_elts_ = 0;
        return;
    }
    
    btm_elts_ = 0;
    int AGC_size = AGC_.A_.size();
    for (int i = 0; i < AGC_size; ++i){
        for (int j = 0; j < static_cast<int>(AGC_.A_[i].size()); ++j){
            if (AGC_.A_[i][j] == -1){//* If A_[i] means the
                 //conductance between current grid and the CONT_ neighbor. */
                par_grid * cur_grid =
                    hoct_contour_[hoct_contour_.size()-1][i].self();
                double btm_area = cur_grid->sizes(X_) * cur_grid->sizes(Y_);
                double old_cond = AGC_.G_[i][j];
                double new_cond = -1.0/(-1.0/AGC_.G_[i][j] +
                    ambient_resis/(2.0*btm_area));
                //* Adjust the according AGC data structure when considering
                //* the resistance&heat capacity from air flow. */
                AGC_.G_[i][j] = new_cond;
                AGC_.G_[i][0] += (old_cond - new_cond);
                int new_AGC_id = AGC_.A_.size();
                AGC_.A_[i][j] = new_AGC_id;
                vector<int> A_;
                vector<double> G_;
                A_.resize(3);
                G_.resize(3);
                A_[0] = new_AGC_id;
                A_[1] = i;
                A_[2] = -1;
                G_[1] = new_cond;
                G_[2] = -1.0/(ambient_resis/(2.0*btm_area));
                G_[0] = -(G_[1] + G_[2]);
                AGC_.A_.push_back(A_);
                AGC_.G_.push_back(G_);
                AGC_.C_.push_back(ambient_heatcap * btm_area);
                AGC_.size_ += 1;
                btm_elts_ ++;
            }
        }
    }

    //broaden_AGC_elt();
}

//*===========================================================================*/
void Solution::broaden_AGC_elt( ){

    int size = AGC_.A_.size() - btm_elts_;
    int deepest_level = hoct_contour_.size()-1;
    
    for (int i = size; i < (size+btm_elts_); ++i){
        grid_contour & cur_cont = hoct_contour_[deepest_level][i-btm_elts_];
            for (int j = 0; j < 4; ++j){
                int lid = cur_cont.nei(j,0)->levelid(deepest_level);
                cout <<"lid"<<lid<<endl;
                if (lid >=0){
                    AGC_.A_[i].push_back(lid+btm_elts_);
                    AGC_.G_[i].push_back(-0.3);
                    AGC_.G_[i][0] += 0.3;
                }

            }
    }

}

//*===========================================================================*/
//* Generate the correct power_ and temp_ information when considering 
//* convection effect. /
void Solution::broaden_temp_power(double temp_value){
    for (int i = 0; i < btm_elts_; ++i){
        power_.push_back(0.0);
        temp_.push_back(temp_value);
    }
    assert(power_.size() == temp_.size());
    assert(power_.size() == AGC_.A_.size());
}


//*===========================================================================*/
//* Partition all the elements into 2 in Z dimension */
void Solution::refine_zpar(){
   
    vector<int> par;
    for (int i = 0; i < DIM_C_; ++i){
        par.push_back(1);
    } 
    par[Z_] = 2;
    for (unsigned int i = 0; i < root_.int_chi_array().size(); ++i){
        for (unsigned int j = 0; j < 
            root_.int_chi_array(i)->int_chi_array().size(); ++j){
            refine_par(i, j, par[X_], par[Y_], par[Z_]);
        }
    }

    spatial_update();
}

//*===========================================================================*/
//* Do custom spatial refinement for any required element. */
void Solution::FC_spatial_refinement(int elem_index, int x_div, int y_div,
        int z_div) {
    
    vector<grid_contour> & c_contour = hoct_contour_[hoct_contour_.size()-1];

    int i = elem_index;
        
    par_grid * cur_grid = c_contour[i].self();

    int nei_cnt = AGC_.A_[i].size();
    for(int j = 0; j < nei_cnt; j++) {
        int nei_index = AGC_.A_[i][j];
        if (nei_index >= 0 && nei_index < N_){
            par_grid * nei_grid = c_contour[nei_index].self();
            if(nei_grid->self_level() >= (init_level_ + 5)){
                continue;
            }
            int dir_flag = cur_grid -> nei_check(nei_grid);
            if (i != nei_index){
                assert(dir_flag >= 0);
            }
        }
    }

    vector<int> par_dim;
    par_dim.resize(DIM_C_, 1);
    par_dim[X_] = x_div;
    par_dim[Y_] = y_div;
    par_dim[Z_] = z_div;

    gen_par_tree(cur_grid, par_dim);
    //spatial_update();
}


/*===========================================================================*/
void Solution::FC_spatial_update(
    const vector<vector<double> > & power) {
    
    spatial_update();
    
    int cont_level = hoct_contour_.size() - 1;
    find_power_rela();
    cal_power_value(cont_level, power);
}


