/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** solver.cc: code for the solver_type class
****************************************************************************/

#include <vector>
#include <cmath>
#include <assert.h>
#include <sys/time.h>
#include "solver.h"
#include "util.h"
#include "hoct_model.h"

using namespace std;

/*##########################################################################*/
/* solver_type constructor */
solver_type::solver_type():
G_(),
Gx_(),
f_(),
residue_()
{

}


/*===========================================================================*/
/* solver_type constructor */
solver_type::solver_type(RC_network_type& AGC):
G_(),
Gx_(),
f_(),
residue_()
{
	int size = AGC.size_;
	Gx_.resize(size, 0);
	G_.resize(size);

    for(int i = 0; i < size; i++) {
		int nei_cnt = AGC.G_[i].size();
		G_[i].reserve(nei_cnt);
		double d_inv_ = 1.0/AGC.G_[i][0];
		G_[i].push_back(d_inv_);
		for(int j = 1; j < nei_cnt; j++) {
			G_[i].push_back(d_inv_*(-AGC.G_[i][j]));
		}
	}
}


/*===========================================================================*/
/* Deconstructor */
solver_type::~solver_type() 
{

}


/*===========================================================================*/
/* Iterative Gauss Seidel method to calculate T in AT=P */
void solver_type::ssmoother( RC_network_type & AGC,
	vector<double>& temp, vector<double>& power, double err_thre) {

	error_type errors(err_thre);
	int size = AGC.size_;
	int iter = 0;

	vector<double> errs(size, -1.0);

	vector<vector<int> > & A = AGC.A_;
	while (true) {
		iter++;
        //cout << iter << endl;
		for(register int i = 0; i < size; i++) {
			double & Gx_i = Gx_[i];
			Gx_i = 0;
			vector<int> & AI = A[i];
			int nei_cnt = AI.size();
			const vector<double> & G_i = G_[i];
			for (register int j = 1; j < nei_cnt; j++) {
				int id_ = AI[j];
				if(id_ >= 0){
					Gx_i += G_i[j] * temp[id_];
				}
			}
			double new_temp = Gx_i;
			new_temp += G_i[0]*power[i];
			double & temp_i = temp[i];
			errors.update_flag(fabs(new_temp - temp_i));
			errs[i] = fabs(new_temp - temp_i);
            /*#warning test the successive over-relaxation here */
			double w_coeff = 1.9;
			temp_i = (1.0 - w_coeff) * temp_i + w_coeff * new_temp;
			//temp_i = new_temp;
		} 
		if(!errors.flag()) {
			break;
		}
		errors.reset_flag();
	}//end while

}

/*===========================================================================*/
/* Iterative Gauss Seidel method to calculate T in AT=P */
void solver_type::ssmoother2( RC_network_type & AGC,
	vector<double>& temp, vector<double>& power, double err_thre) {

	error_type errors(err_thre);
	int size = AGC.size_;
	int iter = 0;
    int iter_max = 10000;

	vector<double> errs(size, -1.0);
	double w_coeff = 1.9;

	vector<vector<int> > & A = AGC.A_;
	while (true) {
		iter++;
        //cout << iter << endl;
		for(register int i = 0; i < size; i++) {
			double & Gx_i = Gx_[i];
			Gx_i = 0;
			vector<int> & AI = A[i];
			int nei_cnt = AI.size();
			const vector<double> & G_i = G_[i];
			for (register int j = 1; j < nei_cnt; j++) {
				int id_ = AI[j];
				if(id_ >= 0){
					Gx_i += G_i[j] * temp[id_];
				}
			}
			double new_temp = Gx_i;
			new_temp += G_i[0]*power[i];
			double & temp_i = temp[i];
			errors.update_flag(fabs(new_temp - temp_i));
			errs[i] = fabs(new_temp - temp_i);
            /*#warning test the successive over-relaxation here */
            if(iter > iter_max && (fabs(w_coeff - 1.0) > 1.0E-7)) {
                w_coeff = 1.0;
            }
			temp_i = (1.0 - w_coeff) * temp_i + w_coeff * new_temp;
			//temp_i = new_temp;
		} 
		if(!errors.flag()) {
			break;
		}
		errors.reset_flag();
	}//end while

}



/*===========================================================================*/
/* Iterative Gauss Seidel method to calculate T in AT=P */
/* Do the smoother inside the AGC accoring to the start and end id (only do
 * the smoother for those elements that reside inside {start_id, end_id} */
void solver_type::ssmoother( RC_network_type & AGC,
    vector<double>& temp, vector<double>& power, double err_thre,
    int start_id, int end_id) {

    error_type errors(err_thre);
    int size = AGC.size_;
    int iter = 0;

    vector<double> errs(size, -1.0);

    vector<vector<int> > & A = AGC.A_;
    while (true) {
        iter++;
        for(register int i = start_id; i < end_id; i++) {
            double & Gx_i = Gx_[i];
            Gx_i = 0;
            vector<int> & AI = A[i];
            int nei_cnt = AI.size();
            const vector<double> & G_i = G_[i];
            for (register int j = 1; j < nei_cnt; j++) {
                int id = AI[j];
            //    if((id >= 0) && (id >= start_id) && (id < end_id)){
				if (id >= 0){
                    Gx_i += G_i[j] * temp[id];
				}
            //    }
            }
            double new_temp = Gx_i;
            new_temp += G_i[0]*power[i];
            double & temp_i = temp[i];
            errors.update_flag(fabs(new_temp - temp_i));
			cout << "diff " << new_temp - temp_i << endl;
            errs[i] = fabs(new_temp - temp_i);
            /*#warning test the successive over-relaxation here */
            double w_coeff = 1.9;
            temp_i = (1.0 - w_coeff) * temp_i + w_coeff * new_temp;
            temp_i = new_temp;
        }
        if(!errors.flag()) {
            break;
        }
        errors.reset_flag();
    }//end while

}


/*===========================================================================*/
/* Iterative Gauss Seidel method to calculate T in AT=P */
void solver_type::ssmoother( RC_network_type & AGC,
	vector<double>& temp, vector<double>& power, double err_thre, int index) {

	error_type errors(err_thre);
	int size = AGC.size_;

	vector<vector<int> > & A = AGC.A_;

	while (true) {
		for(register int i = 0; i < size; i++) {
			double & Gx_i = Gx_[i];
			Gx_i = 0;
			vector<int> & AI = A[i];
			int nei_cnt = AI.size();
			const vector<double> & G_i = G_[i];
			for (register int j = 1; j < nei_cnt; j++) {
				int id_ = AI[j];
				if(id_ >= 0){
					Gx_i += G_i[j] * temp[id_];
				}
			}
			double new_temp = Gx_i;
			if(i == index) {
				new_temp += G_i[0]*power[i];
			}
			double & temp_i = temp[i];
			errors.update_flag(fabs(new_temp - temp_i));
			temp_i = new_temp;
		}
		if(!errors.flag()) {
			break;
		}
		errors.reset_flag();
	}
}


/*===========================================================================*/
/* Iterative Gauss Seidel method to calculate T in AT=P */
void solver_type::ssmoother(vector<vector<int> >& A, 
	vector<double>& temp, vector<double>& power, double err_thre, int index) {

	error_type errors(err_thre);
	int size = A.size();

	while (true) {
		for(register int i = 0; i < size; i++) {
			double & Gx_i = Gx_[i];
			Gx_i = 0;
			vector<int> & AI = A[i];
			int nei_cnt = AI.size();
			const vector<double> & G_i = G_[i];
			for (register int j = 1; j < nei_cnt; j++) {
				int id_ = AI[j];
				if(id_ >= 0){
					Gx_i += G_i[j] * temp[id_];
				}
			}
			double new_temp = Gx_i;
			if(i == index) {
				new_temp += G_i[0]*power[i];
			}
			double & temp_i = temp[i];
			errors.update_flag(fabs(new_temp - temp_i));
			temp_i = new_temp;
		}
		if(!errors.flag()) {
			break;
		}
		errors.reset_flag();
	}//end while
}

